//-------------------------------------------------------------------------------------------------
// <copyright file="BurnController.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"

enum WM_BURN
{
    WM_BURN_FIRST = WM_APP + 0xFFF, // this enum value must always be first.

    WM_BURN_VIEW_INITIALIZED,
    WM_BURN_PREPARE_PACKAGES,
    WM_BURN_CACHE_PACKAGES,
    WM_BURN_EXECUTE_PACKAGES,

    WM_BURN_LAST, // this enum value must always be last.
};


class CBurnController : public CUnknownImpl2<IBurnController, IBurnControllerTestable>
{
public: // IBurnController
    //
    // Run - start doing that which the controller does.
    //
    STDMETHODIMP Run()
    {
        HRESULT hr = S_OK;
        BURN_COMMAND command;
        BURN_MODE mode;

        // First, know what the Model expects of us.
        m_pModel->GetCommand(&command);
        mode = m_pModel->Mode();

        // Ensure the last requested state is reset.
        m_LastRequested = REQUEST_STATE_NONE;

        if (BURN_MODE_ELEVATED == mode)
        {
            hr = RunViewElevated();
            ExitOnFailure(hr, "Failure while running elevated.");
        }
        else if (BURN_MODE_EMBEDDED == mode)
        {
            hr = E_NOTIMPL;
            ExitOnRootFailure(hr, "Failure while running embedded.");
        }
        else
        {
            hr = RunViewUX();
            ExitOnFailure(hr, "Failure while running user experience.");
        }

    LExit:
        // Schedule restart if necessary.
        if (BURN_RESTART_NEVER != command.restart)
        {
            if (BURN_RESTART_ALWAYS == command.restart || (m_fRestartRequested && BURN_RESTART_AUTOMATIC == command.restart) || (m_fRestartAllowed  && BURN_RESTART_PROMPT == command.restart))
            {
                m_pInstallationAbstractionLayer->ScheduleRestart();
            }
        }

        return hr;
    }


    //
    // LoadUXInterface - loads the UX dll to get the IBurnUserExperience interface.
    //
    STDMETHODIMP LoadUXInterface(
        __out IBurnUserExperience** ppUX
        )
    {
        HRESULT hr = S_OK;
        HMODULE hUXModule = NULL;
        LPWSTR sczUXDLL = NULL;
        BURN_COMMAND command = {};

        hr = m_pModel->GetCommand(&command);
        ExitOnFailure(hr, "Failed to get command from model.");

        hr = CacheUXPayloads(&sczUXDLL);
        ExitOnFailure(hr, "Failed to cache UX payloads.");

        hUXModule = ::LoadLibraryW(sczUXDLL);
        ExitOnNullWithLastError(hUXModule, hr, "Failed to load UX dll.");

        PFN_CREATE_USER_EXPERIENCE pfn = reinterpret_cast<PFN_CREATE_USER_EXPERIENCE>(::GetProcAddress(hUXModule, "SetupUXCreate"));
        ExitOnNullWithLastError(pfn, hr, "Failed to get create user experience entry point.");

        hr = (*pfn)(&command, ppUX);
        ExitOnFailure(hr, "Failed to create user experience interface.");

        m_hUXModule = hUXModule;
        hUXModule = NULL;

    LExit:
        ReleaseStr(sczUXDLL);
        if (hUXModule)
        {
            ::FreeLibrary(hUXModule);
        }

        return hr;
    }


    //
    // Prepare - command to start preparing packages.
    //
    STDMETHODIMP Prepare(
        __in DWORD dwPackageIndex,
        __in REQUEST_STATE requested
        )
    {
        HRESULT hr = S_OK;

        Trace(REPORT_STANDARD, "CBurnController::Prepare()");

        if (!::PostThreadMessageW(m_dwThreadId, WM_BURN_PREPARE_PACKAGES, dwPackageIndex, requested))
        {
            ExitWithLastError(hr, "Failed to post prepare packages message.");
        }

    LExit:
        return hr;
    }


    //
    // Cache - command to start caching.
    //
    STDMETHODIMP Cache(
        __in DWORD dwPackageIndex
        )
    {
        HRESULT hr = S_OK;

        Trace(REPORT_STANDARD, "CBurnController::Cache()");

        if (!::PostThreadMessageW(m_dwThreadId, WM_BURN_CACHE_PACKAGES, dwPackageIndex, 0))
        {
            ExitWithLastError(hr, "Failed to post cache packages message.");
        }

    LExit:
        return hr;
    }


    //
    // CacheElevatedPackage - ensures an elevated package is cached and 
    //
    STDMETHODIMP CacheElevatedPackage(
        __in DWORD dwPackageIndex,
        __in ACTION_STATE execute,
        __in ACTION_STATE rollback
        )
    {
        HRESULT hr = S_OK;
        IBurnPackage* pPackage = NULL;

        Trace(REPORT_STANDARD, "CBurnController::ExecuteElevatedPackage()");

        hr = m_pModel->GetPackage(dwPackageIndex, &pPackage);
        ExitOnFailure1(hr, "Failed to find package with index: %u", dwPackageIndex);

        // TODO: cache the elevated pacakge in the locked down location.
        //hr = CacheCompletePackage(pPackage);
        //ExitOnFailure(hr, "Failed to cache package elevated.");

        pPackage->SetActionState(execute, rollback); // remember the action states for the package.

    LExit:
        ReleaseObject(pPackage);
        return hr;
    }


    //
    // Elevate - re-launches burn elevated with a pipe to communicate back and forth.
    //
    STDMETHODIMP Elevate(
        __in_opt HWND hwndParent
        )
    {
        HRESULT hr = S_OK;

        if (m_hElevatedProcess || INVALID_HANDLE_VALUE != m_hElevatedPipe)
        {
            hr = HRESULT_FROM_WIN32(ERROR_ALREADY_INITIALIZED);
            ExitOnRootFailure(hr, "Already launched elevated process and/or created elevated pipe.");
        }

        hr = ElevationParentProcessConnect(hwndParent, m_pInstallationAbstractionLayer, &m_hElevatedProcess, &m_hElevatedPipe);
        ExitOnFailure(hr, "Failed to connect to elevated process.");

    LExit:
        return hr;
    }


    //
    // Execute - command to start execution.
    //
    STDMETHODIMP Execute(
        __in REQUEST_STATE requested
        )
    {
        HRESULT hr = S_OK;

        Trace(REPORT_STANDARD, "CBurnController::Execute()");

        if (!::PostThreadMessageW(m_dwThreadId, WM_BURN_EXECUTE_PACKAGES, 0, requested))
        {
            ExitWithLastError(hr, "Failed to post execute packages message.");
        }

    LExit:
        return hr;
    }


    //
    // ExecuteElevatedPackage - start execution for an elevated package.
    //
    STDMETHODIMP ExecuteElevatedPackage(
        __in DWORD dwPackageIndex,
        __in BOOL fExecute
        )
    {
        HRESULT hr = S_OK;
        IBurnPackage* pPackage = NULL;
        BOOL fRestart = FALSE;

        Trace(REPORT_STANDARD, "CBurnController::ExecuteElevatedPackage()");

        hr = m_pModel->GetPackage(dwPackageIndex, &pPackage);
        ExitOnFailure1(hr, "Failed to find package with index: %u", dwPackageIndex);

        hr = ExecutePackage(pPackage, fExecute, TRUE, &fRestart, NULL);
        ExitOnFailure(hr, "Failed to execute package elevated.");

    LExit:
        ReleaseObject(pPackage);
        return hr;
    }


private: // privates
    //
    // RunViewElevated - runs the "elevated" view. This view takes over the
    //                   main thread. Since the view is in lock step with
    //                   this controller there is no need to separate them
    //                   with a thread.
    //
    HRESULT RunViewElevated()
    {
        HRESULT hr = S_OK;
        BOOL fInitializedView = FALSE;

        hr = m_pView->Initialize(m_pModel, this);
        ExitOnFailure(hr, "Failed to initialize view elevated.");

        fInitializedView = TRUE;

        hr = m_pView->Run();
        ExitOnFailure(hr, "Failed while running view elevated.");

    LExit:
        if (fInitializedView)
        {
            m_pView->Uninitialize();
        }

        return hr;
    }


    //
    // RunViewUX - runs the "user experience" view. This view is
    //             moved out to a separate thread to allow the
    //             controller to process long running tasks (like
    //             a download or install) while the view UX thread
    //             remains unblocked.
    //
    HRESULT RunViewUX()
    {
        HRESULT hr = S_OK;
        MSG msg;
        HANDLE hUIThread = NULL;
        DWORD dwUIThreadId = 0;

        // Initialize this thread's message queue and create the UI thread.
        Assert(0 == m_dwThreadId);
        m_dwThreadId = ::GetCurrentThreadId();
        ::PeekMessageW(&msg, NULL, WM_USER, WM_USER, PM_NOREMOVE);

        hUIThread = ::CreateThread(NULL, 0, CBurnController::UIThread, this, 0, &dwUIThreadId);
        ExitOnNullWithLastError(hUIThread, hr, "Failed to create UI thread.");

        // Message pump.
        for (;;)
        {
            // Wait for the UI thread to exit or a message to be posted.
            DWORD dwWait = ::MsgWaitForMultipleObjects(1, &hUIThread, FALSE, INFINITE, QS_POSTMESSAGE);
            if (WAIT_OBJECT_0 == dwWait) // UI thread exited, quit.
            {
                if (!::GetExitCodeThread(hUIThread, reinterpret_cast<DWORD*>(&hr)))
                {
                    hr = S_FALSE;
                }
                break; // break out of the loop completely
            }
            else if (WAIT_OBJECT_0 + 1 == dwWait) // message posted.
            {
                while (::PeekMessageW(&msg, reinterpret_cast<HWND>(-1), WM_BURN_FIRST, WM_BURN_LAST, PM_REMOVE | PM_QS_POSTMESSAGE))
                {
                    ProcessMessage(&msg);
                }
            }
            else // error.
            {
                ExitWithLastError(hr, "Unexpected failure while waiting for message.");
            }
        }

    LExit:
        if (hUIThread)
        {
            ::CloseHandle(hUIThread);
        }

        m_dwThreadId = 0;
        return hr;
    }


    //
    // ProcessMessage - called to process a controller message.
    //
    void ProcessMessage(
        __in const MSG* pmsg
        )
    {
        switch (pmsg->message)
        {
        case WM_BURN_VIEW_INITIALIZED:
            break;

        case WM_BURN_PREPARE_PACKAGES:
            OnPreparePackages(static_cast<DWORD>(pmsg->wParam), static_cast<REQUEST_STATE>(pmsg->lParam));
            break;

        case WM_BURN_CACHE_PACKAGES:
            OnCachePackages(static_cast<DWORD>(pmsg->wParam));
            break;

        case WM_BURN_EXECUTE_PACKAGES:
            OnExecutePackages(static_cast<REQUEST_STATE>(pmsg->lParam));
            break;

        default:
            TraceError1(E_UNEXPECTED, "Unexpected controller message: %u", pmsg->message);
            break;
        }
    }


    //
    // OnPreparePackages - calculate package states.
    //
    void OnPreparePackages(
        __in DWORD dwPackageId,
        __in REQUEST_STATE requested
        )
    {
        HRESULT hr = S_OK;
        IBurnProperties* pProperties = NULL;
        IBurnPackage* pPackage = NULL;

        Trace(REPORT_STANDARD, "CBurnController::OnPreparePackages() - enter");

        m_pModel->GetProperties(&pProperties);

        if (dwPackageId)
        {
            int nResult = m_pView->OnPrepareBegin(1);
            hr = HRESULT_FROM_VIEW(nResult);
            ExitOnRootFailure(hr, "UX aborted prepare begin.");

            hr = m_pModel->GetPackage(dwPackageId, &pPackage);
            ExitOnFailure1(hr, "Failed to find package with id: %u", dwPackageId);

            hr = PreparePackage(pPackage, pProperties, requested);
            ExitOnFailure(hr, "Failed to prepare package.");

            // Clear the last requested state any time we prepare only one package with a different
            // requested state, since at least one package doesn't match the overall plan.
            if (m_LastRequested != requested)
            {
                m_LastRequested = REQUEST_STATE_NONE;
            }
        }
        else
        {
            DWORD cPackages = m_pModel->PackageCount();

            int nResult = m_pView->OnPrepareBegin(cPackages);
            hr = HRESULT_FROM_VIEW(nResult);
            ExitOnRootFailure(hr, "UX aborted prepare begin.");

            hr = PreparePackages(pProperties, requested);
            ExitOnFailure(hr, "Failed to prepare all packages.");

            // Since all packages should be aligned to the same requested state
            // remember it.
            m_LastRequested = requested;
        }

    LExit:
        m_pView->OnPrepareComplete(hr);

        ReleaseObject(pPackage);
        ReleaseObject(pProperties);

        Trace(REPORT_STANDARD, "CBurnController::OnPreparePackages() - exit");
        return;
    }


    //
    // OnCachePackages - extract/download/copy-locally packages.
    //
    STDMETHODIMP_(void) OnCachePackages(
        __in DWORD dwPackageId
        )
    {
        HRESULT hr = S_OK;
        DWORD cCachePackages = 0;
        IBurnPackage** prgpCachePackages = NULL;
        DWORD cPackages = m_pModel->PackageCount();
        IBurnPackage* pPackage = NULL;
        IBurnPayload* pPayload = NULL;
        DWORD cCachePayloads = 0;
        DWORD64 dw64TotalCacheSize = 0;

        Trace(REPORT_STANDARD, "CBurnController::OnCachePackages() - enter");

        prgpCachePackages = static_cast<IBurnPackage**>(MemAlloc(sizeof(IBurnPackage*) * cPackages, TRUE));
        ExitOnNull(prgpCachePackages, hr, E_OUTOFMEMORY, "Failed to allocate memory to track packages to cache.");

        if (dwPackageId)
        {
            cPackages = 1;
            cCachePackages = 1;

            hr = m_pModel->GetPackage(dwPackageId, prgpCachePackages + 0);
            ExitOnFailure1(hr, "Failed to find package with id: %u", dwPackageId);
        }
        else
        {
            // Add all packages that have uncached payloads to the "cache package" array and
            // calculate the total size to be cached.
            for (DWORD i = 0; i < cPackages; ++i)
            {
                hr = m_pModel->GetPackage(i + 1, &pPackage);
                ExitOnFailure1(hr, "Failed to find package with index: %u", i);

                // Calculate the amount to be cached across the package.
                DWORD64 dw64CacheSize = 0;
                DWORD cPackagePayloadsToCache = 0;
                for (DWORD j = 1; j <= pPackage->PayloadCount(); ++j)
                {
                    hr = pPackage->GetPayload(j, &pPayload);
                    ExitOnFailure1(hr, "Failed to find payload for index: %u", j);

                    BOOL fCached = FALSE;
                    hr = pPayload->IsCached(&fCached);
                    ExitOnFailure(hr, "Failed to check if payload is cached.");

                    if (!fCached)
                    {
                        dw64CacheSize += pPayload->Size();
                        ++cPackagePayloadsToCache;
                    }

                    ReleaseObject(pPayload);
                }

                // If the package had anything to cache...
                if (cPackagePayloadsToCache)
                {
                    dw64TotalCacheSize += dw64CacheSize;

                    prgpCachePackages[cCachePackages] = pPackage;
                    pPackage = NULL;

                    cCachePayloads += cPackagePayloadsToCache;
                    ++cCachePackages;
                }

                ReleaseNullObject(pPackage);
            }
        }

        int nResult = m_pView->OnCacheBegin(cCachePackages, cCachePayloads, dw64TotalCacheSize);
        hr = HRESULT_FROM_VIEW(nResult);
        ExitOnRootFailure(hr, "UX aborted cache begin.");

        for (DWORD i = 0; i < cCachePackages; ++i)
        {
            hr = CachePackage(prgpCachePackages[i]);
            ExitOnFailure(hr, "Failed to cache package.");
        }

    LExit:
        m_pView->OnCacheComplete(hr);

        ReleaseObject(pPackage);

        if (prgpCachePackages)
        {
            for (DWORD i = 0; i < cCachePackages; ++i)
            {
                ReleaseObject(prgpCachePackages[i]);
            }

            MemFree(prgpCachePackages);
        }

        Trace(REPORT_STANDARD, "CBurnController::OnCachePackages() - exit");
        return;
    }


    //
    // OnExecutePackages
    //
    void OnExecutePackages(
        __in REQUEST_STATE requested
        )
    {
        HRESULT hr = S_OK;
        IBurnProperties* pProperties = NULL;
        DWORD cPackages = 0;
        DWORD cChangingPackages = 0;
        DWORD cExecutedPackages = 0;
        IBurnPackage* pPackage = NULL;
        ACTION_STATE execute;
        ACTION_STATE rollback;
        HANDLE hChangeOfOwnerEvent = NULL;

        BOOL fRestart = FALSE;
        BOOL fRollback = FALSE;
        BOOL fMsiMultiPackage = FALSE;

        Trace(REPORT_STANDARD, "CBurnController::OnExecutePackages() - enter");

        m_pModel->GetProperties(&pProperties);

        // If a new requested state was requested, re-prepare all of the packages.
        if (REQUEST_STATE_NONE != requested && m_LastRequested != requested)
        {
            hr = PreparePackages(pProperties, requested);
            m_pView->OnPrepareComplete(hr);
            ExitOnFailure(hr, "Failed to prepare packages before execution.");
        }

        // Forward, charge!
        cPackages = m_pModel->PackageCount();

        // Count the number of packages that actually have change associated with them.
        for (DWORD i = 0; i < cPackages; ++i)
        {
            m_pModel->GetPackage(i + 1, &pPackage);

            pPackage->GetActionState(&execute, &rollback);
            if (ACTION_STATE_NONE < execute || ACTION_STATE_NONE < rollback)
            {
                ++cChangingPackages;
            }

            ReleaseNullObject(pPackage);
        }

        // Begin.
        int nResult = m_pView->OnExecuteBegin(cChangingPackages);
        hr = HRESULT_FROM_VIEW(nResult);
        ExitOnRootFailure(hr, "UX aborted execute begin.");

        hr = m_pMsiEngine->BeginTransaction(&hChangeOfOwnerEvent);
        if (E_NOTIMPL == hr)
        {
            TraceError(hr, "Multi package begin transaction not supported.");
            hr = S_OK;
        }
        else if (SUCCEEDED(hr))
        {
            fMsiMultiPackage = TRUE;
        }
        else
        {
            ExitOnFailure(hr, "Failed to begin transaction");
        }

        // Actually execute the packages.
        for (cExecutedPackages = 0; cExecutedPackages < cPackages; ++cExecutedPackages)
        {
            hr = m_pModel->GetPackage(cExecutedPackages + 1, &pPackage);
            if (FAILED(hr))
            {
                TraceError1(hr, "Failed to find package with index: %u", cExecutedPackages);
                break;
            }

            pPackage->GetActionState(&execute, &rollback);
            if (ACTION_STATE_NONE < execute || ACTION_STATE_NONE < rollback)
            {
                hr = ExecutePackage(pPackage, TRUE, FALSE, &fRestart, &fRollback);
                if (fRollback)
                {
                    TraceError(hr, "Failed to execute package.");
                    break;
                }
                if (fMsiMultiPackage && WAIT_OBJECT_0 == ::WaitForSingleObject(hChangeOfOwnerEvent, 0))
                {
                    // Some other process took the transaction. Take it back.
                    ::ResetEvent(hChangeOfOwnerEvent);
                    hr = m_pMsiEngine->JoinTransaction(&hChangeOfOwnerEvent);
                }
            }

            ReleaseNullObject(pPackage);
        }

        // If we need to rollback, do so.
        if (fRollback)
        {
            ReleaseNullObject(pPackage);

            if (fMsiMultiPackage)
            {
                hr = m_pMsiEngine->EndTransaction(INSTALL_TRANSACTION_ROLLBACK);
                ExitOnFailure(hr, "Failed to rollback transaction.");
            }
            else
            {
                for (DWORD index = cExecutedPackages; 0 < index; --index)
                {
                    m_pModel->GetPackage(index, &pPackage);

                    // TODO: If pPackage is a "transaction boundary package" break out.

                    pPackage->GetActionState(&execute, &rollback);
                    if (ACTION_STATE_NONE < execute || ACTION_STATE_NONE < rollback)
                    {
                        ExecutePackage(pPackage, FALSE, FALSE, &fRestart, NULL);
                    }

                    ReleaseNullObject(pPackage);
                }
            }
        }

        if (fMsiMultiPackage)
        {
            hr = m_pMsiEngine->EndTransaction(INSTALL_TRANSACTION_COMMIT);
            ExitOnFailure(hr, "Failed to commit transaction.");
        }

        // If a restart is required and we haven't schedule a restart already, notify the
        // view and let it decide whether to allow the restart.
        if (fRestart && !m_fRestartRequested)
        {
            m_fRestartRequested = TRUE;
            m_fRestartAllowed = m_pView->OnRestartRequired();
        }

    LExit:
        m_pView->OnExecuteComplete(hr);

        ReleaseObject(pPackage);
        ReleaseObject(pProperties);

        Trace(REPORT_STANDARD, "CBurnController::OnExecutePackages() - exit");
        return;
    }


    //
    // PreparePackages - loops through all packages preparing each one to the requested state.
    //
    HRESULT PreparePackages(
        __in IBurnProperties* pProperties,
        __in REQUEST_STATE requested
        )
    {
        HRESULT hr = S_OK;
        IBurnPackage* pPackage = NULL;
        DWORD cPackages = m_pModel->PackageCount();

        for (DWORD i = 0; i < cPackages; ++i)
        {
            hr = m_pModel->GetPackage(i + 1, &pPackage);
            ExitOnFailure1(hr, "Failed to find package with index: %u", i);

            hr = PreparePackage(pPackage, pProperties, requested);
            ExitOnFailure(hr, "Failed to prepare package.");

            ReleaseNullObject(pPackage);
        }

    LExit:
        return hr;
    }


    //
    // PreparePackage - detect and plan the next step for the provided package.
    //
    HRESULT PreparePackage(
        __in IBurnPackage* pPackage,
        __in IBurnProperties* pProperties,
        __in REQUEST_STATE requested
        )
    {
        HRESULT hr = S_OK;
        IBurnExecuteEngine* pEngine = NULL;
        PACKAGE_STATE state = PACKAGE_STATE_UNKNOWN;
        DWORD64 qwInstalledVersion = 0;
        ACTION_STATE execute = ACTION_STATE_NONE;
        ACTION_STATE rollback = ACTION_STATE_NONE;

        int nResult = m_pView->OnPreparePackageBegin(pPackage);
        hr = HRESULT_FROM_VIEW(nResult);
        ExitOnRootFailure(hr, "UX aborted prepare package begin.");

        hr = GetExecuteEngineForPackage(pPackage, &pEngine);
        ExitOnFailure(hr, "Failed to get execute engine to prepare.");

        // If we haven't detected this package yet, do so.
        if (!pPackage->IsDetected())
        {
            hr = pEngine->Detect(pPackage, pProperties, &state, &qwInstalledVersion);
            ExitOnFailure(hr, "Failed to detect package.");

            pPackage->Detected(state, qwInstalledVersion);
        }

        // If a particular requested state was, uh, requested then plan it out.
        if (REQUEST_STATE_NONE != requested)
        {
            hr = PlanPackage(pEngine, pPackage, pProperties, requested, &execute, &rollback);
            ExitOnFailure(hr, "Failed to plan package.");
        }

        // Always reset the package "activity" states based on the passed in paramter.
        pPackage->SetActionState(execute, rollback);

    LExit:
        m_pView->OnPreparePackageComplete(pPackage, hr, state, requested, execute, rollback);

        ReleaseObject(pEngine);
        return hr;
    }


    //
    // PlanPackage - use the provided engine to determine execute and rollback action states for a package.
    //
    HRESULT PlanPackage(
        __in IBurnExecuteEngine* pEngine,
        __in IBurnPackage* pPackage,
        __in IBurnProperties* pProperties,
        __in REQUEST_STATE requested,
        __out ACTION_STATE* pExecute,
        __out ACTION_STATE* pRollback
        )
    {
        HRESULT hr = S_OK;

#if 0
        // If the package is permanent but we're asked to uninstall it, take no action.
        if (pPackage->fPermanent && PACKAGE_STATE_ABSENT == requested)
        {
            *pExecute = ACTION_STATE_NONE;
            *pRollback = ACTION_STATE_NONE;
            ExitFunction();
        }

        // If the package has a condition, check to see if it is conditioned out.
        if (pPackage->pwzCondition && *pPackage->pwzCondition)
        {
            hr = EvaluateCondition(pPackage->pwzCondition, pProperties, &fCondition);
            ExitOnFailure1(hr, "Failed to evaluate package condition: '%S'", pPackage->pwzCondition);

            if (!fCondition)
            {
                *pExecute = ACTION_STATE_NONE;
                *pRollback = ACTION_STATE_NONE;
                ExitFunction();
            }
        }
#endif

        hr = pEngine->Plan(pPackage, pProperties, requested, pExecute, pRollback);
        ExitOnFailure(hr, "Failed to have engine plan package.");

#if 0
        // If the package is being installed in some way, ensure the prequisites are checked.
        if (ACTION_STATE_NONE != *pExecute && ACTION_STATE_UNINSTALL != *pExecute)
        {
            BOOL fPrereqChecks = TRUE; // assume all of the prequisites will be satisifed.

            for (DWORD i = 0; i < pPackage->cPrereqChecks; ++i)
            {
                SETUP_PREREQUISITE_CHECK* pPrereqCheck = &pSetupPackage->rgPrereqChecks[i];

                // Do prerequisite check.
                fPrereqCheck = TRUE; // a blank prerequisite check should always pass
                if (pPrereqCheck->pwzCondition && *pPrereqCheck->pwzCondition)
                {
                    hr = EvaluateCondition(pPrereqCheck->pwzCondition, pProperties, &fPrereqCheck);
                    ExitOnFailure1(hr, "Failed to evaluate prerequisite check: '%S'", pPrereqCheck->pwzCondition);
                }

                // Send prerequisite check message.
                //hr = SendSuxPrerequisiteCheckMessage(pExecuteEngineInfo, pPackage->dwId, i, fPrereqCheck);
                //ExitOnFailure(hr, "Failed to send prerequisite check message.");

                // If the individal check failed, remember that at least one package has failed prerequisite check.
                if (!fPrereqCheck)
                {
                    fPrereqChecks = FALSE;
                }
            }

            if (!fPrereqChecks)
            {
                LogStringLine(REPORT_STANDARD, "Package prerequisite check failed. The package will not be installed. Id = '%S'", pSetupPackage->wzId);

                // Prerequisite check failed, set status accordingly.
                pSetupPackage->Status = PACKAGE_STATUS_PREREQUISITE_CHECK_FAILED;
                ExitFunction();
            }
        }
#endif

    LExit:
        return hr;
    }


    //
    // CachePackage - loop through all the payloads in a package caching them.
    //
    HRESULT CachePackage(
        __in IBurnPackage* pPackage
        )
    {
        HRESULT hr = S_OK;
        IBurnPayload* pPayload = NULL;
        DWORD cPayloads = pPackage->PayloadCount();
        DWORD64 dw64PackageCacheSize = 0;

        // TODO: calculate the count and total size of all payloads in the package that need caching.

        int nResult = m_pView->OnCachePackageBegin(pPackage, cPayloads, dw64PackageCacheSize);
        hr = HRESULT_FROM_VIEW(nResult);
        ExitOnRootFailure(hr, "UX aborted cache package begin.");

        for (DWORD dwIndex = 1; dwIndex <= cPayloads; ++dwIndex)
        {
            hr = pPackage->GetPayload(dwIndex, &pPayload);
            ExitOnFailure1(hr, "Failed to get payload %d from package.", dwIndex);

            hr = CachePayload(pPackage, pPayload);
            ExitOnFailure1(hr, "Failed to cache payload %d.", dwIndex);

            ReleaseNullObject(pPayload);
        }

    LExit:
        ReleaseObject(pPayload);

        m_pView->OnCachePackageComplete(pPackage, hr);
        return hr;
    }


    HRESULT CacheUXPayloads(
        __out_z LPWSTR* psczUXDLL
        )
    {
        HRESULT hr = S_OK;
        IBurnPayload* pUXPayload = NULL;
        DWORD cPayloads = m_pModel->UXPayloadCount();

        for (DWORD dwIndex = 1; dwIndex <= cPayloads; ++dwIndex)
        {
            m_pModel->GetUXPayload(dwIndex, &pUXPayload);
            ExitOnFailure1(hr, "Failed to get UX payload %d.", dwIndex);

            hr = CachePayload(NULL, pUXPayload);
            ExitOnFailure(hr, "Failed to cache payload.");

            // The UX DLL is always #1, so get its path.
            if (1 == dwIndex)
            {
                hr = pUXPayload->GetCompletedPath(psczUXDLL);
                ExitOnFailure(hr, "Failed to get payload completed path for UX DLL.");
            }

            ReleaseNullObject(pUXPayload);
        }

    LExit:
        ReleaseObject(pUXPayload);

        return hr;
    }


    HRESULT CachePayload(
        __in_opt IBurnPackage* pPackage,
        __in IBurnPayload* pPayload
        )
    {
        HRESULT hr = S_OK;
        LPWSTR sczSource = NULL;
        SOURCE_TYPE sourceType = SOURCE_TYPE_NONE;
        IBurnCacheEngine* pEngine = NULL;

        hr = pPayload->GetSource(&sourceType, &sczSource);
        ExitOnFailure(hr, "Failed to get payload source type.");

        int nResult = m_pView->OnCachePackagePayloadBegin(pPackage, pPayload, sourceType, sczSource);
        hr = HRESULT_FROM_VIEW(nResult);
        ExitOnRootFailure(hr, "UX aborted package cache payload begin.");

        switch (sourceType)
        {
        case SOURCE_TYPE_EMBEDDED: __fallthrough;
        case SOURCE_TYPE_EXTERNAL:
            m_pExtractEngine->AddRef();
            pEngine = m_pExtractEngine;
            break;

        case SOURCE_TYPE_DOWNLOAD:
            m_pDownloadEngine->AddRef();
            pEngine = m_pDownloadEngine;
            break;

        default:
            hr = E_INVALIDARG;
            ExitOnRootFailure(hr, "Invalid payload source type.");
        }

        // We have a payload and the right engine for it, so cache away
        hr = pEngine->CachePayload(pPackage, pPayload, m_pView);
        ExitOnFailure(hr, "Failed to cache payload.");

        hr = pEngine->Complete(pPackage, pPayload, m_pView);
        ExitOnFailure(hr, "Failed to complete payload.");

    LExit:
        m_pView->OnCachePackagePayloadComplete(pPackage, pPayload, hr);
        ReleaseObject(pEngine);
        ReleaseStr(sczSource);

        return hr;
    }


    //
    // ExecutePackage - execute or rollback a package.
    //
    HRESULT ExecutePackage(
        __in IBurnPackage* pPackage,
        __in BOOL fExecute,
        __in BOOL fElevated,
        __out BOOL* pfRestart,
        __out_opt BOOL* pfRollback
        )
    {
        HRESULT hr = S_OK;
        IBurnExecuteEngine* pEngine = NULL;

        // Tell the view we're starting a new package and if we're executing
        // (opposed to rollingback) then check to see if they canceled.
        int nResult = m_pView->OnExecutePackageBegin(pPackage, fExecute);
        if (fExecute)
        {
            hr = HRESULT_FROM_VIEW(nResult);
            ExitOnRootFailure(hr, "UX aborted package execute begin.");
        }

        // If this is a per-machine package but we are not the elevated process, request an elevated package
        // install.
        if (pPackage->IsPerMachine() && !fElevated)
        {
            if (fExecute)
            {
                hr = ElevationParentPackageExecute(m_hElevatedProcess, m_hElevatedPipe, m_pModel, m_pView, pPackage);
                ExitOnFailure(hr, "Failed to execute elevated package.");
            }
            else
            {
                hr = ElevationParentPackageRollback(m_hElevatedProcess, m_hElevatedPipe, m_pModel, m_pView, pPackage);
                ExitOnFailure(hr, "Failed to rollback elevated package.");
            }
        }
        else
        {
            hr = GetExecuteEngineForPackage(pPackage, &pEngine);
            ExitOnFailure(hr, "Failed to get execute engine to prepare.");

            if (fExecute)
            {
                hr = pEngine->Execute(pPackage, m_pView);
                ExitOnFailure(hr, "Failed to execute package.");
            }
            else
            {
                hr = pEngine->Rollback(pPackage, m_pView);
                ExitOnFailure(hr, "Failed to rollback package.");
            }
        }

    LExit:
        // TODO: handle "force reboot" which is returned as "ERROR_INSTALL_SUSPEND".
        if (HRESULT_FROM_WIN32(ERROR_SUCCESS_REBOOT_REQUIRED) == hr || HRESULT_FROM_WIN32(ERROR_SUCCESS_RESTART_REQUIRED) == hr || HRESULT_FROM_WIN32(ERROR_INSTALL_SUSPEND) == hr)
        {
            *pfRestart = TRUE;
            hr = S_OK;
        }
        else if (FAILED(hr))
        {
            if (pfRollback)
            {
                *pfRollback = TRUE;
            }
        }

        m_pView->OnExecutePackageComplete(pPackage, hr);
        ReleaseObject(pEngine);
        return hr;
    }


    //
    // GetExecuteEngineForPackage
    //
    HRESULT GetExecuteEngineForPackage(
        __in IBurnPackage* pPackage,
        __out IBurnExecuteEngine** ppEngine
        )
    {
        HRESULT hr = S_OK;
        IBurnExePackage* pExePackage = NULL;
        IBurnMsiPackage* pMsiPackage = NULL;

        hr = pPackage->QueryInterface(__uuidof(IBurnExePackage), reinterpret_cast<void**>(&pExePackage));
        if (SUCCEEDED(hr))
        {
            hr = m_pExeEngine->QueryInterface(__uuidof(IBurnExecuteEngine), reinterpret_cast<void**>(ppEngine));
        }
        else
        {
            hr = pPackage->QueryInterface(__uuidof(IBurnMsiPackage), reinterpret_cast<void**>(&pMsiPackage));
            if (SUCCEEDED(hr))
            {
                hr = m_pMsiEngine->QueryInterface(__uuidof(IBurnExecuteEngine), reinterpret_cast<void**>(ppEngine));
            }
        }
        ExitOnRootFailure(hr,"Could not find engine for package type.");

    LExit:
        ReleaseObject(pMsiPackage);
        ReleaseObject(pExePackage);
        return hr;
    }


    //
    // UIThread - creates window and processes messages until window is closed.
    //
    static DWORD WINAPI UIThread(
        __in LPVOID pvContext
        )
    {
        HRESULT hr = S_OK;
        CBurnController* pThis = static_cast<CBurnController*>(pvContext);
        IBurnView* pView = NULL;
        BOOL fInitializedView = FALSE;

        // Initialize the view.
        pThis->m_pView->AddRef();
        pView = pThis->m_pView;

        hr = pView->Initialize(pThis->m_pModel, pThis);
        ExitOnFailure(hr, "Failed to initialize view.");

        fInitializedView = TRUE;

        if (!::PostThreadMessageW(pThis->m_dwThreadId, WM_BURN_VIEW_INITIALIZED, 0, 0))
        {
            ExitWithLastError(hr, "Failed to notify controller that view loaded.");
        }

        // Let the view run.
        hr = pView->Run();
        ExitOnFailure(hr, "Failed while running view.");

    LExit:
        if (fInitializedView)
        {
            pView->Uninitialize();
        }
        ReleaseObject(pView);

        return hr;
    }


public:
    //
    // Constructor - intitialize member variables.
    //
    CBurnController(
        __in IBurnModel* pModel,
        __in IBurnView* pView,
        __in IBurnExecuteEngine* pExeEngine,
        __in IBurnExecuteEngine* pMsiEngine,
        __in IBurnCacheEngine* pExtractEngine,
        __in IBurnCacheEngine* pDownloadEngine,
        __in IInstallationAbstractionLayer* pInstallationAbstractionLayer
        )
    {
        m_dwThreadId = 0;
        m_LastRequested = REQUEST_STATE_NONE;
        m_fRestartRequested = FALSE;
        m_fRestartAllowed = FALSE;

        m_hUXModule = NULL;
        m_hElevatedProcess = NULL;
        m_hElevatedPipe = INVALID_HANDLE_VALUE;

        pModel->AddRef();
        m_pModel = pModel;

        pView->AddRef();
        m_pView = pView;

        pExeEngine->AddRef();
        m_pExeEngine = pExeEngine;

        pMsiEngine->AddRef();
        m_pMsiEngine = pMsiEngine;

        pExtractEngine->AddRef();
        m_pExtractEngine = pExtractEngine;

        pDownloadEngine->AddRef();
        m_pDownloadEngine = pDownloadEngine;

        pInstallationAbstractionLayer->AddRef();
        m_pInstallationAbstractionLayer = pInstallationAbstractionLayer;
    }

    //
    // Destructor - release member variables.
    //
    ~CBurnController()
    {
        ReleaseObject(m_pInstallationAbstractionLayer);
        ReleaseObject(m_pDownloadEngine);
        ReleaseObject(m_pExtractEngine);
        ReleaseObject(m_pMsiEngine);
        ReleaseObject(m_pExeEngine);

        ReleaseFileHandle(m_hElevatedPipe);
        if (m_hElevatedProcess)
        {
            ::CloseHandle(m_hElevatedProcess);
        }

        if (m_hUXModule)
        {
            ::FreeLibrary(m_hUXModule);
        }

        ReleaseObject(m_pView);
        ReleaseObject(m_pModel);
    }

private:
    DWORD m_dwThreadId;
    REQUEST_STATE m_LastRequested;
    BOOL m_fRestartRequested;
    BOOL m_fRestartAllowed;

    HMODULE m_hUXModule;
    HANDLE m_hElevatedProcess;
    HANDLE m_hElevatedPipe;

    IBurnModel* m_pModel;
    IBurnView* m_pView;
    IBurnCacheEngine* m_pExtractEngine;
    IBurnCacheEngine* m_pDownloadEngine;
    IBurnExecuteEngine* m_pExeEngine;
    IBurnExecuteEngine* m_pMsiEngine;

    IInstallationAbstractionLayer* m_pInstallationAbstractionLayer;
};


//
// BurnCreateController - creates a new IBurnController object for the provided model, view, and engines.
//
HRESULT BurnCreateController(
    __in IBurnModel* pModel,
    __in IBurnView* pView,
    __in IBurnExecuteEngine* pExeEngine,
    __in IBurnExecuteEngine* pMsiEngine,
    __in IBurnCacheEngine* pExtractEngine,
    __in IBurnCacheEngine* pDownloadEngine,
    __in IInstallationAbstractionLayer* pInstallationAbstractionLayer,
    __out IBurnController **ppController
    )
{
    HRESULT hr = S_OK;

    CBurnController* pController = new CBurnController(pModel, pView, pExeEngine, pMsiEngine, pExtractEngine, pDownloadEngine, pInstallationAbstractionLayer);
    ExitOnNull(pController, hr, E_OUTOFMEMORY, "Failed to create new model object.");

    *ppController = pController;
    pController = NULL;

LExit:
    ReleaseObject(pController);
    return hr;
}
