//-------------------------------------------------------------------------------------------------
// <copyright file="BurnExePackage.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"


class CBurnExePackage : public CBurnPackageBase<IBurnExePackage, IBurnPackageTestable>
{
public: // IBurnExePackage
    //
    // GetProductCode - gets the product code.
    //
    STDMETHODIMP GetProductCode(
        __out LPWSTR* psczProductCode
        )
    {
        HRESULT hr = StrAllocString(psczProductCode, m_sczProductCode, 0);
        ExitOnFailure(hr, "Failed to copy product code.");

    LExit:
        return hr;
    }


    //
    // IsRepairAllowed - specifies whether executable understands has special logic for repair operations.
    //
    STDMETHODIMP_(BOOL) IsRepairAllowed()
    {
        return FALSE; // TODO: return the correct value.
    }


    //
    // GetDetectCondition - gets the condition to detect the product. Returns
    //                     E_NOTIMPL if the package does not support detection.
    //
    STDMETHODIMP GetDetectCondition(
        __out LPWSTR* psczDetectCondition
        )
    {
        HRESULT hr = E_NOTIMPL;

        if (m_sczDetectCondition)
        {
            hr = StrAllocString(psczDetectCondition, m_sczDetectCondition, 0);
            ExitOnFailure(hr, "Failed to copy detect condition.");
        }

    LExit:
        return hr;
    }


    STDMETHODIMP GetExitInfo(
        __in DWORD dwIndex,
        __out BURN_EXE_EXIT_CODE_TYPE* pType,
        __out DWORD* pdwExitCode,
        __out BOOL* pfWildcard
        )
    {
        if (m_cExitInfos <= dwIndex)
        {
            return E_INVALIDARG;
        }

        *pType = m_prgExitInfo[dwIndex].type;
        *pdwExitCode = m_prgExitInfo[dwIndex].dwCode;
        *pfWildcard = m_prgExitInfo[dwIndex].fWildcard;

        return S_OK;
    }


    STDMETHODIMP_(DWORD) ExitInfoCount()
    {
        return m_cExitInfos;
    }


    //
    // GetInstallSwitch - gets the switch to install the product.
    //
    STDMETHODIMP GetInstallSwitch(
        __out LPWSTR* psczInstallSwitch
        )
    {
        HRESULT hr = StrAllocString(psczInstallSwitch, m_sczInstallSwitch, 0);
        ExitOnFailure(hr, "Failed to copy install switch.");

    LExit:
        return hr;
    }


    //
    // GetProgressSwitch - gets the optional swith to turn on progress. Returns
    //                     E_NOTIMPL if the package does not support progress.
    //
    STDMETHODIMP GetProgressSwitch(
        __out LPWSTR* psczProgressSwitch
        )
    {
        HRESULT hr = E_NOTIMPL;

        if (m_sczProgressSwitch)
        {
            hr = StrAllocString(psczProgressSwitch, m_sczProgressSwitch, 0);
            ExitOnFailure(hr, "Failed to copy progress switch.");
        }

    LExit:
        return hr;
    }


    //
    // GetUninstallSwitch - gets the switch to uninstall the product.
    //
    STDMETHODIMP GetUninstallSwitch(
        __out LPWSTR* psczUninstallSwitch
        )
    {
        HRESULT hr = StrAllocString(psczUninstallSwitch, m_sczUninstallSwitch, 0);
        ExitOnFailure(hr, "Failed to copy uninstall switch.");

    LExit:
        return hr;
    }


public: // internals
    //
    // SetPayloadCacheRoot - sets the cache root path for each payload.
    //
    STDMETHODIMP SetPayloadCacheRoot()
    {
        HRESULT hr = S_OK;
        LPWSTR sczLocalAppData = NULL;
        LPWSTR sczCacheRoot = NULL;
        WCHAR wzVersion[36] = {0};

        hr = PathGetKnownFolder(m_fPerMachine ? CSIDL_COMMON_APPDATA : CSIDL_LOCAL_APPDATA, &sczLocalAppData);
        ExitOnFailure1(hr, "Failed to find local %s appdata directory.", m_fPerMachine ? "per-machine" : "per-user");

        hr = ::StringCchPrintfW(wzVersion, countof(wzVersion), L"%I64d.%I64d.%I64d.%I64d",
            (m_qwVersion & 0xffff000000000000) >> 48, 
            (m_qwVersion & 0x0000ffff00000000) >> 32, 
            (m_qwVersion & 0x00000000ffff0000) >> 16, 
             m_qwVersion & 0x000000000000ffff);
        ExitOnFailure(hr, "Failed to format version string.");

        hr = StrAllocFormatted(&sczCacheRoot, L"%lsApplications\\Cache\\%lsv%ls", sczLocalAppData, m_sczProductCode, wzVersion);
        ExitOnFailure(hr, "Failed to format cache path.");

        // now inform the payloads
        for (DWORD dwIndex = 0; dwIndex < m_cPayloads; ++dwIndex)
        {
            m_rgpPayloads[dwIndex]->SetCompletedRoot(sczCacheRoot);
        }

    LExit:
        ReleaseStr(sczCacheRoot);

        return hr;
    }




public:
    //
    // Constructor - intitialize member variables.
    //
    CBurnExePackage(
        __in_z LPCWSTR wzId,
        __in_z LPCWSTR wzProductCode,
        __in DWORD64 qwVersion,
        __in BOOL fPerMachine,
        __in_z_opt LPCWSTR wzDetectCondition,
        __in_z LPCWSTR wzInstallArgs,
        __in_z_opt LPCWSTR wzRepairArgs,
        __in_z_opt LPCWSTR wzUninstallArgs,
        __in DWORD cExitInfos,
        __in_ecount(cExitInfos) BURN_EXE_PACKAGE_EXIT_INFO* rgExitInfo,
        __in DWORD cPayloads,
        __in_ecount_opt(cPayloads) IBurnPayload** rgpPayloads,
        __in_opt IBurnProperties* pProperties,
        __inout HRESULT *pHR
        ) : CBurnPackageBase(wzId, qwVersion, fPerMachine, cPayloads, rgpPayloads, pHR)
    {
        m_sczProductCode = NULL;
        m_sczDetectCondition = NULL;
        m_sczInstallSwitch = NULL;
        m_sczRepairSwitch = NULL;
        m_sczProgressSwitch = NULL;
        m_sczUninstallSwitch = NULL;
        m_prgExitInfo = NULL;

        HRESULT hr = *pHR;
        ExitOnFailure(hr, "Failed to initialize base package.");

        if (pProperties)
        {
            m_pProperties = pProperties;
            m_pProperties->AddRef();
        }

        hr = StrAllocString(&m_sczProductCode, wzProductCode, 0);
        ExitOnFailure(hr, "Failed to copy product code.");

        if (wzDetectCondition)
        {
            hr = StrAllocString(&m_sczDetectCondition, wzDetectCondition, 0);
            ExitOnFailure(hr, "Failed to copy detect condition.");
        }

        hr = StrAllocString(&m_sczInstallSwitch, wzInstallArgs, 0);
        ExitOnFailure(hr, "Failed to copy install arguments.");

        if (wzRepairArgs)
        {
            hr = StrAllocString(&m_sczRepairSwitch, wzRepairArgs, 0);
            ExitOnFailure(hr, "Failed to copy repair arguments.");
        }

        if (wzUninstallArgs)
        {
            hr = StrAllocString(&m_sczUninstallSwitch, wzUninstallArgs, 0);
            ExitOnFailure(hr, "Failed to copy uninstall arguments.");
        }

        m_cExitInfos = cExitInfos;
        m_prgExitInfo = rgExitInfo;

    LExit:
        *pHR = hr;
    }

    //
    // Destructor - release member variables.
    //
    ~CBurnExePackage()
    {
        ReleaseMem(m_prgExitInfo);
        ReleaseStr(m_sczUninstallSwitch);
        ReleaseStr(m_sczProgressSwitch);
        ReleaseStr(m_sczRepairSwitch);
        ReleaseStr(m_sczInstallSwitch);
        ReleaseStr(m_sczDetectCondition);
        ReleaseStr(m_sczProductCode);
    }

private:
    LPWSTR m_sczProductCode;

    LPWSTR m_sczDetectCondition;
    LPWSTR m_sczInstallSwitch;
    LPWSTR m_sczRepairSwitch;
    LPWSTR m_sczProgressSwitch;
    LPWSTR m_sczUninstallSwitch;

    BURN_EXE_PACKAGE_EXIT_INFO* m_prgExitInfo;
    DWORD m_cExitInfos;
};


//
// BurnCreateExePackage - creates a new IBurnExePackage object.
//
HRESULT BurnCreateExePackage(
    __in_z LPCWSTR wzId,
    __in_z LPCWSTR wzProductCode,
    __in DWORD64 qwVersion,
    __in BOOL fPerMachine,
    __in_z_opt LPCWSTR wzDetectCondition,
    __in_z LPCWSTR wzInstallArgs,
    __in_z_opt LPCWSTR wzRepairArgs,
    __in_z_opt LPCWSTR wzUninstallArgs,
    __in DWORD cExitInfos,
    __in_ecount(cExitInfos) BURN_EXE_PACKAGE_EXIT_INFO* rgExitInfo,
    __in DWORD cPayloads,
    __in_ecount_opt(cPayloads) IBurnPayload** rgpPayloads,
    __in_opt IBurnProperties* pProperties,
    __out IBurnExePackage** ppExePackage
    )
{
    HRESULT hr = S_OK;

    CBurnExePackage* pExePackage = new CBurnExePackage(wzId, wzProductCode, qwVersion, fPerMachine, wzDetectCondition, wzInstallArgs, wzRepairArgs, wzUninstallArgs, cExitInfos, rgExitInfo, cPayloads, rgpPayloads, pProperties, &hr);
    ExitOnNull(pExePackage, hr, E_OUTOFMEMORY, "Failed to create new EXE package.");
    ExitOnFailure(hr, "Failed to initialize EXE package.");

    hr = pExePackage->SetPayloadCacheRoot();
    ExitOnFailure(hr, "Failed to set payload cache roots.");

    *ppExePackage = pExePackage;
    pExePackage = NULL;

LExit:
    ReleaseObject(pExePackage);
    return hr;
}
