//-------------------------------------------------------------------------------------------------
// <copyright file="BurnModel.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"

class CBurnModel : public CUnknownImpl2<IBurnModel, IBurnModelTestable>
{
public: // IBurnModel
    //
    // GetCommand - gets the commands Burn should be executing.
    //
    STDMETHODIMP GetCommand(
        __out BURN_COMMAND* pCommand
        )
    {
        memcpy_s(pCommand, sizeof(BURN_COMMAND), &m_command, sizeof(m_command));
        return S_OK;
    }


    //
    // GetCommandLineParameters - gets the command line parameters passed in with the exe
    //
    STDMETHODIMP GetCommandLineParameters(
        __out LPWSTR* psczCommandLine
        )
    {
        HRESULT hr = S_OK;

        if (m_sczCommandLine)
        {
            hr = StrAllocString(psczCommandLine, m_sczCommandLine, 0);
            ExitOnFailure(hr, "Error getting command line.");
        }
        else
        {
            ExitFunction1(hr = S_FALSE);
        }

LExit:
        return hr;
    }


    //
    // GetElevatedInfo - gets the pipe name and token from the elevated process
    //                   verify the identity.
    //
    STDMETHODIMP GetElevatedInfo(
        __out_z LPWSTR* psczPipeName,
        __out_z LPWSTR* psczToken
        )
    {
        HRESULT hr = S_OK;
        if (BURN_MODE_ELEVATED != m_mode)
        {
            ExitFunction1(hr = E_UNEXPECTED);
        }

        AssertSz(m_sczElevatedPipeName, "Burn model is in elevated mode, there should have been a pipe name provided.");
        hr = StrAllocString(psczPipeName, m_sczElevatedPipeName, 0);
        ExitOnFailure(hr, "Failed to copy elevated pipe name");

        AssertSz(m_sczElevatedToken, "Burn model is in elevated mode, there should have been an elevated token provided.");
        hr = StrAllocString(psczToken, m_sczElevatedToken, 0);
        ExitOnFailure(hr, "Failed to copy elevated token");

    LExit:
        return hr;
    }


    //
    // GetLogPath - gets the path to the log file. May return NULL.
    //
    STDMETHODIMP GetLogPath(
        __out_z LPWSTR* psczLogPath,
        __out_opt BURN_LOGGING_LEVEL* pLogLevel
        )
    {
        HRESULT hr = S_OK;

        if (!m_sczLogFile)
        {
            // create a default log if one hasn't already been initialized
            hr = InitializeLog(BURN_LOGGING_LEVEL_NORMAL, BURN_LOGGING_MODE_UNIQUE, NULL);
            ExitOnFailure(hr, "Failed to initialize default log");
        }

        hr = StrAllocString(psczLogPath, m_sczLogFile, 0);
        ExitOnFailure(hr, "Failed to copy log path.");

        if (pLogLevel)
        {
            *pLogLevel = m_logLevel;
        }

    LExit:
        return hr;
    }


    //
    // Mode - returns whether we are elevated, embedded or just running normal.
    //
    STDMETHODIMP_(BURN_MODE) Mode()
    {
        return m_mode;
    }

    STDMETHODIMP GetPackage(
        __in DWORD dwIndex,
        __out IBurnPackage** ppPackage
        )
    {
        HRESULT hr = S_OK;

        if (0 == dwIndex || m_cPackages < dwIndex)
        {
            hr = E_INVALIDARG;
            ExitOnRootFailure(hr, "Invalid package index.");
        }

        m_rgpPackages[dwIndex - 1]->AddRef();
        *ppPackage = m_rgpPackages[dwIndex - 1];

    LExit:
        return hr;
    }


    STDMETHODIMP GetPackageIndex(
        __in LPCWSTR wzId,
        __out DWORD* pdwIndex
        )
    {
        HRESULT hr = S_OK;
        DWORD i;

        for (i = 0; i < m_cPackages; ++i)
        {
            if (CSTR_EQUAL == m_rgpPackages[i]->CompareId(wzId))
            {
                *pdwIndex = i + 1; // remember, package indices are "1" based.
                break;
            }
        }

        if (i == m_cPackages)
        {
            hr = E_INVALIDARG;
            ExitOnRootFailure1(hr, "Failed to find package with id: %ls", wzId);
        }

    LExit:
        return hr;
    }


    STDMETHODIMP GetProperties(
        __out IBurnProperties** ppProperties
        )
    {
        HRESULT hr = S_OK;

        m_pGlobalProperties->AddRef();
        *ppProperties = m_pGlobalProperties;

        return hr;
    }


    STDMETHODIMP_(DWORD) PackageCount()
    {
        return m_cPackages;
    }

    STDMETHODIMP_(DWORD) UXPayloadCount()
    {
        return m_cUXPayloads;
    }

    STDMETHODIMP GetUXPayload(
        __in DWORD dwIndex,
        __out IBurnPayload** ppPayload
        )
    {
        HRESULT hr = S_OK;

        if (0 == dwIndex || m_cUXPayloads < dwIndex)
        {
            hr = E_INVALIDARG;
            ExitOnRootFailure(hr, "Invalid UX payload index.");
        }

        m_rgpUXPayloads[dwIndex - 1]->AddRef();
        *ppPayload = m_rgpUXPayloads[dwIndex - 1];

    LExit:
        return hr;
    }


    //
    // Load - loads the embedded or external manifests.
    //
    STDMETHODIMP Load(
        __in LPCWSTR sczBundleName
        )
    {
        HRESULT hr = S_OK;

        LPWSTR sczFileName = NULL;
        DWORD cPackages = 0;
        IBurnPackage** rgpPackages = NULL;
        DWORD cUXPayloads = 0;
        IBurnPayload** rgpUXPayloads = NULL;
        BURN_SEARCHES searches = { 0 };

        ManifestInitialize();

        if (m_sczManifestFile)
        {
            // TODO
            // ManifestParseFromFile(...)
        }
        else
        {
            if (!sczBundleName)
            {
                hr = PathForCurrentProcess(&sczFileName, NULL);
                ExitOnFailure(hr, "Failed to get path for current process.");
            }

            hr = ManifestParseFromEmbedded(sczBundleName ? sczBundleName : sczFileName, this, &rgpPackages, &cPackages, &rgpUXPayloads, &cUXPayloads, &searches);
            ExitOnFailure(hr, "Failed to load embedded manifest");
        }

        hr = SearchesSort(&searches);
        ExitOnFailure(hr, "Failed to sort searches.");

        m_cPackages = cPackages;
        m_rgpPackages = rgpPackages;
        rgpPackages = NULL;

        m_cUXPayloads = cUXPayloads;
        m_rgpUXPayloads = rgpUXPayloads;
        rgpUXPayloads = NULL;

        memcpy_s(&this->m_Searches, sizeof(BURN_SEARCHES), &searches, sizeof(BURN_SEARCHES));
        memset(&searches, 0, sizeof(BURN_SEARCHES));

    LExit:
        ManifestUninitialize();
        ReleaseStr(sczFileName);
        ReleaseObjectArray(rgpPackages, cPackages);
        ReleaseObjectArray(rgpUXPayloads, cUXPayloads);
        SearchesUninitialize(&searches);

        return hr;
    }


    //
    // InitializeLog - create the chain-wide log file based on command-line switch, manifest, or defaults.
    //
    STDMETHODIMP InitializeLog(
        __in BURN_LOGGING_LEVEL logLevel,
        __in BURN_LOGGING_MODE logMode,
        __in_opt LPCWSTR wzLogPath
        )
    {
        HRESULT hr = S_OK;
        LPWSTR sczNewLogPath = NULL;
        LPWSTR sczLogDirectory = NULL;
        HANDLE hLogFile = INVALID_HANDLE_VALUE;

        // the chain log set by the /log command-line switch overrides the defaults set in the manifest
        if (m_sczLogFile)
        {
            Trace(REPORT_STANDARD, "Model log has already been initialized");
            hr = S_FALSE;
            ExitFunction();
        }

        if (wzLogPath)
        {
            // resolve %environment_variables%
            hr = FileResolvePath(wzLogPath, &sczNewLogPath);
            ExitOnFailure(hr, "Failed to resolve incoming log path.");
        }
        else
        {
            // the default chain log is <nameOfBundle>.exe.log in the current directory
            hr = PathForCurrentProcess(&sczNewLogPath, NULL);
            ExitOnFailure(hr, "Failed to get current process path for log.");

            hr = StrAllocString(&sczNewLogPath, FileFromPath(sczNewLogPath), 0);
            ExitOnFailure(hr, "Failed to copy current process file name for log.");

            hr = FileChangeExtension(sczNewLogPath, L".log", &sczNewLogPath);
            ExitOnFailure(hr, "Failed to rename current process log.");
        }

        // create the directory if one was provided
        hr = PathGetDirectory(sczNewLogPath, &sczLogDirectory);
        ExitOnFailure1(hr, "Failed to get directory from log path: %ls", wzLogPath);

        if (sczLogDirectory)
        {
            hr = DirEnsureExists(sczLogDirectory, NULL);
            ExitOnFailure(hr, "Failed to create log directory.");
        }

        // handle rename-mode collision-avoidance using PathCreateTempFile's templating
        // (the PathCreateTempFile template is <logBaseName>.%04x.<logExtension>);
        // otherwise we're appending/overwriting and LogInitialize handles that for us
        if (BURN_LOGGING_MODE_UNIQUE == logMode)
        {
            hr = FileAddSuffixToBaseName(FileFromPath(sczNewLogPath), L".%04x", &sczNewLogPath);
            ExitOnFailure(hr, "Failed to allocate renamed log template.");

            hr = PathCreateTempFile(sczLogDirectory, sczNewLogPath, SHRT_MAX, FILE_ATTRIBUTE_NORMAL, &sczNewLogPath, NULL);
            ExitOnFailure(hr, "Failed to create log file.");
        }

        hr = LogInitialize(::GetModuleHandleW(NULL), sczNewLogPath, NULL, BURN_LOGGING_MODE_APPEND == logMode, TRUE);
        ExitOnFailure(hr, "Failed to initialize logging system.");

        m_logLevel = logLevel;
        m_sczLogFile = sczNewLogPath;
        sczNewLogPath = NULL;

    LExit:
        ReleaseStr(sczLogDirectory);
        ReleaseStr(sczNewLogPath);
        ReleaseFile(hLogFile);

        return hr;
    }


public: // IBurnModelTestable
    STDMETHODIMP AddPackageForTestingPurposes(
        __in IBurnPackage* pPackage
        )
    {
        HRESULT hr = S_OK;

        ++m_cPackages;
        m_rgpPackages = static_cast<IBurnPackage**>(MemAlloc(sizeof(IBurnPackage*) * m_cPackages, TRUE));
        ExitOnNull(m_rgpPackages, hr, E_OUTOFMEMORY, "Failed to allocate packages.");

        pPackage->AddRef();
        m_rgpPackages[m_cPackages - 1] = pPackage;

    LExit:
        return hr;
    }

public: // internals
    //
    // Initialize - process the provided command line arguments.
    //
    HRESULT Initialize(
        __in_z LPCWSTR wzCommandLine
        )
    {
        HRESULT hr = S_OK;
        LPWSTR sczName = NULL;

        int argc = 0;
        LPWSTR* argv = ::CommandLineToArgvW(wzCommandLine, &argc);
        ExitOnNullWithLastError(argv, hr, "Failed to get command line.");

        hr = BurnCreateProperties(&m_pGlobalProperties);
        ExitOnFailure(hr, "Failed to create empty properties object.");

        for (int i = 0; i < argc; ++i)
        {
            if (argv[i][0] == L'-' || argv[i][0] == L'/')
            {
                if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, NORM_IGNORECASE, &argv[i][1], -1, L"l", -1) ||
                    CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, NORM_IGNORECASE, &argv[i][1], -1, L"log", -1))
                {
                    if (i + 1 >= argc)
                    {
                        ExitOnRootFailure(hr = E_INVALIDARG, "Must specify a path for log.");
                    }

                    ++i;

                    hr = StrAllocString(&m_sczLogFile, argv[i], 0);
                    ExitOnFailure(hr, "Failed to copy log file path.");
                }
                else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, NORM_IGNORECASE, &argv[i][1], -1, L"help", -1))
                {
                    m_command.action = BURN_ACTION_HELP;
                }
                else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, NORM_IGNORECASE, &argv[i][1], -1, L"quiet", -1))
                {
                    m_command.display = BURN_DISPLAY_NONE;

                    if (BURN_RESTART_UNKNOWN == m_command.restart)
                    {
                        m_command.restart = BURN_RESTART_AUTOMATIC;
                    }
                }
                else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, NORM_IGNORECASE, &argv[i][1], -1, L"passive", -1))
                {
                    m_command.display = BURN_DISPLAY_PASSIVE;

                    if (BURN_RESTART_UNKNOWN == m_command.restart)
                    {
                        m_command.restart = BURN_RESTART_AUTOMATIC;
                    }
                }
                else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, NORM_IGNORECASE, &argv[i][1], -1, L"norestart", -1))
                {
                    m_command.restart = BURN_RESTART_NEVER;
                }
                else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, NORM_IGNORECASE, &argv[i][1], -1, L"forcerestart", -1))
                {
                    m_command.restart = BURN_RESTART_ALWAYS;
                }
                else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, NORM_IGNORECASE, &argv[i][1], -1, L"promptrestart", -1))
                {
                    m_command.restart = BURN_RESTART_PROMPT;
                }
                else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, NORM_IGNORECASE, &argv[i][1], -1, L"uninstall", -1))
                {
                    if (BURN_ACTION_HELP != m_command.action)
                    {
                        m_command.action = BURN_ACTION_UNINSTALL;
                    }
                }
                else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, NORM_IGNORECASE, &argv[i][1], -1, L"package", -1) ||
                         CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, NORM_IGNORECASE, &argv[i][1], -1, L"update", -1))
                {
                    if (BURN_ACTION_UNKNOWN == m_command.action)
                    {
                        m_command.action = BURN_ACTION_INSTALL;
                    }
                }
                else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, NORM_IGNORECASE, &argv[i][1], -1, L"burn.elevated", -1))
                {
                    if (i + 2 >= argc)
                    {
                        ExitOnRootFailure(hr = E_INVALIDARG, "Must specify the parent and child elevation tokens.");
                    }

                    m_mode = BURN_MODE_ELEVATED;

                    ++i;

                    hr = StrAllocString(&m_sczElevatedPipeName, argv[i], 0);
                    ExitOnFailure(hr, "Failed to copy elevated pipe name.");

                    ++i;

                    hr = StrAllocString(&m_sczElevatedToken, argv[i], 0);
                    ExitOnFailure(hr, "Failed to copy elevation token.");
                }
                else
                {
                    // Remember command-line switch to pass off to UX.
                    hr = StrAllocConcat(&m_sczCommandLine, &argv[i][0], 0);
                    ExitOnFailure(hr, "Failed to copy command line parameter.");

                    hr = StrAllocConcat(&m_sczCommandLine, L" ", 0);
                    ExitOnFailure(hr, "Failed to copy space.");
                }
            }
            else
            {
                // if it looks like PROPERTY=VALUE, store it as a global property.
                LPCWSTR wzEquals = wcschr(argv[i], '=');
                if (wzEquals)
                {
                    hr = StrAllocString(&sczName, argv[i], wzEquals - argv[i]);
                    ExitOnFailure(hr, "Failed to extract command-line property name.");

                    hr = m_pGlobalProperties->SetString(sczName, wzEquals + 1);
                    ExitOnFailure(hr, "Failed to set command-line property value.");

                    ReleaseNullStr(sczName);
                }
                else
                {
                    // Remember command-line switch to pass off to UX.
                    hr = StrAllocConcat(&m_sczCommandLine, argv[i], 0);
                    ExitOnFailure(hr, "Failed to copy command line parameter.");

                    hr = StrAllocConcat(&m_sczCommandLine, L" ", 0);
                    ExitOnFailure(hr, "Failed to copy space.");
                }
            }
        }

        // Set the defaults if nothing was set above.
        if (BURN_ACTION_UNKNOWN == m_command.action)
        {
            m_command.action = BURN_ACTION_INSTALL;
        }

        if (BURN_DISPLAY_UNKNOWN == m_command.display)
        {
            m_command.display = BURN_DISPLAY_FULL;
        }

        if (BURN_RESTART_UNKNOWN == m_command.restart)
        {
            m_command.restart = BURN_RESTART_PROMPT;
        }

        Trace3(REPORT_STANDARD, "CBurnModel::Initialize() - Command action = %d, display = %d, restart = %d", m_command.action, m_command.display, m_command.restart);

    LExit:
        ReleaseStr(sczName);

        if (argv)
        {
            ::LocalFree(argv);
        }

        return hr;
    }

public:
    //
    // Constructor - intitialize member variables.
    //
    CBurnModel()
    {
        m_mode = BURN_MODE_NORMAL;
        m_sczElevatedPipeName = NULL;
        m_sczElevatedToken = NULL;

        memset(&m_command, 0, sizeof(m_command));
        m_sczCommandLine = NULL;
        m_sczManifestFile = NULL;
        m_pGlobalProperties = NULL;
        m_logLevel = BURN_LOGGING_LEVEL_UNKNOWN;
        m_sczLogFile = NULL;

        m_rgpPackages = NULL;
        m_cPackages = 0;
        m_rgpUXPayloads = NULL;
        m_cUXPayloads = 0;

        memset(&this->m_Searches, 0, sizeof(BURN_SEARCHES));
    }

    //
    // Destructor - release member variables.
    //
    ~CBurnModel()
    {
        ReleaseObject(m_pGlobalProperties);
        if (m_rgpPackages)
        {
            for (DWORD i = 0; i < m_cPackages; ++i)
            {
                ReleaseObject(m_rgpPackages[i]);
            }
            MemFree(m_rgpPackages);
        }
        ReleaseObjectArray(m_rgpUXPayloads, m_cUXPayloads);
        LogUninitialize(TRUE);
        ReleaseStr(m_sczElevatedToken);
        ReleaseStr(m_sczElevatedPipeName);
        ReleaseStr(m_sczLogFile);
        ReleaseStr(m_sczManifestFile);
        ReleaseStr(m_sczCommandLine);
        SearchesUninitialize(&this->m_Searches);
    }

private:
    BURN_MODE m_mode;
    LPWSTR m_sczElevatedPipeName;
    LPWSTR m_sczElevatedToken;

    BURN_COMMAND m_command;
    BURN_LOGGING_LEVEL m_logLevel;
    LPWSTR m_sczCommandLine;
    LPWSTR m_sczLogFile;
    LPWSTR m_sczManifestFile;
    IBurnProperties* m_pGlobalProperties;

    IBurnPackage** m_rgpPackages;
    DWORD m_cPackages;

    IBurnPayload** m_rgpUXPayloads;
    DWORD m_cUXPayloads;

    BURN_SEARCHES m_Searches;
};


//
// BurnCreateModel - creates a new IBurnModel object.
//
HRESULT BurnCreateModel(
    __in_z LPCWSTR wzCommandLine,
    __out IBurnModel **ppModel
    )
{
    HRESULT hr = S_OK;

    CBurnModel* pModel = new CBurnModel();
    ExitOnNull(pModel, hr, E_OUTOFMEMORY, "Failed to create new model object.");

    hr = pModel->Initialize(wzCommandLine);
    ExitOnFailure(hr, "Failed to initialize command line object.");

    *ppModel = pModel;
    pModel = NULL;

LExit:
    ReleaseObject(pModel);
    return hr;
}
