//-------------------------------------------------------------------------------------------------
// <copyright file="BurnViewElevated.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"


//
// BurnViewElevated - implements IBurnView for elevated.
//
class CBurnViewElevated : public CUnknownImpl<IBurnView>
{
public: // IBurnView
    //
    // Initialize - initialize the view.
    //
    STDMETHODIMP Initialize(
        __in IBurnModel* pModel,
        __in IBurnController* pController
        )
    {
        HRESULT hr = S_OK;
        HANDLE hPipe = INVALID_HANDLE_VALUE;

        hr = ConnectToParent(pModel, &hPipe);
        ExitOnFailure(hr, "Failed to connect to parent process.");

        pController->AddRef();
        m_pController = pController;

        pModel->AddRef();
        m_pModel = pModel;

        m_hPipe = hPipe;
        hPipe = INVALID_HANDLE_VALUE;

        m_fInitialized = TRUE;

    LExit:
        ReleaseFileHandle(hPipe);
        return hr;
    }


    //
    // Uninitialize - ensure all the contained objects are freed.
    //
    STDMETHODIMP_(void) Uninitialize()
    {
        m_fInitialized = FALSE;

        if (INVALID_HANDLE_VALUE != m_hPipe)
        {
            HRESULT hr = ElevationChildPostMessage(m_hPipe, BURN_ELEVATION_MESSAGE_TYPE_CHILD_UNINITIALIZE, NULL, 0);
            TraceError(hr, "Failed to post elevated message for uninitailize.");
        }

        ReleaseFileHandle(m_hPipe);
        ReleaseNullObject(m_pController);
        ReleaseNullObject(m_pModel);
    }


    //
    // Run - run the view.
    //
    STDMETHODIMP Run()
    {
        HRESULT hr = S_OK;
        BURN_ELEVATION_MESSAGE msg = { };
        DWORD dwAck = 1;
        DWORD cb = 0;

        if (INVALID_HANDLE_VALUE == m_hPipe)
        {
            hr = E_UNEXPECTED;
            ExitOnRootFailure(hr, "Cannot run elevated view until parent pipe is open.");
        }

        if (!::WriteFile(m_hPipe, &dwAck, sizeof(dwAck), &cb, NULL))
        {
            ExitWithLastError(hr, "Failed to inform parent process that we are running.");
        }

        // Pump messages from parent process.
        while (S_OK == (hr = ElevationChildGetMessage(m_hPipe, &msg)))
        {
            // Process the message.
            switch (msg.dwMessage)
            {
            case BURN_ELEVATION_MESSAGE_TYPE_CACHE:
                {
                if (sizeof(DWORD) * 3 != msg.cbData)
                {
                    hr = E_INVALIDARG;
                    ExitOnRootFailure(hr, "Elevated cache message did not provide correct data.");
                }

                DWORD dwPackageId = *static_cast<DWORD*>(msg.pvData);
                ACTION_STATE execute = *reinterpret_cast<ACTION_STATE*>(static_cast<DWORD*>(msg.pvData) + 1);
                ACTION_STATE rollback = *reinterpret_cast<ACTION_STATE*>(static_cast<DWORD*>(msg.pvData) + 2);
                HRESULT hrResult = m_pController->CacheElevatedPackage(dwPackageId, execute, rollback);

                hr = ElevationChildPostMessage(m_hPipe, BURN_ELEVATION_MESSAGE_TYPE_COMPLETE, &hrResult, sizeof(hrResult));
                ExitOnFailure(hr, "Failed to post complete cached message to parent process.");
                }
                break;

            case BURN_ELEVATION_MESSAGE_TYPE_EXECUTE: __fallthrough;
            case BURN_ELEVATION_MESSAGE_TYPE_ROLLBACK:
                {
                if (sizeof(DWORD) != msg.cbData)
                {
                    hr = E_INVALIDARG;
                    ExitOnRootFailure(hr, "Elevated execute or rollback message did not provide correct data.");
                }

                DWORD dwPackageId = *static_cast<DWORD*>(msg.pvData);
                HRESULT hrResult = m_pController->ExecuteElevatedPackage(dwPackageId, BURN_ELEVATION_MESSAGE_TYPE_EXECUTE == msg.dwMessage);

                hr = ElevationChildPostMessage(m_hPipe, BURN_ELEVATION_MESSAGE_TYPE_COMPLETE, &hrResult, sizeof(hrResult));
                ExitOnFailure(hr, "Failed to post complete execute or rollback message to parent process.");
                }
                break;

            default:
                hr = E_INVALIDARG;
                ExitOnRootFailure1(hr, "Unexpected elevated message sent to child process, msg: %u", msg.dwMessage);
            }

            ElevationMessageUninitialize(&msg);
        }

        if (S_FALSE == hr)
        {
            hr = S_OK;
        }

    LExit:
        ElevationMessageUninitialize(&msg);
        return hr;
    }


    STDMETHODIMP_(int) OnPrepareBegin(
        __in DWORD cPackages
        )
    {
        AssertSz(FALSE, "Elevated view should never be called here.");
        return IDERROR;
    }


    STDMETHODIMP_(int) OnPreparePackageBegin(
        __in IBurnPackage* pPackage
        )
    {
        AssertSz(FALSE, "Elevated view should never be called here.");
        return IDERROR;
    }


    STDMETHODIMP_(void) OnPreparePackageComplete(
        __in IBurnPackage* pPackage,
        __in HRESULT hrStatus,
        __in PACKAGE_STATE state,
        __in REQUEST_STATE requested,
        __in ACTION_STATE execute,
        __in ACTION_STATE rollback
        )
    {
        AssertSz(FALSE, "Elevated view should never be called here.");
        return;
    }


    STDMETHODIMP_(void) OnPrepareComplete(
        __in HRESULT hrStatus
        )
    {
        AssertSz(FALSE, "Elevated view should never be called here.");
        return;
    }


    STDMETHODIMP_(int) OnCacheBegin(
        __in DWORD cCachingPackages,
        __in DWORD cTotalPayloads,
        __in DWORD64 dw64TotalCacheSize
        )
    {
        AssertSz(FALSE, "Elevated view should never be called here.");
        return IDERROR;
    }


    STDMETHODIMP_(int) OnCachePackageBegin(
        __in IBurnPackage* pPackage,
        __in DWORD cPayloads,
        __in DWORD64 dw64PackageCacheSize
        )
    {
        AssertSz(FALSE, "Elevated view should never be called here.");
        return IDERROR;
    }


    STDMETHODIMP_(int) OnCachePackagePayloadBegin(
        __in_opt IBurnPackage* pPackage,
        __in IBurnPayload* pPayload,
        __in SOURCE_TYPE sourceType,
        __in_z LPCWSTR wzSourcePath
        )
    {
        AssertSz(FALSE, "Elevated view should never be called here.");
        return IDERROR;
    }


    STDMETHODIMP_(int) OnCacheProgress(
        __in_opt IBurnPackage* pPackage,
        __in IBurnPayload* pPayload,
        __in DWORD64 dw64PayloadCached,
        __in DWORD64 dw64PayloadSize
        )
    {
        AssertSz(FALSE, "Elevated view should never be called here.");
        return IDERROR;
    }


    STDMETHODIMP_(void) OnCachePackagePayloadComplete(
        __in_opt IBurnPackage* pPackage,
        __in IBurnPayload* pPayload,
        __in HRESULT hrExitCode
        )
    {
        AssertSz(FALSE, "Elevated view should never be called here.");
        return;
    }


    STDMETHODIMP_(void) OnCachePackageComplete(
        __in IBurnPackage* pPackage,
        __in HRESULT hrExitCode
        )
    {
        AssertSz(FALSE, "Elevated view should never be called here.");
        return;
    }


    STDMETHODIMP_(void) OnCacheComplete(
        __in HRESULT hrStatus
        )
    {
        AssertSz(FALSE, "Elevated view should never be called here.");
        return;
    }


    STDMETHODIMP_(int) OnExecuteBegin(
        __in DWORD cExecutingPackages
        )
    {
        AssertSz(FALSE, "Elevated view should never be called here.");
        return IDERROR;
    }


    STDMETHODIMP_(int) OnExecutePackageBegin(
        __in IBurnPackage* pPackage,
        __in BOOL fExecute
        )
    {
        HRESULT hr = S_OK;
        LPCWSTR wzPackageId = NULL;
        DWORD dwData[2] = { };
        int nResult = IDNOACTION;

        wzPackageId = pPackage->Id();
        hr = m_pModel->GetPackageIndex(wzPackageId, &dwData[0]);
        ExitOnFailure(hr, "Failed to get package index.");

        dwData[1] = fExecute;

        hr = ElevationChildSendMessage(m_hPipe, BURN_ELEVATION_MESSAGE_TYPE_EXECUTE_PACKAGE_BEGIN, dwData, sizeof(dwData), &nResult);
        ExitOnFailure(hr, "Failed to send elevated message for execute package begin.");

    LExit:
        return SUCCEEDED(hr) ? nResult : IDERROR;
    }


    STDMETHODIMP_(int) OnExecuteError(
        __in IBurnPackage* pPackage,
        __in DWORD dwCode,
        __in_z LPCWSTR wzMessage,
        __in UINT uiFlags
        )
    {
        HRESULT hr = S_OK;
        LPCWSTR wzPackageId = NULL;
        DWORD dwPackageId = 0;
        BYTE* pbData = NULL;
        DWORD cbData = 0;
        int nResult = IDNOACTION;

        wzPackageId = pPackage->Id();
        hr = m_pModel->GetPackageIndex(wzPackageId, &dwPackageId);
        ExitOnFailure(hr, "Failed to get package index.");

        hr = BuffWriteNumber(&pbData, &cbData, dwPackageId);
        ExitOnFailure(hr, "Failed to copy package index into message.");

        hr = BuffWriteNumber(&pbData, &cbData, dwCode);
        ExitOnFailure(hr, "Failed to copy error code into message.");

        hr = BuffWriteString(&pbData, &cbData, wzMessage);
        ExitOnFailure(hr, "Failed to copy error message into message.");

        hr = BuffWriteNumber(&pbData, &cbData, uiFlags);
        ExitOnFailure(hr, "Failed to copy UI flags into message.");

        hr = ElevationChildSendMessage(m_hPipe, BURN_ELEVATION_MESSAGE_TYPE_EXECUTE_PACKAGE_ERROR, pbData, cbData, &nResult);
        ExitOnFailure(hr, "Failed to send elevated message for execute package error.");

    LExit:
        ReleaseBuffer(pbData);
        return SUCCEEDED(hr) ? nResult : IDERROR;
    }


    STDMETHODIMP_(int) OnExecuteProgress(
        __in IBurnPackage* pPackage,
        __in DWORD dwPercentage
        )
    {
        HRESULT hr = S_OK;
        LPCWSTR wzPackageId = NULL;
        DWORD dwData[2] = { };
        int nResult = IDNOACTION;

        wzPackageId = pPackage->Id();
        hr = m_pModel->GetPackageIndex(wzPackageId, &dwData[0]);
        ExitOnFailure(hr, "Failed to get package index.");

        dwData[1] = dwPercentage;

        hr = ElevationChildSendMessage(m_hPipe, BURN_ELEVATION_MESSAGE_TYPE_EXECUTE_PACKAGE_PROGRESS, dwData, sizeof(dwData), &nResult);
        ExitOnFailure(hr, "Failed to send elevated message for execute package progress.");

    LExit:
        return SUCCEEDED(hr) ? nResult : IDERROR;
    }


    STDMETHODIMP_(int) OnExecuteMsiMessage(
        __in IBurnMsiPackage* pPackage,
        __in INSTALLMESSAGE mt,
        __in UINT uiFlags,
        __in_z LPCWSTR wzMessage,
        __in_opt MSIHANDLE hRecord
        )
    {
        HRESULT hr = S_OK;
        LPCWSTR wzPackageId = NULL;
        DWORD dwPackageId = 0;
        BYTE* pbData = NULL;
        DWORD cbData = 0;
        int nResult = IDNOACTION;

        wzPackageId = pPackage->Id();
        hr = m_pModel->GetPackageIndex(wzPackageId, &dwPackageId);
        ExitOnFailure(hr, "Failed to get package index.");

        hr = BuffWriteNumber(&pbData, &cbData, dwPackageId);
        ExitOnFailure(hr, "Failed to copy package index into message.");

        hr = BuffWriteNumber(&pbData, &cbData, mt);
        ExitOnFailure(hr, "Failed to copy message type into message.");

        hr = BuffWriteNumber(&pbData, &cbData, uiFlags);
        ExitOnFailure(hr, "Failed to copy UI flags into message.");

        hr = BuffWriteString(&pbData, &cbData, wzMessage);
        ExitOnFailure(hr, "Failed to copy msi message into message.");

        hr = ElevationChildSendMessage(m_hPipe, BURN_ELEVATION_MESSAGE_TYPE_EXECUTE_PACKAGE_MSI_MESSAGE, pbData, cbData, &nResult);
        ExitOnFailure(hr, "Failed to send elevated message for execute package error.");

    LExit:
        ReleaseBuffer(pbData);
        return SUCCEEDED(hr) ? nResult : IDERROR;
    }


    STDMETHODIMP_(void) OnExecutePackageComplete(
        __in IBurnPackage* pPackage,
        __in HRESULT hrExitCode
        )
    {
        HRESULT hr = S_OK;
        LPCWSTR wzPackageId = NULL;
        DWORD dwData[2] = { };
        int nResult = IDNOACTION;

        wzPackageId = pPackage->Id();
        hr = m_pModel->GetPackageIndex(wzPackageId, &dwData[0]);
        ExitOnFailure(hr, "Failed to get package index.");

        dwData[1] = hrExitCode;

        hr = ElevationChildSendMessage(m_hPipe, BURN_ELEVATION_MESSAGE_TYPE_EXECUTE_PACKAGE_COMPLETE, dwData, sizeof(dwData), &nResult);
        ExitOnFailure(hr, "Failed to send elevated message for execute package complete.");

    LExit:
        return;
    }


    STDMETHODIMP_(void) OnExecuteComplete(
        __in HRESULT hrStatus
        )
    {
        AssertSz(FALSE, "Elevated view should never be called here.");
        return;
    }


    STDMETHODIMP_(BOOL) OnRestartRequired()
    {
        if (!m_fInitialized)
        {
            return FALSE;
        }

        HRESULT hr = S_OK;
        BOOL fRestart = FALSE;

        hr = ElevationChildSendMessage(m_hPipe, BURN_ELEVATION_MESSAGE_TYPE_RESTART_REQUIRED, NULL, 0, &fRestart);
        ExitOnFailure(hr, "Failed to send elevated message for restart required.");

    LExit:
        return SUCCEEDED(hr) ? fRestart : FALSE;
    }


    STDMETHODIMP GetControlText(
        __in DWORD dwControlID,
        __out LPWSTR* psczText
        )
    {
        return E_NOTIMPL;
    }


    STDMETHODIMP WaitIfPaused()
    {
        return E_NOTIMPL;
    }

public: // internals
    static HRESULT ConnectToParent(
        __in IBurnModel* pModel,
        __out HANDLE* phPipe
        )
    {
        HRESULT hr = S_OK;
        LPWSTR sczPipeName = NULL;
        LPWSTR sczToken = NULL;

        HANDLE hPipe = INVALID_HANDLE_VALUE;

        DWORD cbVerificationToken = 0;
        LPWSTR sczVerificationToken = NULL;
        DWORD dwRead = 0;

        hr = pModel->GetElevatedInfo(&sczPipeName, &sczToken);
        ExitOnFailure(hr, "Failed to get parent and child elevated tokens.");

        hr = StrAllocPrefix(&sczPipeName, L"\\\\.\\pipe\\", 0);
        ExitOnFailure(hr, "Failed to allocate name of parent pipe.");

        // Connect to the parent.
        hPipe = ::CreateFileW(sczPipeName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
        if (INVALID_HANDLE_VALUE == hPipe)
        {
            ExitWithLastError1(hr, "Failed to open parent pipe: %S", sczPipeName)
        }

        // Read the verification token.
        if (!::ReadFile(hPipe, &cbVerificationToken, sizeof(cbVerificationToken), &dwRead, NULL))
        {
            ExitWithLastError(hr, "Failed to read size of verification token from parent pipe.");
        }

        if (255 < cbVerificationToken / sizeof(WCHAR))
        {
            hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
            ExitOnRootFailure(hr, "Verification token from parent is too big.");
        }

        hr = StrAlloc(&sczVerificationToken, cbVerificationToken / sizeof(WCHAR) + 1);
        ExitOnFailure(hr, "Failed to allocate buffer for verification token.");

        if (!::ReadFile(hPipe, sczVerificationToken, cbVerificationToken, &dwRead, NULL))
        {
            ExitWithLastError(hr, "Failed to read size of verification token from parent pipe.");
        }

        // Verify the tokens match.
        if (CSTR_EQUAL != ::CompareStringW(LOCALE_NEUTRAL, 0, sczVerificationToken, -1, sczToken, -1))
        {
            hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
            ExitOnRootFailure(hr, "Verification token from parent does not match.");
        }

        *phPipe = hPipe;
        hPipe = INVALID_HANDLE_VALUE;

    LExit:
        ReleaseStr(sczVerificationToken);
        ReleaseFileHandle(hPipe);
        ReleaseStr(sczToken);
        ReleaseStr(sczPipeName);
        return hr;
    }

public:
    //
    // Constructor - intitialize member variables.
    //
    CBurnViewElevated()
    {
        m_fInitialized = FALSE;

        m_pModel = NULL;
        m_pController = NULL;
        m_hPipe = INVALID_HANDLE_VALUE;
    }

    //
    // Destructor - release member variables.
    //
    ~CBurnViewElevated()
    {
        AssertSz(!m_fInitialized, "Should not still be initialized in CBurnViewElevated::~CBurnViewElevated().");

        AssertSz(INVALID_HANDLE_VALUE == m_hPipe, "Pipe should have been freed via CBurnViewElevated::Uninitialize().");
        AssertSz(!m_pController, "Controller should have been freed via CBurnViewElevated::Uninitialize().");
        AssertSz(!m_pModel, "Model should have been freed via CBurnViewElevated::Uninitialize().");
    }

private:
    BOOL m_fInitialized;

    IBurnModel* m_pModel;
    IBurnController* m_pController;
    HANDLE m_hPipe;
};


//
// BurnCreateViewElevated - creates a new IBurnView object.
//
HRESULT BurnCreateViewElevated(
    __out IBurnView **ppView
    )
{
    HRESULT hr = S_OK;

    CBurnViewElevated* pView = new CBurnViewElevated();
    ExitOnNull(pView, hr, E_OUTOFMEMORY, "Failed to create new elevated view object.");

    *ppView = pView;
    pView = NULL;

LExit:
    ReleaseObject(pView);
    return hr;
}
