//-------------------------------------------------------------------------------------------------
// <copyright file="elevated.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
// </copyright>
// 
// <summary>
//    Burn elevated process handler.
// </summary>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"


static HRESULT CreateElevatedInfo(
    __inout_z LPWSTR *psczPipeName,
    __inout_z LPWSTR *psczClientToken
    );
static HRESULT CreateElevatedPipe(
    __in LPCWSTR wzPipeName,
    __out HANDLE* phPipe
    );
static HRESULT CreateElevatedProcess(
    __in_opt HWND hwndParent,
    __in IInstallationAbstractionLayer* pInstallationAbstractionLayer,
    __in_z LPCWSTR wzPipeName,
    __in_z LPCWSTR wzToken,
    __out HANDLE* phProcess
    );
static HRESULT ConnectElevatedProcess(
    __in LPCWSTR wzToken,
    __in HANDLE hPipe,
    __in HANDLE hProcess
    );
static HRESULT CacheElevatedPackage(
    __in HANDLE hProcess,
    __in HANDLE hPipe,
    __in IBurnModel* pModel,
    __in IBurnView* pView,
    __in IBurnPackage* pPackage
    );
static HRESULT DoElevatedPackage(
    __in HANDLE hProcess,
    __in HANDLE hPipe,
    __in IBurnModel* pModel,
    __in IBurnView* pView,
    __in IBurnPackage* pPackage,
    __in BOOL fExecute
    );
static HRESULT PumpElevatedMessages(
    __in HANDLE hPipe,
    __in IBurnView* pView,
    __in IBurnPackage* pPackage,
    __in BOOL fExecute
    );
static HRESULT GetElevatedMessage(
    __in HANDLE hPipe,
    __in BURN_ELEVATION_MESSAGE *pMsg
    );
static int ProcessElevatedMessage(
    __in IBurnView* pView,
    __in IBurnPackage* pPackage,
    __in BOOL fExecute,
    __in BURN_ELEVATION_MESSAGE* pMsg
    );
static int ProcessElevatedExecutePackageBegin(
    __in IBurnView* pView,
    __in IBurnPackage* pPackage,
    __in BOOL fExecute,
    __in_bcount_opt(cbData) LPVOID pvData,
    __in DWORD cbData
    );
static int ProcessElevatedExecutePackageError(
    __in IBurnView* pView,
    __in IBurnPackage* pPackage,
    __in_bcount_opt(cbData) LPVOID pvData,
    __in DWORD cbData
    );
static int ProcessElevatedExecutePackageProgress(
    __in IBurnView* pView,
    __in IBurnPackage* pPackage,
    __in_bcount_opt(cbData) LPVOID pvData,
    __in DWORD cbData
    );
static int ProcessElevatedExecutePackageMsiMessage(
    __in IBurnView* pView,
    __in IBurnPackage* pPackage,
    __in_bcount_opt(cbData) LPVOID pvData,
    __in DWORD cbData
    );
static int ProcessElevatedExecutePackageComplete(
    __in IBurnView* pView,
    __in IBurnPackage* pPackage,
    __in_bcount_opt(cbData) LPVOID pvData,
    __in DWORD cbData
    );
static HRESULT WriteMessage(
    __in HANDLE hPipe,
    __in DWORD dwMessage,
    __in_bcount_opt(cbData) LPVOID pvData,
    __in DWORD cbData
    );
static HRESULT AllocateMessage(
    __in DWORD dwMessage,
    __in_bcount_opt(cbData) LPVOID pvData,
    __in DWORD cbData,
    __out_bcount(cb) LPVOID* ppvMessage,
    __out DWORD* cbMessage
    );


extern "C" void ElevationMessageUninitialize(
    __in BURN_ELEVATION_MESSAGE *pMsg
    )
{
    if (pMsg->fAllocatedData)
    {
        ReleaseNullMem(pMsg->pvData);
        pMsg->fAllocatedData = FALSE;
    }
}


extern "C" HRESULT ElevationParentProcessConnect(
    __in HWND hwndParent,
    __in IInstallationAbstractionLayer* pInstallationAbstractionLayer,
    __out HANDLE* phElevatedProcess,
    __out HANDLE* phElevatedPipe
    )
{
    HRESULT hr = S_OK;
    LPWSTR sczPipeName = NULL;
    LPWSTR sczClientToken = NULL;
    HANDLE hPipe = INVALID_HANDLE_VALUE;
    HANDLE hProcess = NULL;

    hr = CreateElevatedInfo(&sczPipeName, &sczClientToken);
    ExitOnFailure(hr, "Failed to allocate elevation information.");

    hr = CreateElevatedPipe(sczPipeName, &hPipe);
    ExitOnFailure(hr, "Failed to create elevated pipe.");

    hr = CreateElevatedProcess(hwndParent, pInstallationAbstractionLayer, sczPipeName, sczClientToken, &hProcess);
    ExitOnFailure(hr, "Failed to create elevated process.");

    hr = ConnectElevatedProcess(sczClientToken, hPipe, hProcess);
    ExitOnFailure(hr, "Failed to connect to elevated process.");

    *phElevatedPipe = hPipe;
    hPipe = INVALID_HANDLE_VALUE;
    *phElevatedProcess = hProcess;
    hProcess = NULL;

LExit:
    if (hProcess)
    {
        ::CloseHandle(hProcess);
    }

    ReleaseFileHandle(hPipe);
    ReleaseStr(sczClientToken);
    ReleaseStr(sczPipeName);
    return hr;
}


extern "C" HRESULT ElevationParentPackageExecute(
    __in HANDLE hElevatedProcess,
    __in HANDLE hElevatedPipe,
    __in IBurnModel* pModel,
    __in IBurnView* pView,
    __in IBurnPackage* pPackage
    )
{
    HRESULT hr = S_OK;

    hr = CacheElevatedPackage(hElevatedProcess, hElevatedPipe, pModel, pView, pPackage);
    ExitOnFailure(hr, "Failed to cache elevated package.");

    hr = DoElevatedPackage(hElevatedProcess, hElevatedPipe, pModel, pView, pPackage, TRUE);
    ExitOnFailure(hr, "Failed to execute elevated package.");

LExit:
    return hr;
}


extern "C" HRESULT ElevationParentPackageRollback(
    __in HANDLE hElevatedProcess,
    __in HANDLE hElevatedPipe,
    __in IBurnModel* pModel,
    __in IBurnView* pView,
    __in IBurnPackage* pPackage
    )
{
    return DoElevatedPackage(hElevatedProcess, hElevatedPipe, pModel, pView, pPackage, FALSE);
}


extern "C" HRESULT ElevationChildGetMessage(
    __in HANDLE hPipe,
    __in BURN_ELEVATION_MESSAGE* pMsg
    )
{
    HRESULT hr = S_OK;
    DWORD rgdwMessageAndByteCount[2] = { };
    DWORD cb = 0;
    DWORD cbRead = 0;

    while (cbRead < sizeof(rgdwMessageAndByteCount))
    {
        if (!::ReadFile(hPipe, reinterpret_cast<BYTE*>(rgdwMessageAndByteCount) + cbRead, sizeof(rgdwMessageAndByteCount) - cbRead, &cb, NULL))
        {
            DWORD er = ::GetLastError();
            if (ERROR_MORE_DATA == er)
            {
                hr = S_OK;
            }
            else if (ERROR_BROKEN_PIPE == er) // parent process shut down, time to exit.
            {
                memset(rgdwMessageAndByteCount, 0, sizeof(rgdwMessageAndByteCount));
                hr = S_FALSE;
                break;
            }
            else
            {
                hr = HRESULT_FROM_WIN32(er);
            }
            ExitOnRootFailure(hr, "Failed to read message from pipe.");
        }

        cbRead += cb;
    }

    pMsg->dwMessage = rgdwMessageAndByteCount[0];
    pMsg->cbData = rgdwMessageAndByteCount[1];
    if (pMsg->cbData)
    {
        pMsg->pvData = MemAlloc(pMsg->cbData, FALSE);
        ExitOnNull(pMsg->pvData, hr, E_OUTOFMEMORY, "Failed to allocate data for message.");

        if (!::ReadFile(hPipe, pMsg->pvData, pMsg->cbData, &cb, NULL))
        {
            ExitWithLastError(hr, "Failed to read data for message.");
        }

        pMsg->fAllocatedData = TRUE;
    }

LExit:
    if (!pMsg->fAllocatedData && pMsg->pvData)
    {
        MemFree(pMsg->pvData);
    }

    return hr;
}


extern "C" HRESULT ElevationChildPostMessage(
    __in HANDLE hPipe,
    __in DWORD dwMessage,
    __in_bcount_opt(cbData) LPVOID pvData,
    __in DWORD cbData
    )
{
    HRESULT hr = S_OK;

    hr = WriteMessage(hPipe, dwMessage, pvData, cbData);
    ExitOnFailure(hr, "Failed to write message to parent process.");

LExit:
    return hr;
}


extern "C" HRESULT ElevationChildSendMessage(
    __in HANDLE hPipe,
    __in DWORD dwMessage,
    __in_bcount_opt(cbData) LPVOID pvData,
    __in DWORD cbData,
    __out int* pnResult
    )
{
    HRESULT hr = S_OK;
    DWORD cb = 0;

    // TODO: change these write/read calls into single TransactMessage (aka: TransactNamedPipe) call.
    hr = WriteMessage(hPipe, dwMessage, pvData, cbData);
    ExitOnFailure(hr, "Failed to write message to parent process.");

    if (!::ReadFile(hPipe, pnResult, sizeof(int), &cb, NULL))
    {
        ExitWithLastError(hr, "Failed to read response from pipe.");
    }

LExit:
    return hr;
}


static HRESULT CreateElevatedInfo(
    __inout_z LPWSTR *psczPipeName,
    __inout_z LPWSTR *psczClientToken
    )
{
    HRESULT hr = S_OK;
    RPC_STATUS rs = RPC_S_OK;
    UUID guid = { };
    WCHAR wzGuid[39];

    // Create the unique pipe name.
    rs = ::UuidCreate(&guid);
    hr = HRESULT_FROM_RPC(rs);
    ExitOnFailure(hr, "Failed to create pipe guid.");

    if (!::StringFromGUID2(guid, wzGuid, countof(wzGuid)))
    {
        hr = E_OUTOFMEMORY;
        ExitOnRootFailure(hr, "Failed to convert pipe guid into string.");
    }

    hr = StrAllocFormatted(psczPipeName, L"BurnPipe.%s", wzGuid);
    ExitOnFailure(hr, "Failed to allocate pipe name.");

    // Create the unique client token.
    rs = ::UuidCreate(&guid);
    hr = HRESULT_FROM_RPC(rs);
    ExitOnRootFailure(hr, "Failed to create pipe guid.");

    if (!::StringFromGUID2(guid, wzGuid, countof(wzGuid)))
    {
        hr = E_OUTOFMEMORY;
        ExitOnRootFailure(hr, "Failed to convert pipe guid into string.");
    }

    hr = StrAllocFormatted(psczClientToken, L"%s", wzGuid);
    ExitOnFailure(hr, "Failed to allocate client token.");

LExit:
    return hr;
}


static HRESULT CreateElevatedPipe(
    __in LPCWSTR wzPipeName,
    __out HANDLE* phPipe
    )
{
    HRESULT hr = S_OK;
    LPWSTR sczFullPipeName = NULL;

    hr = StrAllocFormatted(&sczFullPipeName, L"\\\\.\\pipe\\%s", wzPipeName);
    ExitOnFailure1(hr, "Failed to allocate full name of pipe: %S", wzPipeName);

    // TODO: use a different security descriptor to lock down this named pipe even more than it is now.
    // TODO: consider using overlapped IO to do waits on the pipe and still be able to cancel and such.
    *phPipe = ::CreateNamedPipeW(sczFullPipeName, PIPE_ACCESS_DUPLEX | FILE_FLAG_FIRST_PIPE_INSTANCE, PIPE_TYPE_BYTE | PIPE_READMODE_BYTE | PIPE_WAIT, 1, 64 * 1024, 64 * 1024, 1, NULL);
    if (INVALID_HANDLE_VALUE == *phPipe)
    {
        ExitWithLastError1(hr, "Failed to create elevated pipe: %S", sczFullPipeName);
    }

LExit:
    ReleaseStr(sczFullPipeName);
    return hr;
}


static HRESULT CreateElevatedProcess(
    __in_opt HWND hwndParent,
    __in IInstallationAbstractionLayer* pInstallationAbstractionLayer,
    __in_z LPCWSTR wzPipeName,
    __in_z LPCWSTR wzToken,
    __out HANDLE* phProcess
    )
{
    HRESULT hr = S_OK;
    LPWSTR sczBurnPath = NULL;
    LPWSTR sczParameters = NULL;
    SHELLEXECUTEINFOW info = { };

    hr = PathForCurrentProcess(&sczBurnPath, NULL);
    ExitOnFailure(hr, "Failed to get current process path.");

    hr = StrAllocFormatted(&sczParameters, L"-burn.elevated %s %s", wzPipeName, wzToken);
    ExitOnFailure(hr, "Failed to allocate parameters for elevated process.");

    info.cbSize = sizeof(SHELLEXECUTEINFOW);
    info.fMask = SEE_MASK_FLAG_DDEWAIT | SEE_MASK_FLAG_NO_UI | SEE_MASK_NOCLOSEPROCESS;
    info.hwnd = hwndParent;
    info.lpFile = sczBurnPath;
    info.lpParameters = sczParameters;
    info.lpVerb = L"runas";
    info.nShow = SW_HIDE;

    hr = pInstallationAbstractionLayer->ShellExecute(&info);
    ExitOnFailure2(hr, "Failed to launch elevated process: %S %S", sczBurnPath, sczParameters);

    *phProcess = info.hProcess;
    info.hProcess = NULL;

LExit:
    if (info.hProcess)
    {
        ::CloseHandle(info.hProcess);
    }

    ReleaseStr(sczParameters);
    ReleaseStr(sczBurnPath);
    return hr;
}


static HRESULT ConnectElevatedProcess(
    __in LPCWSTR wzToken,
    __in HANDLE hPipe,
    __in HANDLE hProcess
    )
{
    HRESULT hr = S_OK;
    DWORD cbToken = lstrlenW(wzToken) * sizeof(WCHAR);
    DWORD dwAck = 0;
    DWORD cb = 0;

    UNREFERENCED_PARAMETER(hProcess); // TODO: use the hProcess to determine if the elevated process dies before/while waiting for pipe communcation.

    if (!::ConnectNamedPipe(hPipe, NULL))
    {
        DWORD er = ::GetLastError();
        if (ERROR_PIPE_CONNECTED != er)
        {
            hr = HRESULT_FROM_WIN32(er);
            ExitOnRootFailure(hr, "Failed to connect elevated pipe.");
        }
    }

    // Prove we are the one that created the elevated process by passing the token.
    if (!::WriteFile(hPipe, &cbToken, sizeof(cbToken), &cb, NULL))
    {
        ExitWithLastError(hr, "Failed to write token length to elevated pipe.");
    }

    if (!::WriteFile(hPipe, wzToken, cbToken, &cb, NULL))
    {
        ExitWithLastError(hr, "Failed to write token to elevated pipe.");
    }

    // Wait until the elevated process responds that it is ready to go.
    if (!::ReadFile(hPipe, &dwAck, sizeof(dwAck), &cb, NULL))
    {
        ExitWithLastError(hr, "Failed to read ACK from elevated pipe.");
    }

    if (1 != dwAck)
    {
        hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
        ExitOnRootFailure1(hr, "Incorrect ACK from elevated pipe: %u", dwAck);
    }

LExit:
    return hr;
}


static HRESULT CacheElevatedPackage(
    __in HANDLE hProcess,
    __in HANDLE hPipe,
    __in IBurnModel* pModel,
    __in IBurnView* pView,
    __in IBurnPackage* pPackage
    )
{
    HRESULT hr = S_OK;
    DWORD rgdwData[3] = { };
    ACTION_STATE execute = ACTION_STATE_NONE;
    ACTION_STATE rollback = ACTION_STATE_NONE;

    LPCWSTR wzPackageId = pPackage->Id();
    hr = pModel->GetPackageIndex(wzPackageId, &rgdwData[0]);
    ExitOnFailure(hr, "Failed to get package index for elevated package to cache.");

    pPackage->GetActionState(&execute, &rollback);
    rgdwData[1] = execute;
    rgdwData[2] = rollback;

    hr = WriteMessage(hPipe, BURN_ELEVATION_MESSAGE_TYPE_CACHE, &rgdwData, sizeof(rgdwData));
    ExitOnFailure(hr, "Failed to write cache message to child process.");

    hr = PumpElevatedMessages(hPipe, pView, pPackage, FALSE);
    ExitOnFailure(hr, "Failure while pumping elevated messages for caching package.");

LExit:
    return hr;
}


static HRESULT DoElevatedPackage(
    __in HANDLE hProcess,
    __in HANDLE hPipe,
    __in IBurnModel* pModel,
    __in IBurnView* pView,
    __in IBurnPackage* pPackage,
    __in BOOL fExecute
    )
{
    HRESULT hr = S_OK;
    DWORD dwMessage = fExecute ? BURN_ELEVATION_MESSAGE_TYPE_EXECUTE : BURN_ELEVATION_MESSAGE_TYPE_ROLLBACK;
    DWORD dwPackageIndex = 0;

    LPCWSTR wzPackageId = pPackage->Id();
    hr = pModel->GetPackageIndex(wzPackageId, &dwPackageIndex);
    ExitOnFailure(hr, "Failed to get package index.");

    // TODO: serialize the Model's properities to pass to the elevated process.

    hr = WriteMessage(hPipe, dwMessage, &dwPackageIndex, sizeof(dwPackageIndex));
    ExitOnFailure(hr, "Failed to write message to child process.");

    hr = PumpElevatedMessages(hPipe, pView, pPackage, fExecute);
    ExitOnFailure(hr, "Failure while pumping elevated messages.");

LExit:
    return hr;
}


static HRESULT PumpElevatedMessages(
    __in HANDLE hPipe,
    __in IBurnView* pView,
    __in IBurnPackage* pPackage,
    __in BOOL fExecute
    )
{
    HRESULT hr = S_OK;
    BURN_ELEVATION_MESSAGE msg = { };
    DWORD cb = 0;

    while (S_OK == (hr = GetElevatedMessage(hPipe, &msg)))
    {
        if (BURN_ELEVATION_MESSAGE_TYPE_COMPLETE == msg.dwMessage)
        {
            hr = *static_cast<HRESULT*>(msg.pvData);

            Trace1(REPORT_STANDARD, "Worker thread received complete message, hr = 0x%x.", hr);
            break;
        }

        // Process the message then send the view's response back to the child process so it
        // can continue or cancel as appropriate.
        int nResult = ProcessElevatedMessage(pView, pPackage, fExecute, &msg);
        if (!::WriteFile(hPipe, &nResult, sizeof(nResult), &cb, NULL))
        {
            ExitWithLastError(hr, "Failed to write message response to elevated pipe.");
        }

        ElevationMessageUninitialize(&msg);
    }

LExit:
    ElevationMessageUninitialize(&msg);
    return hr;
}


static HRESULT GetElevatedMessage(
    __in HANDLE hPipe,
    __in BURN_ELEVATION_MESSAGE *pMsg
    )
{
    HRESULT hr = S_OK;
    DWORD cbRead = 0;
    LPVOID pv = NULL;

    // Blank out the message in case anything goes wrong below.
    memset(pMsg, 0, sizeof(*pMsg));

    // Read the pMsg->dwMessage (first DWORD) and pMsg->cbData (second DWORD)
    // for the message out of the pipe.
    if (!::ReadFile(hPipe, pMsg, sizeof(DWORD) * 2, &cbRead, NULL))
    {
        ExitWithLastError(hr, "Failed to read message from pipe.");
    }

    // Finally, if there is data for this message then read all that.
    if (pMsg->cbData)
    {
        pv = MemAlloc(pMsg->cbData, TRUE);
        ExitOnNull(pv, hr, E_OUTOFMEMORY, "Failed to allocate memory for data in pipe.");

        if (!::ReadFile(hPipe, pv, pMsg->cbData, &cbRead, NULL))
        {
            ExitWithLastError(hr, "Failed to read extra data from pipe.");
        }

        pMsg->fAllocatedData = TRUE;
        pMsg->pvData = pv;
        pv = NULL;
    }

LExit:
    ReleaseMem(pv);

    return hr;
}


static int ProcessElevatedMessage(
    __in IBurnView* pView,
    __in IBurnPackage* pPackage,
    __in BOOL fExecute,
    __in BURN_ELEVATION_MESSAGE* pMsg
    )
{
    int nResult = IDERROR;

    switch (pMsg->dwMessage)
    {
    case BURN_ELEVATION_MESSAGE_TYPE_EXECUTE_PACKAGE_BEGIN:
        nResult = ProcessElevatedExecutePackageBegin(pView, pPackage, fExecute, pMsg->pvData, pMsg->cbData);
        break;

    case BURN_ELEVATION_MESSAGE_TYPE_EXECUTE_PACKAGE_ERROR:
        nResult = ProcessElevatedExecutePackageError(pView, pPackage, pMsg->pvData, pMsg->cbData);
        break;

    case BURN_ELEVATION_MESSAGE_TYPE_EXECUTE_PACKAGE_PROGRESS:
        nResult = ProcessElevatedExecutePackageProgress(pView, pPackage, pMsg->pvData, pMsg->cbData);
        break;

    case BURN_ELEVATION_MESSAGE_TYPE_EXECUTE_PACKAGE_MSI_MESSAGE:
        nResult = ProcessElevatedExecutePackageMsiMessage(pView, pPackage, pMsg->pvData, pMsg->cbData);
        break;

    case BURN_ELEVATION_MESSAGE_TYPE_EXECUTE_PACKAGE_COMPLETE:
        nResult = ProcessElevatedExecutePackageComplete(pView, pPackage, pMsg->pvData, pMsg->cbData);
        break;

    case BURN_ELEVATION_MESSAGE_TYPE_RESTART_REQUIRED:
        {
        BOOL fResult = pView->OnRestartRequired();
        nResult = static_cast<int>(fResult);
        }
        break;

    // TODO: process the remaining messages.
    //case BURN_ELEVATION_MESSAGE_TYPE_CHILD_UNINITIALIZE:
    //    break;

    default: // unknown message type, abort.
        AssertSz(FALSE, "Unknown elevated message type to process.");
        break;
    }

    return nResult;
}


static int ProcessElevatedExecutePackageBegin(
    __in IBurnView* pView,
    __in IBurnPackage* pPackage,
    __in BOOL fExecute,
    __in_bcount_opt(cbData) LPVOID pvData,
    __in DWORD cbData
    )
{
    HRESULT hr = S_OK;
    int nResult = IDNOACTION;

    if (sizeof(DWORD) * 2 != cbData)
    {
        hr = E_INVALIDARG;
        ExitOnRootFailure(hr, "Invalid data size passed to execute package begin.");
    }

#ifdef DEBUG
    DWORD* rgdwData = static_cast<DWORD*>(pvData);
    AssertSz(fExecute == rgdwData[1], "Elevated execute package begin fExecute didn't match expectations");
#endif

    nResult = pView->OnExecutePackageBegin(pPackage, fExecute);

LExit:
    return FAILED(hr) ? IDERROR : nResult;
}


static int ProcessElevatedExecutePackageError(
    __in IBurnView* pView,
    __in IBurnPackage* pPackage,
    __in_bcount_opt(cbData) LPVOID pvData,
    __in DWORD cbData
    )
{
    HRESULT hr = S_OK;
    BYTE* pb = static_cast<BYTE*>(pvData);
    DWORD ib = 0;
    DWORD dwPackageIndex = 0;
    DWORD dwErrorCode = 0;
    LPWSTR sczMessage = NULL;
    UINT uiFlags = 0;
    int nResult = IDNOACTION;

    hr = BuffReadNumber(pb, cbData, &ib, &dwPackageIndex);
    ExitOnFailure(hr, "Failed to read package index from elevated package error message.");

    hr = BuffReadNumber(pb, cbData, &ib, &dwErrorCode);
    ExitOnFailure(hr, "Failed to read error code from elevated package error message.");

    hr = BuffReadString(pb, cbData, &ib, &sczMessage);
    ExitOnFailure(hr, "Failed to read UI flags from elevated package error message.");

    hr = BuffReadNumber(pb, cbData, &ib, reinterpret_cast<DWORD*>(&uiFlags));
    ExitOnFailure(hr, "Failed to read UI flags from elevated package error message.");

    nResult = pView->OnExecuteError(pPackage, dwErrorCode, sczMessage, uiFlags);

LExit:
    ReleaseStr(sczMessage);
    return FAILED(hr) ? IDERROR : nResult;
}


static int ProcessElevatedExecutePackageProgress(
    __in IBurnView* pView,
    __in IBurnPackage* pPackage,
    __in_bcount_opt(cbData) LPVOID pvData,
    __in DWORD cbData
    )
{
    HRESULT hr = S_OK;
    DWORD* rgdwData = NULL;
    int nResult = IDNOACTION;

    if (sizeof(DWORD) * 2 != cbData)
    {
        hr = E_INVALIDARG;
        ExitOnRootFailure(hr, "Invalid data size passed to execute package progress.");
    }

    rgdwData = static_cast<DWORD*>(pvData);

    nResult = pView->OnExecuteProgress(pPackage, rgdwData[1]);

LExit:
    return FAILED(hr) ? IDERROR : nResult;
}


static int ProcessElevatedExecutePackageMsiMessage(
    __in IBurnView* pView,
    __in IBurnPackage* pPackage,
    __in_bcount_opt(cbData) LPVOID pvData,
    __in DWORD cbData
    )
{
    HRESULT hr = S_OK;
    BYTE* pb = static_cast<BYTE*>(pvData);
    DWORD ib = 0;
    DWORD dwPackageIndex = 0;
    INSTALLMESSAGE mt = INSTALLMESSAGE_FATALEXIT;
    UINT uiFlags = 0;
    LPWSTR sczMessage = NULL;
    IBurnMsiPackage* pMsiPackage = NULL;
    int nResult = IDNOACTION;

    hr = BuffReadNumber(pb, cbData, &ib, &dwPackageIndex);
    ExitOnFailure(hr, "Failed to read package index from elevated package error message.");

    hr = BuffReadNumber(pb, cbData, &ib, reinterpret_cast<DWORD*>(&mt));
    ExitOnFailure(hr, "Failed to read message type from elevated package error message.");

    hr = BuffReadNumber(pb, cbData, &ib, reinterpret_cast<DWORD*>(&uiFlags));
    ExitOnFailure(hr, "Failed to read UI flags from elevated package error message.");

    hr = BuffReadString(pb, cbData, &ib, &sczMessage);
    ExitOnFailure(hr, "Failed to read UI flags from elevated package error message.");

    hr = pPackage->QueryInterface(IID_PPV_ARGS(&pMsiPackage));
    ExitOnFailure(hr, "Failed to query for MSI package.");

    nResult = pView->OnExecuteMsiMessage(pMsiPackage, mt, uiFlags, sczMessage, NULL);

LExit:
    ReleaseStr(sczMessage);
    ReleaseObject(pMsiPackage);
    return FAILED(hr) ? IDERROR : nResult;
}


static int ProcessElevatedExecutePackageComplete(
    __in IBurnView* pView,
    __in IBurnPackage* pPackage,
    __in_bcount_opt(cbData) LPVOID pvData,
    __in DWORD cbData
    )
{
    HRESULT hr = S_OK;
    DWORD* rgdwData = NULL;

    if (sizeof(DWORD) * 2 != cbData)
    {
        hr = E_INVALIDARG;
        ExitOnRootFailure(hr, "Invalid data size passed to execute package complete.");
    }

    rgdwData = static_cast<DWORD*>(pvData);

    pView->OnExecutePackageComplete(pPackage, rgdwData[1]);

LExit:
    return FAILED(hr) ? IDERROR : IDNOACTION;
}


static HRESULT WriteMessage(
    __in HANDLE hPipe,
    __in DWORD dwMessage,
    __in_bcount_opt(cbData) LPVOID pvData,
    __in DWORD cbData
    )
{
    HRESULT hr = S_OK;
    LPVOID pv = NULL;
    DWORD cb = 0;

    hr = AllocateMessage(dwMessage, pvData, cbData, &pv, &cb);
    ExitOnFailure(hr, "Failed to allocate message to write.");

    // Write the message.
    DWORD cbWrote = 0;
    DWORD cbTotalWritten = 0;
    do
    {
        if (!::WriteFile(hPipe, pv, cb - cbTotalWritten, &cbWrote, NULL))
        {
            ExitWithLastError(hr, "Failed to write message type to pipe.");
        }

        cbTotalWritten += cbWrote;
    } while (cbTotalWritten < cb);

LExit:
    ReleaseMem(pv);
    return hr;
}


static HRESULT AllocateMessage(
    __in DWORD dwMessage,
    __in_bcount_opt(cbData) LPVOID pvData,
    __in DWORD cbData,
    __out_bcount(cb) LPVOID* ppvMessage,
    __out DWORD* cbMessage
    )
{
    HRESULT hr = S_OK;
    LPVOID pv = NULL;
    DWORD cb = 0;

    // If no data was provided, ensure the count of bytes is zero.
    if (!pvData)
    {
        cbData = 0;
    }

    // Allocate the message.
    cb = sizeof(dwMessage) + sizeof(cbData) + cbData;
    pv = MemAlloc(cb, FALSE);
    ExitOnNull(pv, hr, E_OUTOFMEMORY, "Failed to allocate memory for message.");

    memcpy_s(pv, cb, &dwMessage, sizeof(dwMessage));
    memcpy_s(static_cast<BYTE*>(pv) + sizeof(dwMessage), cb - sizeof(dwMessage), &cbData, sizeof(cbData));
    if (cbData)
    {
        memcpy_s(static_cast<BYTE*>(pv) + sizeof(dwMessage) + sizeof(cbData), cb - sizeof(dwMessage) - sizeof(cbData), pvData, cbData);
    }

    *cbMessage = cb;
    *ppvMessage = pv;
    pv = NULL;

LExit:
    ReleaseMem(pv);
    return hr;
}
