//-------------------------------------------------------------------------------------------------
// <copyright file="manifest.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
// 
// <summary>
//    Burn manifest parser.
// </summary>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"


static HRESULT ParseManifestDocument(
    __in REFGUID guidBundleId,
    __in LPCWSTR sczBundleFileName,
    __in IXMLDOMDocument* pixd,
    __in DWORD dwPayloadsOffset,
    __in IBurnModel* pModel,
    __out DWORD* pcPackages,
    __out IBurnPackage*** prgpPackages,
    __out DWORD* pcUXPayloads,
    __out IBurnPayload*** prgpPayloads,
    __in BURN_SEARCHES* pSearches
    );
static HRESULT ParseManifestChain(
    __in REFGUID guidBundleId,
    __in LPCWSTR sczBundleFileName,
    __in IXMLDOMNode* pixnChain,
    __in DWORD dwPayloadsOffset,
    __in IBurnModel* pModel,
    __out DWORD* pcPackages,
    __out IBurnPackage*** prgpPackages
    );
static HRESULT ParseManifestMsiPackage(
    __in REFGUID guidBundleId,
    __in LPCWSTR sczBundleFileName,
    __in IXMLDOMNode* pixnMsiPackage,
    __in DWORD dwPayloadsOffset,
    __in IBurnModel* pModel,
    __out IBurnMsiPackage** ppPackage
    );
static HRESULT ParseManifestExePackage(
    __in REFGUID guidBundleId,
    __in LPCWSTR sczBundleFileName,
    __in IXMLDOMNode* pixnExePackage,
    __in DWORD dwPayloadsOffset,
    __in IBurnModel* pModel,
    __out IBurnExePackage** ppPackage
    );
static HRESULT ParseManifestExeExitInfo(
    __in IXMLDOMNode* pixnExitInfoNode,
    __in BURN_EXE_PACKAGE_EXIT_INFO* pExitInfo
    );
static HRESULT ParseManifestDirectorySearch(
    __in IXMLDOMNode* pixnDirectorySearch,
    __in BURN_SEARCHES* pSearches
    );
static HRESULT ParseManifestFileSearch(
    __in IXMLDOMNode* pixnFileSearch,
    __in BURN_SEARCHES* pSearches
    );
static HRESULT ParseManifestRegistrySearch(
    __in IXMLDOMNode* pixnRegistrySearch,
    __in BURN_SEARCHES* pSearches
    );
static HRESULT ParseCommonSearchAttributes(
    __in BSTR bstrNodeName,
    __in BSTR bstrNodeValue,
    __in BURN_SEARCH* pSearch
    );
static HRESULT ParseCommonSearchChildElements(
    __in BSTR bstrNodeName,
    __in IXMLDOMNode* pNode,
    __in BURN_SEARCH* pSearch
    );
static HRESULT ParseManifestSearchDependency(
    __in IXMLDOMNode* pixnSearchDependency,
    __in BURN_SEARCH* pSearch
    );


/********************************************************************
 ManifestInitialize - Initialize manifest utilities.

*********************************************************************/
HRESULT WIXAPI ManifestInitialize()
{
    HRESULT hr = CabInitialize(FALSE);
    ExitOnFailure(hr, "Failed to initialize cabinet support.");

    hr = XmlInitialize();
    ExitOnFailure(hr, "Failed to initialize XML support.");

LExit:
    return hr;
}


/********************************************************************
 ManifestUninitialize - Uninitialize manifest utilities.

*********************************************************************/
void WIXAPI ManifestUninitialize()
{
    XmlUninitialize();
    CabUninitialize();
}


/********************************************************************
 ManifestParseFromFile - parses out a Burn manifest from a file path.
    TODO: Broken right now because we don't have support for 
    specifying UX information except via embedded manifest. Priority?
*********************************************************************/
HRESULT WIXAPI ManifestParseFromFile(
    __in LPCWSTR wzManifestFile,
    __out SETUP_PACKAGES** ppSetupPackages,
    __out SETUP_UX_RESOURCES** ppUXResources
    )
{
    Assert(wzManifestFile);
    Assert(ppSetupPackages);
    Assert(ppUXResources);

    HRESULT hr = S_OK;
//    SETUP_PACKAGES* pSetupPackages = NULL;
//    SETUP_UX_RESOURCES* pUXResources = NULL;
//    IXMLDOMDocument* pixdManifest = NULL;
//
//    hr = XmlLoadDocumentFromFile(wzManifestFile, &pixdManifest);
//    ExitOnFailure(hr, "Failed to load manifest as XML document.");
//
//    hr = ParseManifestDocument(pixdManifest, &pSetupPackages, &pUXResources);
//    ExitOnFailure(hr, "Failed to parse manifest document.");
//
//    *ppSetupPackages = pSetupPackages;
//    pSetupPackages = NULL;
//    *ppUXResources = pUXResources;
//    pUXResources = NULL;
//
//LExit:
//    ReleaseObject(pixdManifest);
//    FreeSetupPackages(pSetupPackages);
//    FreeUXResources(pUXResources);
//
    return hr;
}


/********************************************************************
 ManifestParseFromEmbedded - parses out a Burn manifest embedded as a
    payload in the current module.
*********************************************************************/
HRESULT WIXAPI ManifestParseFromEmbedded(
    __in LPCWSTR sczFileName,
    __in IBurnModel* pModel,
    __out IBurnPackage*** prgpPackages,
    __out DWORD* pcPackages,
    __out IBurnPayload*** prgpUXPayloads,
    __out DWORD* pcUXPayloads,
    __in BURN_SEARCHES* pSearches
    )
{
    Assert(prgpPackages);
    Assert(pcPackages);
    Assert(prgpUXPayloads);
    Assert(pcUXPayloads);

    HRESULT hr = S_OK;
    LPBYTE pbManifest = NULL;
    DWORD cbManifest = 0;
    IBurnPackage** rgpPackages = NULL;
    IBurnPayload** rgpUXPayloads = NULL;
    IXMLDOMDocument* pixdManifest = NULL;
    GUID guidBundleId = {0};
    DWORD dwPayloadsOffset = 0;

    hr = PayloadGetManifest(sczFileName, &pbManifest, &cbManifest, &guidBundleId, &dwPayloadsOffset);
    ExitOnFailure(hr, "Failed to get embedded manifest");

    hr = XmlLoadDocumentFromBuffer(pbManifest, cbManifest, &pixdManifest);
    ExitOnFailure(hr, "Failed to load manifest as XML document.");

    hr = ParseManifestDocument(guidBundleId, sczFileName, pixdManifest, dwPayloadsOffset, pModel, pcPackages, &rgpPackages, pcUXPayloads, &rgpUXPayloads, pSearches);
    ExitOnFailure(hr, "Failed to parse manifest document.");

    *prgpPackages = rgpPackages;
    rgpPackages = NULL;
    *prgpUXPayloads = rgpUXPayloads;
    rgpUXPayloads = NULL;

LExit:
    ReleaseMem(pbManifest);
    ReleaseObject(pixdManifest);
    ReleaseObjectArray(rgpPackages, *pcPackages);
    ReleaseObjectArray(rgpUXPayloads, *pcUXPayloads);

    return hr;
}


/********************************************************************
 ParseManifestPayload - parses the payload information from UX, 
    Resource, and package elements and returns an IBurnPayload.
*********************************************************************/
static HRESULT ParseManifestPayload(
    __in REFGUID guidBundleId,
    __in LPCWSTR sczBundleFileName,
    __in IXMLDOMNode* pNode,
    __in DWORD dwPayloadsOffset,
    __in BOOL fTemporary,
    __out IBurnPayload** ppPayload
    )
{
    HRESULT hr = S_OK;
    IXMLDOMNamedNodeMap* pixnnmAttributes = NULL;
    IXMLDOMNode* pixnAttribute = NULL;
    BSTR bstrNodeName = NULL;
    BSTR bstrNodeValue = NULL;
    LPWSTR sczFileName = NULL;
    LPWSTR sczLocalName = NULL;
    LPWSTR sczDownloadSource = NULL;
    LPWSTR sczRelativeDirectory = NULL;
    LPWSTR sczSourceDirectory = NULL;
    LPWSTR sczEmbeddedId = NULL;
    DWORD64 cbSize = 0;
    DWORD dwOffset = 0;
    SOURCE_TYPE sourceType = SOURCE_TYPE_EMBEDDED;

    hr = pNode->get_attributes(&pixnnmAttributes);
    ExitOnFailure(hr, "Failed to get attributes on Resource element.");

    while (S_OK == (hr = pixnnmAttributes->nextNode(&pixnAttribute)))
    {
        // get name and value for each attribute
        hr = pixnAttribute->get_nodeName(&bstrNodeName);
        ExitOnFailure(hr, "Failed to get Resource attribute name.");

        hr = XmlGetText(pixnAttribute, &bstrNodeValue);
        ExitOnFailure1(hr, "Failed to get Resource attribute value: %S", bstrNodeName);

        if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"FileName", -1))
        {
            hr = StrAllocString(&sczFileName, bstrNodeValue, 0);
            ExitOnFailure(hr, "Failed to allocate @FileName.");
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"DownloadSource", -1))
        {
            hr = StrAllocString(&sczDownloadSource, bstrNodeValue, 0);
            ExitOnFailure(hr, "Failed to allocate @DownloadSource.");
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"EmbeddedId", -1))
        {
            hr = StrAllocString(&sczEmbeddedId, bstrNodeValue, 0);
            ExitOnFailure(hr, "Failed to allocate @EmbeddedId.");
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"FileSize", -1))
        {
            cbSize = static_cast<DWORD64>(_wcstoui64(bstrNodeValue, NULL, 10));
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Offset", -1))
        {
            dwOffset = static_cast<DWORD>(wcstoul(bstrNodeValue, NULL, 10));
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Packaging", -1))
        {
            if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"embedded", -1))
            {
                sourceType = SOURCE_TYPE_EMBEDDED;
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"download", -1))
            {
                sourceType = SOURCE_TYPE_DOWNLOAD;
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"external", -1))
            {
                sourceType = SOURCE_TYPE_EXTERNAL;
            }
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"RelativeDirectory", -1))
        {
            hr = StrAllocString(&sczRelativeDirectory, bstrNodeValue, 0);
            ExitOnFailure(hr, "Failed to allocate @RelativePath.");
        }

        ReleaseNullBSTR(bstrNodeValue);
        ReleaseNullBSTR(bstrNodeName);
    }

    if (S_FALSE == hr)
    {
        hr = S_OK;
    }
    ExitOnFailure(hr, "Failed to enumerate all payload attributes.");

    // we have the data from the manifest, now create the payload object
    hr = PathConcat(sczRelativeDirectory, sczFileName, &sczLocalName);
    ExitOnFailure(hr, "Failed to allocate local name for resource.");

    if (SOURCE_TYPE_EMBEDDED == sourceType)
    {
        hr = BurnCreateEmbeddedPayload(guidBundleId, sczBundleFileName, dwPayloadsOffset, sczEmbeddedId, sczLocalName, fTemporary, ppPayload);
    }
    else if (SOURCE_TYPE_EXTERNAL == sourceType)
    {
        hr = PathGetDirectory(sczBundleFileName, &sczSourceDirectory);
        ExitOnFailure(hr, "Failed to get source directory for external payload.");

        hr = PathConcat(sczSourceDirectory, sczFileName, &sczFileName);
        ExitOnFailure2(hr, "Failed to combine source directory and file name: %ls, %ls", sczSourceDirectory, sczFileName);

        hr = BurnCreatePayload(sourceType, guidBundleId, sczFileName, sczLocalName, cbSize, NULL, 0, fTemporary, ppPayload);
    }
    else
    {
        // TODO: handle download payloads
        hr = S_FALSE;
    }
    ExitOnFailure(hr, "Failed to create payload object.");

LExit:
    ReleaseBSTR(bstrNodeValue);
    ReleaseBSTR(bstrNodeName);
    ReleaseObject(pixnAttribute);
    ReleaseObject(pixnnmAttributes);
    ReleaseStr(sczFileName);
    ReleaseStr(sczLocalName);
    ReleaseStr(sczDownloadSource);
    ReleaseStr(sczRelativeDirectory);
    ReleaseStr(sczSourceDirectory);
    ReleaseStr(sczEmbeddedId);

    return hr;
}


/********************************************************************
 ParseResources - parses child Resource elements, plus the parent
    element (package/UX), into an array of payloads.
*********************************************************************/
static HRESULT ParseResources(
    __in REFGUID guidBundleId,
    __in LPCWSTR sczBundleFileName,
    __in IXMLDOMNode* pixn,
    __in DWORD dwPayloadsOffset,
    __in BOOL fTemporary,
    __out DWORD* pcPayloads,
    __out IBurnPayload*** prgpPayloads
    )
{
    HRESULT hr = S_OK;
    IXMLDOMNodeList* pNodeList = NULL;
    IXMLDOMNode* pNode = NULL;
    BSTR bstrNodeName = NULL;
    IBurnPayload** rgpNewPayloads = NULL;
    DWORD cPayloads = 0;
    DeclareConstBSTR(bstrResource, L"Resource");

    hr = pixn->selectNodes(UseConstBSTR(bstrResource), &pNodeList);
    ExitOnFailure(hr, "Failed to select all Resource children.");

    hr = pNodeList->get_length(reinterpret_cast<long*>(&cPayloads));
    ExitOnFailure(hr, "Failed to get the Resource count.");

    // plus one for the parent [package,UX dll] itself
    ++cPayloads;

    rgpNewPayloads = static_cast<IBurnPayload**>(MemAlloc(sizeof(IBurnPayload*) * cPayloads, TRUE));
    ExitOnNull(rgpNewPayloads, hr, E_OUTOFMEMORY, "Failed to allocate memory for payloads.");

    // first, create the first payload object from the parent's payload attributes
    hr = ParseManifestPayload(guidBundleId, sczBundleFileName, pixn, dwPayloadsOffset, fTemporary, rgpNewPayloads);
    ExitOnFailure(hr, "Failed to parse Resource payload from manifest.");

    // now create payload objects for the Resources
    DWORD dwCurrentPayload = 1;
    while (S_OK == (hr = XmlNextElement(pNodeList, &pNode, &bstrNodeName)))
    {
        hr = ParseManifestPayload(guidBundleId, sczBundleFileName, pNode, dwPayloadsOffset, fTemporary, rgpNewPayloads + dwCurrentPayload);
        ExitOnFailure(hr, "Failed to parse package Resource from manifest.");

        ++dwCurrentPayload;
        ReleaseNullBSTR(bstrNodeName);
        ReleaseNullObject(pNode);
    }

    if (S_FALSE == hr)
    {
        hr = S_OK;
    }

    *pcPayloads = cPayloads;
    *prgpPayloads = rgpNewPayloads;
    rgpNewPayloads = NULL;

LExit:
    ReleaseBSTR(bstrNodeName);
    ReleaseObject(pNode);
    ReleaseObject(pNodeList);
    ReleaseObjectArray(rgpNewPayloads, cPayloads);

    return hr;
}


/********************************************************************
 ParseManifestLog - parses a Log element for Chain and MsiPackage
    parent elements.
*********************************************************************/
static HRESULT ParseManifestLog(
    __in IXMLDOMNode* pixn,
    __out BURN_LOGGING_LEVEL* pLogLevel,
    __out BURN_LOGGING_MODE* pLogMode,
    __out_z LPWSTR* psczLogPath
    )
{
    Assert(pixn && pLogLevel && pLogMode && psczLogPath);

    HRESULT hr = S_OK;
    IXMLDOMNamedNodeMap* pixnnmAttributes = NULL;
    IXMLDOMNode* pixnAttribute = NULL;
    BSTR bstrNodeName = NULL;
    BSTR bstrNodeValue = NULL;
    BURN_LOGGING_LEVEL logLevel = BURN_LOGGING_LEVEL_UNKNOWN;
    BURN_LOGGING_MODE logMode = BURN_LOGGING_MODE_UNKNOWN;
    LPWSTR sczLogPath = NULL;

    hr = pixn->get_attributes(&pixnnmAttributes);
    ExitOnFailure(hr, "Failed to get attributes on Log element.");

    while (S_OK == (hr = pixnnmAttributes->nextNode(&pixnAttribute)))
    {
        // get name and value for each attribute
        hr = pixnAttribute->get_nodeName(&bstrNodeName);
        ExitOnFailure(hr, "Failed to get Log attribute name.");

        hr = XmlGetText(pixnAttribute, &bstrNodeValue);
        ExitOnFailure1(hr, "Failed to get Log attribute value: %S", bstrNodeName);

        if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Level", -1))
        {
            if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"normal", -1))
            {
                logLevel = BURN_LOGGING_LEVEL_NORMAL;
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"verbose", -1))
            {
                logLevel = BURN_LOGGING_LEVEL_VERBOSE;
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"debug", -1))
            {
                logLevel = BURN_LOGGING_LEVEL_DEBUG;
            }
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Mode", -1))
        {
            if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"append", -1))
            {
                logMode = BURN_LOGGING_MODE_APPEND;
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"overwrite", -1))
            {
                logMode = BURN_LOGGING_MODE_OVERWRITE;
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"unique", -1))
            {
                logMode = BURN_LOGGING_MODE_UNIQUE;
            }
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Path", -1))
        {
            hr = StrAllocString(&sczLogPath, bstrNodeValue, 0);
            ExitOnFailure(hr, "Failed to allocate @Path.");
        }
        else
        {
            ExitOnFailure1(hr = E_UNEXPECTED, "Log element unknown attribute: %S", bstrNodeName);
        }

        ReleaseNullBSTR(bstrNodeValue);
        ReleaseNullBSTR(bstrNodeName);
    }

    if (S_FALSE == hr)
    {
        hr = S_OK;
    }
    ExitOnFailure(hr, "Failed to enumerate all Log attributes.");

    // defaults
    if (BURN_LOGGING_LEVEL_UNKNOWN == logLevel)
    {
        logLevel = BURN_LOGGING_LEVEL_NORMAL;
    }

    if (BURN_LOGGING_MODE_UNKNOWN == logMode)
    {
        logMode = BURN_LOGGING_MODE_UNIQUE;
    }

    *pLogLevel = logLevel;
    *pLogMode = logMode;
    *psczLogPath = sczLogPath;
    sczLogPath = NULL;

LExit:
    ReleaseBSTR(bstrNodeValue);
    ReleaseBSTR(bstrNodeName);
    ReleaseObject(pixnAttribute);
    ReleaseObject(pixnnmAttributes);
    ReleaseStr(sczLogPath);

    return hr;
}


/********************************************************************
 ParseManifestDocument - parses out a manifest from a loaded XML DOM 
    document.
*********************************************************************/
static HRESULT ParseManifestDocument(
    __in REFGUID guidBundleId,
    __in LPCWSTR sczBundleFileName,
    __in IXMLDOMDocument* pixd,
    __in DWORD dwPayloadsOffset,
    __in IBurnModel* pModel,
    __out DWORD* pcPackages,
    __out IBurnPackage*** prgpPackages,
    __out DWORD* pcUXPayloads,
    __out IBurnPayload*** prgpPayloads,
    __in BURN_SEARCHES* pSearches
    )
{
    Assert(pixd);
    Assert(prgpPackages);
    Assert(pcPackages);
    Assert(prgpPayloads);
    Assert(pcUXPayloads);

    HRESULT hr = S_OK;
    IXMLDOMElement* pManifestElement = NULL;
    IXMLDOMNodeList* pNodes = NULL;
    IXMLDOMNode* pNode = NULL;
    BSTR bstrNodeName = NULL;

    IBurnPackage** rgpNewPackages = NULL;
    IBurnPayload** rgpNewUXPayloads = NULL;

    BURN_LOGGING_LEVEL logLevel = BURN_LOGGING_LEVEL_UNKNOWN;
    BURN_LOGGING_MODE logMode = BURN_LOGGING_MODE_UNKNOWN;
    LPWSTR sczLogPath = NULL;

    //
    // Get the document element and start processing packages.
    //
    hr = pixd->get_documentElement(&pManifestElement);
    ExitOnFailure(hr, "Failed to get BurnManifest document element.");

    hr = pManifestElement->get_childNodes(&pNodes);
    ExitOnFailure(hr, "Failed to get child nodes of manifest document element.");

    while (S_OK == (hr = XmlNextElement(pNodes, &pNode, &bstrNodeName)))
    {
        if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Log", -1))
        {
            hr = ParseManifestLog(pNode, &logLevel, &logMode, &sczLogPath);
            ExitOnFailure(hr, "Failed to parse Log from manifest chain.");

            hr = pModel->InitializeLog(logLevel, logMode, sczLogPath);
            ExitOnFailure(hr, "Failed to initialize model log.");
            if (S_FALSE == hr)
            {
                Trace(REPORT_STANDARD, "Didn't initialize log from manifest, because a log file was specified on the command line.");
            }
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"UX", -1))
        {
            hr = ParseResources(guidBundleId, sczBundleFileName, pNode, dwPayloadsOffset, TRUE, pcUXPayloads, &rgpNewUXPayloads);
            ExitOnFailure(hr, "Failed to parse manifest UX.");
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Chain", -1))
        {
            hr = ParseManifestChain(guidBundleId, sczBundleFileName, pNode, dwPayloadsOffset, pModel, pcPackages, &rgpNewPackages);
            ExitOnFailure(hr, "Failed to parse manifest chain.");
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"DirectorySearch", -1))
        {
            hr = ParseManifestDirectorySearch(pNode, pSearches);
            ExitOnFailure(hr, "Failed to parse directory search.");
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"FileSearch", -1))
        {
            hr = ParseManifestFileSearch(pNode, pSearches);
            ExitOnFailure(hr, "Failed to parse file search.");
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"RegistrySearch", -1))
        {
            hr = ParseManifestRegistrySearch(pNode, pSearches);
            ExitOnFailure(hr, "Failed to parse registry search.");
        }

        ReleaseNullBSTR(bstrNodeName);
        ReleaseNullObject(pNode);
    }

    if (S_FALSE == hr)
    {
        hr = S_OK;
    }

    *prgpPackages = rgpNewPackages;
    rgpNewPackages = NULL;
    *prgpPayloads = rgpNewUXPayloads;
    rgpNewUXPayloads = NULL;

LExit:
    ReleaseBSTR(bstrNodeName);
    ReleaseObject(pNode);
    ReleaseObject(pNodes);
    ReleaseObject(pManifestElement);
    ReleaseObjectArray(rgpNewPackages, *pcPackages);
    ReleaseObjectArray(rgpNewUXPayloads, *pcUXPayloads);
    ReleaseStr(sczLogPath);

    return hr;
}


/********************************************************************
 ParseManifestChain - parses out a chain of packages from a loaded 
    manifest XML DOM element.
*********************************************************************/
static HRESULT ParseManifestChain(
    __in REFGUID guidBundleId,
    __in LPCWSTR sczBundleFileName,
    __in IXMLDOMNode* pixnChain,
    __in DWORD dwPayloadsOffset,
    __in IBurnModel* pModel,
    __out DWORD* pcPackages,
    __out IBurnPackage*** prgpPackages
    )
{
    Assert(pixnChain);
    Assert(pcPackages);
    Assert(prgpPackages);

    HRESULT hr = S_OK;
    IXMLDOMNodeList* pNodeList = NULL;
    IXMLDOMNode* pNode = NULL;
    BSTR bstrNodeName = NULL;
    IBurnPackage** rgpNewPackages = NULL;
    long cPackages = 0;
    DWORD dwCurrentPackage = 0;

    //
    // First, calculate how many items in the chain and allocate
    // the SETUP_PACKAGES structure.
    //
    hr = pixnChain->get_childNodes(&pNodeList);
    ExitOnFailure(hr, "Failed to select all packages in the manifest Chain.");

    hr = pNodeList->get_length(&cPackages);
    ExitOnFailure(hr, "Failed to count the number of packages in the manifest Chain.");

    rgpNewPackages = static_cast<IBurnPackage**>(MemAlloc(sizeof(IBurnPackage*) * cPackages, TRUE));
    ExitOnNull(rgpNewPackages, hr, E_OUTOFMEMORY, "Failed to allocate package list.");

    //
    // Process the packages in the chain.
    //
    while (S_OK == (hr = XmlNextElement(pNodeList, &pNode, &bstrNodeName)))
    {
        if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"MsiPackage", -1))
        {
            hr = ParseManifestMsiPackage(guidBundleId, sczBundleFileName, pNode, dwPayloadsOffset, pModel, reinterpret_cast<IBurnMsiPackage**>(rgpNewPackages + dwCurrentPackage));
            ExitOnFailure(hr, "Failed to parse MsiPackage from manifest chain.");

            ++dwCurrentPackage;
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"ExePackage", -1))
        {
            hr = ParseManifestExePackage(guidBundleId, sczBundleFileName, pNode, dwPayloadsOffset, pModel, reinterpret_cast<IBurnExePackage**>(rgpNewPackages + dwCurrentPackage));
            ExitOnFailure(hr, "Failed to parse ExePackage from manifest chain.");

            ++dwCurrentPackage;
        }
        else
        {
            ExitOnFailure1(hr = E_UNEXPECTED, "Chain element unknown: %S", bstrNodeName);
        }

        ReleaseNullBSTR(bstrNodeName);
        ReleaseNullObject(pNode);
    }

    *pcPackages = cPackages;
    *prgpPackages = rgpNewPackages;
    rgpNewPackages = NULL;

LExit:
    ReleaseBSTR(bstrNodeName);
    ReleaseObject(pNode);
    ReleaseObject(pNodeList);
    ReleaseObjectArray(rgpNewPackages, *pcPackages);

    return hr;
}


/********************************************************************
 ParseCommonPackageAttributes - parses attributes common to multiple
    types of packages
*********************************************************************/
static HRESULT ParseCommonPackageAttributes(
    __in BSTR bstrNodeName,
    __in BSTR bstrNodeValue,
    __out LPWSTR* psczId,
    __out BOOL* pfPermanent,
    __out BOOL* pfVital,
    __out BOOL* pfPerMachine
    )
{
    HRESULT hr = S_OK;

    if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Id", -1))
    {
        hr = StrAllocString(psczId, bstrNodeValue, 0);
        ExitOnFailure(hr, "Failed to allocate @Id.");
    }
    else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"PerMachine", -1))
    {
        *pfPerMachine = CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"yes", -1);
    }
    else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Permanent", -1))
    {
        *pfPermanent = CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"yes", -1);
    }
    else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Vital", -1))
    {
        *pfVital = CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"yes", -1);
    }
    else
    {
        // not a common attribute
        hr = S_FALSE;
    }

LExit:
    return hr;
}


/********************************************************************
 ParseManifestSetProperty - parses a SetProperty element to support
    package-level properties with values from chain-level properties
    (e.g., to get command-line [chain-level] properties from the
    command line to the packages.
*********************************************************************/
static HRESULT ParseManifestSetProperty(
    __in IXMLDOMNode* pNode,
    __out_z LPWSTR* sczSetPropertyId,
    __out_z LPWSTR* sczSetPropertyGlobal
    )
{
    HRESULT hr = S_OK;
    IXMLDOMNamedNodeMap* pixnnmAttributes = NULL;
    IXMLDOMNode* pixnAttribute = NULL;
    BSTR bstrNodeName = NULL;
    BSTR bstrNodeValue = NULL;
    LPWSTR sczId = NULL;
    LPWSTR sczGlobal = NULL;

    hr = pNode->get_attributes(&pixnnmAttributes);
    ExitOnFailure(hr, "Failed to get attributes on SetProperty element.");

    while (S_OK == (hr = pixnnmAttributes->nextNode(&pixnAttribute)))
    {
        // get name and value for each attribute
        hr = pixnAttribute->get_nodeName(&bstrNodeName);
        ExitOnFailure(hr, "Failed to get SetProperty  attribute name.");

        hr = XmlGetText(pixnAttribute, &bstrNodeValue);
        ExitOnFailure1(hr, "Failed to get SetProperty  attribute value: %S", bstrNodeName);

        if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Id", -1))
        {
            hr = StrAllocString(&sczId, bstrNodeValue, 0);
            ExitOnFailure(hr, "Failed to allocate @Id.");
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Global", -1))
        {
            hr = StrAllocString(&sczGlobal, bstrNodeValue, 0);
            ExitOnFailure(hr, "Failed to allocate @Global.");
        }

        ReleaseNullBSTR(bstrNodeValue);
        ReleaseNullBSTR(bstrNodeName);
        ReleaseNullObject(pixnAttribute);
    }

    if (S_FALSE == hr)
    {
        hr = S_OK;
    }
    ExitOnFailure(hr, "Failed to enumerate all SetProperty attributes.");

    ExitOnNull(sczId, hr, E_INVALIDARG, "SetProperty requires Id attribute value.");
    ExitOnNull(sczGlobal, hr, E_INVALIDARG, "SetProperty requires Global attribute value.");

    *sczSetPropertyId = sczId;
    sczId = NULL;

    *sczSetPropertyGlobal = sczGlobal;
    sczGlobal = NULL;

LExit:
    ReleaseBSTR(bstrNodeValue);
    ReleaseBSTR(bstrNodeName);
    ReleaseObject(pixnAttribute);
    ReleaseObject(pixnnmAttributes);
    ReleaseStr(sczId);
    ReleaseStr(sczGlobal);

    return hr;
}


/********************************************************************
 ParseManifestMsiPackage - parses an MsiPackage node's attributes and
    child elements into an IBurnMsiPackage.
*********************************************************************/
static HRESULT ParseManifestMsiPackage(
    __in REFGUID guidBundleId,
    __in LPCWSTR sczBundleFileName,
    __in IXMLDOMNode* pixnMsiPackage,
    __in DWORD dwPayloadsOffset,
    __in IBurnModel* pModel,
    __out IBurnMsiPackage** ppPackage
    )
{
    HRESULT hr = S_OK;
    IXMLDOMNamedNodeMap* pixnnmAttributes = NULL;
    IXMLDOMNode* pixnAttribute = NULL;
    IXMLDOMNodeList* pNodeList = NULL;
    IXMLDOMNode* pNode = NULL;
    BSTR bstrNodeName = NULL;
    BSTR bstrNodeValue = NULL;
    LPWSTR sczId = NULL;
    LPWSTR sczProductCode = NULL;
    LPWSTR sczGlobalValue = NULL;
    IBurnProperties* pGlobalProperties = NULL;
    IBurnPayload** rgpPayloads = NULL;

    DWORD cPayloads = 1;
    BOOL fPermanent = FALSE;
    BOOL fVital = FALSE;
    BOOL fPerMachine = FALSE;
    ULARGE_INTEGER uliProductVersion = { 0 };

    BURN_LOGGING_LEVEL logLevel = BURN_LOGGING_LEVEL_UNKNOWN;
    BURN_LOGGING_MODE logMode = BURN_LOGGING_MODE_UNKNOWN;
    LPWSTR sczLogPath = NULL;
    LPWSTR sczChainLogPath = NULL;

    LPWSTR sczSetPropertyId = NULL;
    LPWSTR sczSetPropertyGlobal = NULL;
    IBurnProperties* pProperties = NULL;

    // now process the rest of the package's attributes.
    hr = pixnMsiPackage->get_attributes(&pixnnmAttributes);
    ExitOnFailure(hr, "Failed to get attributes on MsiPackage element.");

    while (S_OK == (hr = pixnnmAttributes->nextNode(&pixnAttribute)))
    {
        // get name and value for each attribute
        hr = pixnAttribute->get_nodeName(&bstrNodeName);
        ExitOnFailure(hr, "Failed to get MsiPackage attribute name.");

        hr = XmlGetText(pixnAttribute, &bstrNodeValue);
        ExitOnFailure1(hr, "Failed to get MsiPackage attribute value: %S", bstrNodeName);

        // try common attributes; if that's a no-go, try MsiPackage-specific attributes
        hr = ParseCommonPackageAttributes(bstrNodeName, bstrNodeValue, &sczId, &fPermanent, &fVital, &fPerMachine);
        ExitOnFailure2(hr, "Failed to parse common package attribute: %S=%S", bstrNodeName, bstrNodeValue);

        if (S_FALSE == hr)
        {
            if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"ProductCode", -1))
            {
                hr = StrAllocString(&sczProductCode, bstrNodeValue, 0);
                ExitOnFailure1(hr, "Failed to convert invalid GUID '%S'", bstrNodeValue);
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"ProductVersion", -1))
            {
                hr = FileVersionFromString(bstrNodeValue, &uliProductVersion.HighPart, &uliProductVersion.LowPart);
                ExitOnFailure1(hr, "Failed to parse MsiPackage product version. '%S'", bstrNodeValue);
            }
        } 

        ReleaseNullBSTR(bstrNodeValue);
        ReleaseNullBSTR(bstrNodeName);
        ReleaseNullObject(pixnAttribute);
    }

    if (S_FALSE == hr)
    {
        hr = S_OK;
    }
    ExitOnFailure(hr, "Failed to enumerate all attributes on MsiPackage element.");

    // first get all the package's Resource children (e.g., external cabinets)
    hr = ParseResources(guidBundleId, sczBundleFileName, pixnMsiPackage, dwPayloadsOffset, FALSE, &cPayloads, &rgpPayloads);
    ExitOnFailure(hr, "Failed to parse MsiPackage/Resource payloads.");

    // now process other child elements
    hr = pixnMsiPackage->get_childNodes(&pNodeList);
    ExitOnFailure(hr, "Failed to select all children of the MsiPackage.");

    while (S_OK == (hr = XmlNextElement(pNodeList, &pNode, &bstrNodeName)))
    {
        if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Log", -1))
        {
            // get log request from manifest and save it for when we create the package object
            hr = ParseManifestLog(pNode, &logLevel, &logMode, &sczLogPath);
            ExitOnFailure(hr, "Failed to parse Log from manifest chain.");
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"SetProperty", -1))
        {
            hr = ParseManifestSetProperty(pNode, &sczSetPropertyId, &sczSetPropertyGlobal);
            ExitOnFailure(hr, "Failed to parse SetProperty from manifest.");

            // get value of specified global property
            hr = pModel->GetProperties(&pGlobalProperties);
            ExitOnFailure(hr, "Failed to get global properties collection.");

            hr = pGlobalProperties->GetStringValueByName(sczSetPropertyGlobal, &sczGlobalValue);
            // a missing global property isn't fatal (mimicking MSI behavior)
            if (E_INVALIDARG != hr)
            {
                ExitOnFailure1(hr, "Failed to get global property: %S.", sczSetPropertyGlobal);

                // don't create an IBurnProperties for the package until it's needed
                if (!pProperties)
                {
                    hr = BurnCreateProperties(&pProperties);
                    ExitOnFailure(hr, "Failed to create empty properties collection.");
                }

                hr = pProperties->SetString(sczSetPropertyId, sczGlobalValue);
                ExitOnFailure(hr, "Failed to add to properties collection.");
            }
        }

        ReleaseNullBSTR(bstrNodeValue);
        ReleaseNullBSTR(bstrNodeName);
        ReleaseNullStr(sczSetPropertyId);
        ReleaseNullStr(sczSetPropertyGlobal);
        ReleaseNullStr(sczGlobalValue);
        ReleaseNullObject(pGlobalProperties);
        ReleaseNullObject(pNode);
    }

    // TODO: sanity check
    ExitOnNull(sczProductCode, hr, E_INVALIDARG, "MsiPackage/@ProductCode invalid or not specified.");

    // we have payload(s), we have data, now create the package object
    hr = BurnCreateMsiPackage(sczId, sczProductCode, uliProductVersion.QuadPart, fPerMachine, cPayloads, rgpPayloads, pProperties, ppPackage);
    ExitOnFailure(hr, "Failed to create IBurnMsiPackage object.");

    // set the package's logging, with defaults based off the chain log if needed.
    BURN_LOGGING_LEVEL chainLogLevel;
    hr = pModel->GetLogPath(&sczChainLogPath, &chainLogLevel);
    ExitOnFailure(hr, "Failed to retrieve chain's log.");

    hr = (*ppPackage)->InitializeLog(BURN_LOGGING_LEVEL_UNKNOWN == logLevel ? chainLogLevel : logLevel, logMode, sczChainLogPath, sczLogPath);
    ExitOnFailure(hr, "Failed to initialize MSI package log.");

    rgpPayloads = NULL;
    pProperties = NULL;

LExit:
    ReleaseStr(sczId);
    ReleaseStr(sczProductCode);
    ReleaseStr(sczSetPropertyId);
    ReleaseStr(sczSetPropertyGlobal);
    ReleaseStr(sczGlobalValue);
    ReleaseStr(sczLogPath);
    ReleaseStr(sczChainLogPath);
    ReleaseBSTR(bstrNodeValue);
    ReleaseBSTR(bstrNodeName);
    ReleaseObject(pNode);
    ReleaseObject(pNodeList);
    ReleaseObject(pixnAttribute);
    ReleaseObject(pixnnmAttributes);
    ReleaseObject(pProperties);
    ReleaseObject(pGlobalProperties);
    ReleaseObjectArray(rgpPayloads, cPayloads);

    return hr;
}


/********************************************************************
 ParseManifestExePackage - parses an ExePackage node's attributes and
    child elements.
*********************************************************************/
static HRESULT ParseManifestExePackage(
    __in REFGUID guidBundleId,
    __in LPCWSTR sczBundleFileName,
    __in IXMLDOMNode* pixnExePackage,
    __in DWORD dwPayloadsOffset,
    __in IBurnModel* pModel,
    __out IBurnExePackage** ppPackage
    )
{
    HRESULT hr = S_OK;
    IXMLDOMNamedNodeMap* pixnnmAttributes = NULL;
    IXMLDOMNode* pixnAttribute = NULL;
    IXMLDOMNodeList* pNodeList = NULL;
    IXMLDOMNode* pNode = NULL;
    BSTR bstrNodeName = NULL;
    BSTR bstrNodeValue = NULL;
    LPWSTR sczId = NULL;
    LPWSTR sczDetectCondition = NULL;
    LPWSTR sczInstallArgs = NULL;
    LPWSTR sczRepairArgs = NULL;
    LPWSTR sczUninstallArgs = NULL;
    LPWSTR sczGlobalValue = NULL;
    IBurnProperties* pGlobalProperties = NULL;
    IBurnPayload** rgpPayloads = NULL;
    DeclareConstBSTR(bstrExitInfo, L"ExitInfo");

    DWORD cExitInfos = 0;
    BURN_EXE_PACKAGE_EXIT_INFO* prgExitInfo = NULL;

    DWORD cPayloads = 1;
    BOOL fPermanent = FALSE;
    BOOL fVital = FALSE;
    BOOL fPerMachine = FALSE;
    ULARGE_INTEGER uliProductVersion = { 0 };

    BURN_LOGGING_LEVEL logLevel = BURN_LOGGING_LEVEL_UNKNOWN;
    BURN_LOGGING_MODE logMode = BURN_LOGGING_MODE_UNKNOWN;
    LPWSTR sczLogPath = NULL;
    LPWSTR sczChainLogPath = NULL;

    LPWSTR sczSetPropertyId = NULL;
    LPWSTR sczSetPropertyGlobal = NULL;
    IBurnProperties* pProperties = NULL;

    // now process the rest of the package's attributes.
    hr = pixnExePackage->get_attributes(&pixnnmAttributes);
    ExitOnFailure(hr, "Failed to get attributes on ExePackage element.");

    while (S_OK == (hr = pixnnmAttributes->nextNode(&pixnAttribute)))
    {
        // get name and value for each attribute
        hr = pixnAttribute->get_nodeName(&bstrNodeName);
        ExitOnFailure(hr, "Failed to get ExePackage attribute name.");

        hr = XmlGetText(pixnAttribute, &bstrNodeValue);
        ExitOnFailure1(hr, "Failed to get ExePackage attribute value: %S", bstrNodeName);

        // try common attributes; if that's a no-go, try ExePackage-specific attributes
        hr = ParseCommonPackageAttributes(bstrNodeName, bstrNodeValue, &sczId, &fPermanent, &fVital, &fPerMachine);
        ExitOnFailure2(hr, "Failed to parse common package attribute: %S=%S", bstrNodeName, bstrNodeValue);

        if (S_FALSE == hr)
        {
            if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"InstallArguments", -1))
            {
                hr = StrAllocString(&sczInstallArgs, bstrNodeValue, 0);
                ExitOnFailure(hr, "Failed to allocate @InstallArguments.");
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"ProductVersion", -1))
            {
                hr = FileVersionFromString(bstrNodeValue, &uliProductVersion.HighPart, &uliProductVersion.LowPart);
                ExitOnFailure1(hr, "Failed to parse ExePackage product version. '%S'", bstrNodeValue);
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"RepairArguments", -1))
            {
                hr = StrAllocString(&sczRepairArgs, bstrNodeValue, 0);
                ExitOnFailure(hr, "Failed to allocate @RepairArguments.");
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"UninstallArguments", -1))
            {
                hr = StrAllocString(&sczUninstallArgs, bstrNodeValue, 0);
                ExitOnFailure(hr, "Failed to allocate @UninstallArguments.");
            }
        }

        ReleaseNullBSTR(bstrNodeValue);
        ReleaseNullBSTR(bstrNodeName);
        ReleaseNullObject(pixnAttribute);
    }

    if (S_FALSE == hr)
    {
        hr = S_OK;
    }
    ExitOnFailure(hr, "Failed to enumerate all attributes on ExePackage element.");

    // first get all the package's Resource children (e.g., external cabinets)
    hr = ParseResources(guidBundleId, sczBundleFileName, pixnExePackage, dwPayloadsOffset, FALSE, &cPayloads, &rgpPayloads);
    ExitOnFailure(hr, "Failed to parse ExePackage/Resource payloads.");

    hr = pixnExePackage->selectNodes(UseConstBSTR(bstrExitInfo), &pNodeList);
    ExitOnFailure(hr, "Failed to select all Resource children.");

    hr = pNodeList->get_length(reinterpret_cast<long*>(&cExitInfos));
    ExitOnFailure(hr, "Failed to get the ExitInfo count.");

    if (cExitInfos)
    {
        prgExitInfo = static_cast<BURN_EXE_PACKAGE_EXIT_INFO*>(MemAlloc(cExitInfos * sizeof(BURN_EXE_PACKAGE_EXIT_INFO), TRUE));
        ExitOnNull(prgExitInfo, hr, E_OUTOFMEMORY, "Failed to allocate ExitInfo array.");
    }

    // now process other child elements
    hr = pixnExePackage->get_childNodes(&pNodeList);
    ExitOnFailure(hr, "Failed to select all children of the ExePackage.");

    DWORD dwNextExitInfo = 0;
    while (S_OK == (hr = XmlNextElement(pNodeList, &pNode, &bstrNodeName)))
    {
        if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"DetectionCondition", -1))
        {
            hr = XmlGetText(pNode, &bstrNodeValue);
            ExitOnFailure(hr, "Failed to parse @DetectionCondition from manifest chain.");

            hr = StrAllocString(&sczDetectCondition, bstrNodeValue, 0);
            ExitOnFailure(hr, "Failed to copy detection condition.");
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"ExitInfo", -1))
        {
            // get log request from manifest and save it for when we create the package object
            hr = ParseManifestExeExitInfo(pNode, prgExitInfo + dwNextExitInfo);
            ExitOnFailure(hr, "Failed to parse Log from manifest chain.");

            ++dwNextExitInfo;
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Log", -1))
        {
            // get log request from manifest and save it for when we create the package object
            hr = ParseManifestLog(pNode, &logLevel, &logMode, &sczLogPath);
            ExitOnFailure(hr, "Failed to parse Log from manifest chain.");
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"SetProperty", -1))
        {
            hr = ParseManifestSetProperty(pNode, &sczSetPropertyId, &sczSetPropertyGlobal);
            ExitOnFailure(hr, "Failed to parse SetProperty from manifest.");

            // get value of specified global property
            hr = pModel->GetProperties(&pGlobalProperties);
            ExitOnFailure(hr, "Failed to get global properties collection.");

            hr = pGlobalProperties->GetStringValueByName(sczSetPropertyGlobal, &sczGlobalValue);
            // a missing global property isn't fatal (mimicking EXE behavior)
            if (E_INVALIDARG != hr)
            {
                ExitOnFailure1(hr, "Failed to get global property: %S.", sczSetPropertyGlobal);

                // don't create an IBurnProperties for the package until it's needed
                if (!pProperties)
                {
                    hr = BurnCreateProperties(&pProperties);
                    ExitOnFailure(hr, "Failed to create empty properties collection.");
                }

                hr = pProperties->SetString(sczSetPropertyId, sczGlobalValue);
                ExitOnFailure(hr, "Failed to add to properties collection.");
            }
        }

        ReleaseNullBSTR(bstrNodeValue);
        ReleaseNullBSTR(bstrNodeName);
        ReleaseNullStr(sczSetPropertyId);
        ReleaseNullStr(sczSetPropertyGlobal);
        ReleaseNullStr(sczGlobalValue);
        ReleaseNullObject(pGlobalProperties);
        ReleaseNullObject(pNode);
    }

    // TODO: sanity check
    ExitOnNull(sczInstallArgs, hr, E_INVALIDARG, "ExePackage/@InstallArguments invalid or not specified.");

    // we have payload(s), we have data, now create the package object
    // TODO: figure out what to pass as the ProductCode since passing the "Id" is not the right thing to do ultimately.
    hr = BurnCreateExePackage(sczId, sczId, uliProductVersion.QuadPart, fPerMachine, sczDetectCondition, sczInstallArgs, sczRepairArgs, sczUninstallArgs, cExitInfos, prgExitInfo, cPayloads, rgpPayloads, pProperties, ppPackage);
    ExitOnFailure(hr, "Failed to create IBurnExePackage object.");

    prgExitInfo = NULL;
    rgpPayloads = NULL;

    // set the package's logging, with defaults based off the chain log if needed.
    BURN_LOGGING_LEVEL chainLogLevel;
    hr = pModel->GetLogPath(&sczChainLogPath, &chainLogLevel);
    ExitOnFailure(hr, "Failed to retrieve chain's log.");

    hr = (*ppPackage)->InitializeLog(BURN_LOGGING_LEVEL_UNKNOWN == logLevel ? chainLogLevel : logLevel, logMode, sczChainLogPath, sczLogPath);
    ExitOnFailure(hr, "Failed to initialize EXE package log.");
    pProperties = NULL;

LExit:
    ReleaseStr(sczId);
    ReleaseStr(sczUninstallArgs);
    ReleaseStr(sczRepairArgs);
    ReleaseStr(sczInstallArgs);
    ReleaseStr(sczDetectCondition);
    ReleaseStr(sczSetPropertyId);
    ReleaseStr(sczSetPropertyGlobal);
    ReleaseStr(sczGlobalValue);
    ReleaseStr(sczLogPath);
    ReleaseStr(sczChainLogPath);
    ReleaseBSTR(bstrNodeValue);
    ReleaseBSTR(bstrNodeName);
    ReleaseObject(pNode);
    ReleaseObject(pNodeList);
    ReleaseObject(pixnAttribute);
    ReleaseObject(pixnnmAttributes);
    ReleaseObject(pProperties);
    ReleaseObject(pGlobalProperties);
    ReleaseObjectArray(rgpPayloads, cPayloads);
    ReleaseMem(prgExitInfo);

    return hr;
}


static HRESULT ParseManifestExeExitInfo(
    __in IXMLDOMNode* pixnExitInfoNode,
    __in BURN_EXE_PACKAGE_EXIT_INFO* pExitInfo
    )
{
    HRESULT hr = S_OK;
    IXMLDOMNamedNodeMap* pixnnmAttributes = NULL;
    IXMLDOMNode* pixnAttribute = NULL;
    BSTR bstrNodeName = NULL;
    BSTR bstrNodeValue = NULL;

    hr = pixnExitInfoNode->get_attributes(&pixnnmAttributes);
    ExitOnFailure(hr, "Failed to get attributes on ExitInfo element.");

    while (S_OK == (hr = pixnnmAttributes->nextNode(&pixnAttribute)))
    {
        // get name and value for each attribute
        hr = pixnAttribute->get_nodeName(&bstrNodeName);
        ExitOnFailure(hr, "Failed to get ExitInfo attribute name.");

        hr = XmlGetText(pixnAttribute, &bstrNodeValue);
        ExitOnFailure1(hr, "Failed to get ExitInfo attribute value: %S", bstrNodeName);

        if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Type", -1))
        {
            if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"success", -1))
            {
                pExitInfo->type = BURN_EXE_EXIT_CODE_TYPE_SUCCESS;
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"error", -1))
            {
                pExitInfo->type = BURN_EXE_EXIT_CODE_TYPE_ERROR;
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"reboot", -1))
            {
                pExitInfo->type = BURN_EXE_EXIT_CODE_TYPE_SCHEDULE_REBOOT;
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"forceReboot", -1))
            {
                pExitInfo->type = BURN_EXE_EXIT_CODE_TYPE_FORCE_REBOOT;
            }
            else
            {
                hr = E_INVALIDARG;
                ExitOnFailure1(hr, "Invalid ExitInfo/@Type: %S", bstrNodeValue);
            }
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"ExitCode", -1))
        {
            // If we have a wildcard ignore the exit code.
            if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"*", -1))
            {
                pExitInfo->dwCode = 0;
                pExitInfo->fWildcard = TRUE;
            }
            else // must be an actual exit code number.
            {
                pExitInfo->dwCode = wcstoul(bstrNodeValue, NULL, 10);
                pExitInfo->fWildcard = FALSE;
            }
        }

        ReleaseNullBSTR(bstrNodeValue);
        ReleaseNullBSTR(bstrNodeName);
        ReleaseNullObject(pixnAttribute);
    }

    if (S_FALSE == hr)
    {
        hr = S_OK;
    }
    ExitOnFailure(hr, "Failed to enumerate all attributes on ExePackage element.");

LExit:
    ReleaseBSTR(bstrNodeValue);
    ReleaseBSTR(bstrNodeName);
    ReleaseObject(pixnAttribute);

    return hr;
}

static HRESULT ParseManifestDirectorySearch(
    __in IXMLDOMNode* pixnDirectorySearch,
    __in BURN_SEARCHES* pSearches
    )
{
    HRESULT hr = S_OK;
    IXMLDOMNamedNodeMap* pixnnmAttributes = NULL;
    IXMLDOMNode* pixnAttribute = NULL;
    IXMLDOMNodeList* pNodeList = NULL;
    IXMLDOMNode* pNode = NULL;
    BSTR bstrNodeName = NULL;
    BSTR bstrNodeValue = NULL;
    BURN_SEARCH search = { 0 };
    BURN_SEARCH* pNewSearch = NULL;

    // process attributes
    hr = pixnDirectorySearch->get_attributes(&pixnnmAttributes);
    ExitOnFailure(hr, "Failed to get attributes.");

    while (S_OK == (hr = pixnnmAttributes->nextNode(&pixnAttribute)))
    {
        // get name and value for each attribute
        hr = pixnAttribute->get_nodeName(&bstrNodeName);
        ExitOnFailure(hr, "Failed to get attribute name.");

        hr = XmlGetText(pixnAttribute, &bstrNodeValue);
        ExitOnFailure1(hr, "Failed to get attribute value: %S", bstrNodeName);

        // try common attributes
        hr = ParseCommonSearchAttributes(bstrNodeName, bstrNodeValue, &search);
        ExitOnFailure2(hr, "Failed to parse common search attribute: %S=%S", bstrNodeName, bstrNodeValue);

        if (S_FALSE == hr)
        {
            if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Path", -1))
            {
                hr = StrAllocString(&search.DirectorySearch.sczPath, bstrNodeValue, 0);
                ExitOnFailure(hr, "Failed to copy @Path.");
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Type", -1))
            {
                if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"exists", -1))
                {
                    search.DirectorySearch.Type = BURN_DIRECTORY_SEARCH_TYPE_EXISTS;
                }
                else
                {
                    ExitOnFailure1(hr = E_INVALIDARG, "Invalid value for @Type: %S", bstrNodeValue);
                }
            }
            else
            {
                ExitOnFailure1(hr = E_INVALIDARG, "Unexpected attribure name: %S", bstrNodeName);
            }
        }

        ReleaseNullBSTR(bstrNodeValue);
        ReleaseNullBSTR(bstrNodeName);
        ReleaseNullObject(pixnAttribute);
    }
    ExitOnFailure(hr, "Failed to enumerate all attributes on DirectorySearch element.");

    // process child elements
    hr = pixnDirectorySearch->get_childNodes(&pNodeList);
    ExitOnFailure(hr, "Failed to get child elements.");

    while (S_OK == (hr = XmlNextElement(pNodeList, &pNode, &bstrNodeName)))
    {
        // try common child elements
        hr = ParseCommonSearchChildElements(bstrNodeName, pNode, &search);
        ExitOnFailure1(hr, "Failed to parse common search child element: %S", bstrNodeName);

        if (S_FALSE == hr)
        {
            ExitOnFailure1(hr = E_INVALIDARG, "Unexpected child element name: %S", bstrNodeName);
        }

        ReleaseNullBSTR(bstrNodeName);
        ReleaseNullObject(pNode);
    }

    // TODO: sanity check
    ExitOnNull(search.sczKey, hr, E_INVALIDARG, "DirectorySearch/@Id invalid or not specified.");

    // add search
    hr = SearchAdd(pSearches, &pNewSearch);
    ExitOnFailure(hr, "Failed to add search.");

    ::memcpy_s(pNewSearch, sizeof(BURN_SEARCH), &search, sizeof(BURN_SEARCH));
    ::memset(&search, 0, sizeof(BURN_SEARCH));

LExit:
    ReleaseBSTR(bstrNodeValue);
    ReleaseBSTR(bstrNodeName);
    ReleaseObject(pNode);
    ReleaseObject(pNodeList);
    ReleaseObject(pixnAttribute);
    ReleaseObject(pixnnmAttributes);
    SearchFree(&search);

    return hr;
}

static HRESULT ParseManifestFileSearch(
    __in IXMLDOMNode* pixnFileSearch,
    __in BURN_SEARCHES* pSearches
    )
{
    HRESULT hr = S_OK;
    IXMLDOMNamedNodeMap* pixnnmAttributes = NULL;
    IXMLDOMNode* pixnAttribute = NULL;
    IXMLDOMNodeList* pNodeList = NULL;
    IXMLDOMNode* pNode = NULL;
    BSTR bstrNodeName = NULL;
    BSTR bstrNodeValue = NULL;
    BURN_SEARCH search = { 0 };
    BURN_SEARCH* pNewSearch = NULL;

    // process attributes
    hr = pixnFileSearch->get_attributes(&pixnnmAttributes);
    ExitOnFailure(hr, "Failed to get attributes.");

    while (S_OK == (hr = pixnnmAttributes->nextNode(&pixnAttribute)))
    {
        // get name and value for each attribute
        hr = pixnAttribute->get_nodeName(&bstrNodeName);
        ExitOnFailure(hr, "Failed to get attribute name.");

        hr = XmlGetText(pixnAttribute, &bstrNodeValue);
        ExitOnFailure1(hr, "Failed to get attribute value: %S", bstrNodeName);

        // try common attributes
        hr = ParseCommonSearchAttributes(bstrNodeName, bstrNodeValue, &search);
        ExitOnFailure2(hr, "Failed to parse common search attribute: %S=%S", bstrNodeName, bstrNodeValue);

        if (S_FALSE == hr)
        {
            if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Path", -1))
            {
                hr = StrAllocString(&search.FileSearch.sczPath, bstrNodeValue, 0);
                ExitOnFailure(hr, "Failed to copy @Path.");
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Type", -1))
            {
                if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"exists", -1))
                {
                    search.FileSearch.Type = BURN_FILE_SEARCH_TYPE_EXISTS;
                }
                else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"version", -1))
                {
                    search.FileSearch.Type = BURN_FILE_SEARCH_TYPE_VERSION;
                }
                else
                {
                    ExitOnFailure1(hr = E_INVALIDARG, "Invalid value for @Type: %S", bstrNodeValue);
                }
            }
            else
            {
                ExitOnFailure1(hr = E_INVALIDARG, "Unexpected attribure name: %S", bstrNodeName);
            }
        }

        ReleaseNullBSTR(bstrNodeValue);
        ReleaseNullBSTR(bstrNodeName);
        ReleaseNullObject(pixnAttribute);
    }
    ExitOnFailure(hr, "Failed to enumerate all attributes on DirectorySearch element.");

    // process child elements
    hr = pixnFileSearch->get_childNodes(&pNodeList);
    ExitOnFailure(hr, "Failed to get child elements.");

    while (S_OK == (hr = XmlNextElement(pNodeList, &pNode, &bstrNodeName)))
    {
        // try common child elements
        hr = ParseCommonSearchChildElements(bstrNodeName, pNode, &search);
        ExitOnFailure1(hr, "Failed to parse common search child element: %S", bstrNodeName);

        if (S_FALSE == hr)
        {
            ExitOnFailure1(hr = E_INVALIDARG, "Unexpected child element name: %S", bstrNodeName);
        }

        ReleaseNullBSTR(bstrNodeName);
        ReleaseNullObject(pNode);
    }

    // TODO: sanity check
    ExitOnNull(search.sczKey, hr, E_INVALIDARG, "FileSearch/@Id invalid or not specified.");

    // add search
    hr = SearchAdd(pSearches, &pNewSearch);
    ExitOnFailure(hr, "Failed to add search.");

    ::CopyMemory(pNewSearch, &search, sizeof(BURN_SEARCH));
    ::ZeroMemory(&search, sizeof(BURN_SEARCH));

LExit:
    ReleaseBSTR(bstrNodeValue);
    ReleaseBSTR(bstrNodeName);
    ReleaseObject(pNode);
    ReleaseObject(pNodeList);
    ReleaseObject(pixnAttribute);
    ReleaseObject(pixnnmAttributes);
    SearchFree(&search);

    return hr;
}

static HRESULT ParseManifestRegistrySearch(
    __in IXMLDOMNode* pixnRegistrySearch,
    __in BURN_SEARCHES* pSearches
    )
{
    HRESULT hr = S_OK;
    IXMLDOMNamedNodeMap* pixnnmAttributes = NULL;
    IXMLDOMNode* pixnAttribute = NULL;
    IXMLDOMNodeList* pNodeList = NULL;
    IXMLDOMNode* pNode = NULL;
    BSTR bstrNodeName = NULL;
    BSTR bstrNodeValue = NULL;
    BURN_SEARCH search = { 0 };
    BURN_SEARCH* pNewSearch = NULL;

    // process attributes
    hr = pixnRegistrySearch->get_attributes(&pixnnmAttributes);
    ExitOnFailure(hr, "Failed to get attributes.");

    while (S_OK == (hr = pixnnmAttributes->nextNode(&pixnAttribute)))
    {
        // get name and value for each attribute
        hr = pixnAttribute->get_nodeName(&bstrNodeName);
        ExitOnFailure(hr, "Failed to get attribute name.");

        hr = XmlGetText(pixnAttribute, &bstrNodeValue);
        ExitOnFailure1(hr, "Failed to get attribute value: %S", bstrNodeName);

        // try common attributes
        hr = ParseCommonSearchAttributes(bstrNodeName, bstrNodeValue, &search);
        ExitOnFailure2(hr, "Failed to parse common search attribute: %S=%S", bstrNodeName, bstrNodeValue);

        if (S_FALSE == hr)
        {
            if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"ExpandEnvironment", -1))
            {
                search.RegistrySearch.fExpandEnvironment = CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"yes", -1);
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Key", -1))
            {
                hr = StrAllocString(&search.RegistrySearch.sczKey, bstrNodeValue, 0);
                ExitOnFailure(hr, "Failed to copy @Key.");
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Root", -1))
            {
                if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"HKCU", -1))
                {
                    search.RegistrySearch.hRoot = HKEY_CURRENT_USER;
                }
                else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"HKLM", -1))
                {
                    search.RegistrySearch.hRoot = HKEY_LOCAL_MACHINE;
                }
                else
                {
                    ExitOnFailure1(hr = E_INVALIDARG, "Invalid value for @Root: %S", bstrNodeValue);
                }
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Value", -1))
            {
                hr = StrAllocString(&search.RegistrySearch.sczValue, bstrNodeValue, 0);
                ExitOnFailure(hr, "Failed to copy @Value.");
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Type", -1))
            {
                if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"exists", -1))
                {
                    search.RegistrySearch.Type = BURN_REGISTRY_SEARCH_TYPE_EXISTS;
                }
                else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"value", -1))
                {
                    search.RegistrySearch.Type = BURN_REGISTRY_SEARCH_TYPE_VALUE;
                }
                else
                {
                    ExitOnFailure1(hr = E_INVALIDARG, "Invalid value for @Type: %S", bstrNodeValue);
                }
            }
            else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"ValueType", -1))
            {
                if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"numeric", -1))
                {
                    search.RegistrySearch.ValueType = BURN_VARIANT_TYPE_NUMERIC;
                }
                else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"string", -1))
                {
                    search.RegistrySearch.ValueType = BURN_VARIANT_TYPE_STRING;
                }
                else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"version", -1))
                {
                    search.RegistrySearch.ValueType = BURN_VARIANT_TYPE_VERSION;
                }
                else
                {
                    ExitOnFailure1(hr = E_INVALIDARG, "Invalid value for @ValueType: %S", bstrNodeValue);
                }
            }
            else
            {
                ExitOnFailure1(hr = E_INVALIDARG, "Unexpected attribure name: %S", bstrNodeName);
            }
        }

        ReleaseNullBSTR(bstrNodeValue);
        ReleaseNullBSTR(bstrNodeName);
        ReleaseNullObject(pixnAttribute);
    }
    ExitOnFailure(hr, "Failed to enumerate all attributes on DirectorySearch element.");

    // process child elements
    hr = pixnRegistrySearch->get_childNodes(&pNodeList);
    ExitOnFailure(hr, "Failed to select all children of the MsiPackage.");

    while (S_OK == (hr = XmlNextElement(pNodeList, &pNode, &bstrNodeName)))
    {
        // try common child elements
        hr = ParseCommonSearchChildElements(bstrNodeName, pNode, &search);
        ExitOnFailure1(hr, "Failed to parse common search child element: %S", bstrNodeName);

        if (S_FALSE == hr)
        {
            ExitOnFailure1(hr = E_INVALIDARG, "Unexpected child element name: %S", bstrNodeName);
        }

        ReleaseNullBSTR(bstrNodeName);
        ReleaseNullObject(pNode);
    }

    // TODO: sanity check
    ExitOnNull(search.sczKey, hr, E_INVALIDARG, "RegistrySearch/@Id invalid or not specified.");

    // add search
    hr = SearchAdd(pSearches, &pNewSearch);
    ExitOnFailure(hr, "Failed to add search.");

    ::CopyMemory(pNewSearch, &search, sizeof(BURN_SEARCH));
    ::ZeroMemory(&search, sizeof(BURN_SEARCH));

LExit:
    ReleaseBSTR(bstrNodeValue);
    ReleaseBSTR(bstrNodeName);
    ReleaseObject(pNode);
    ReleaseObject(pNodeList);
    ReleaseObject(pixnAttribute);
    ReleaseObject(pixnnmAttributes);
    SearchFree(&search);

    return hr;
}

static HRESULT ParseCommonSearchAttributes(
    __in BSTR bstrNodeName,
    __in BSTR bstrNodeValue,
    __in BURN_SEARCH* pSearch
    )
{
    HRESULT hr = S_OK;

    if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Id", -1))
    {
        hr = StrAllocString(&pSearch->sczKey, bstrNodeValue, 0);
        ExitOnFailure(hr, "Failed to copy @Id.");
    }
    else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Property", -1))
    {
        hr = StrAllocString(&pSearch->sczProperty, bstrNodeValue, 0);
        ExitOnFailure(hr, "Failed to copy @Property.");
    }
    else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Condition", -1))
    {
        hr = StrAllocString(&pSearch->sczCondition, bstrNodeValue, 0);
        ExitOnFailure(hr, "Failed to copy @Condition.");
    }
    else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Scheduling", -1))
    {
        if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"detect", -1))
        {
            pSearch->Scheduling = BURN_SEARCH_SCHEDULING_DETECT;
        }
        else if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeValue, -1, L"prepare", -1))
        {
            pSearch->Scheduling = BURN_SEARCH_SCHEDULING_PREPARE;
        }
        else
        {
            ExitOnFailure1(hr = E_INVALIDARG, "Invalid value for @Scheduling: %S", bstrNodeValue);
        }
    }
    else
    {
        // not a common attribute
        hr = S_FALSE;
    }

LExit:
    return hr;
}

static HRESULT ParseCommonSearchChildElements(
    __in BSTR bstrNodeName,
    __in IXMLDOMNode* pNode,
    __in BURN_SEARCH* pSearch
    )
{
    HRESULT hr = S_OK;

    if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"SearchDependency", -1))
    {
        hr = ParseManifestSearchDependency(pNode, pSearch);
        ExitOnFailure(hr, "Failed to parse SearchDependency element.");
    }
    else
    {
        // not a common child element
        hr = S_FALSE;
    }

LExit:
    return hr;
}

static HRESULT ParseManifestSearchDependency(
    __in IXMLDOMNode* pixnSearchDependency,
    __in BURN_SEARCH* pSearch
    )
{
    HRESULT hr = S_OK;
    IXMLDOMNamedNodeMap* pixnnmAttributes = NULL;
    IXMLDOMNode* pixnAttribute = NULL;
    IXMLDOMNodeList* pNodeList = NULL;
    IXMLDOMNode* pNode = NULL;
    BSTR bstrNodeName = NULL;
    BSTR bstrNodeValue = NULL;
    LPWSTR sczDependency = NULL;
    LPWSTR* psczDependency = NULL;

    // process attributes
    hr = pixnSearchDependency->get_attributes(&pixnnmAttributes);
    ExitOnFailure(hr, "Failed to get attributes.");

    while (S_OK == (hr = pixnnmAttributes->nextNode(&pixnAttribute)))
    {
        // get name and value for each attribute
        hr = pixnAttribute->get_nodeName(&bstrNodeName);
        ExitOnFailure(hr, "Failed to get attribute name.");

        hr = XmlGetText(pixnAttribute, &bstrNodeValue);
        ExitOnFailure1(hr, "Failed to get attribute value: %S", bstrNodeName);

        if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, 0, bstrNodeName, -1, L"Id", -1))
        {
            hr = StrAllocString(&sczDependency, bstrNodeValue, 0);
            ExitOnFailure(hr, "Failed to copy @Id.");
        }
        else
        {
            ExitOnFailure1(hr = E_INVALIDARG, "Unexpected attribure name: %S", bstrNodeName);
        }

        ReleaseNullBSTR(bstrNodeValue);
        ReleaseNullBSTR(bstrNodeName);
        ReleaseNullObject(pixnAttribute);
    }
    ExitOnFailure(hr, "Failed to enumerate all attributes on DirectorySearch element.");

    // TODO: sanity check
    ExitOnNull(sczDependency, hr, E_INVALIDARG, "SearchDependency/@Id invalid or not specified.");

    // add search
    hr = SearchAddDependency(pSearch, &psczDependency);
    ExitOnFailure(hr, "Failed to add search dependency.");

    *psczDependency = sczDependency;
    sczDependency = NULL;

LExit:
    ReleaseBSTR(bstrNodeValue);
    ReleaseBSTR(bstrNodeName);
    ReleaseObject(pNode);
    ReleaseObject(pNodeList);
    ReleaseObject(pixnAttribute);
    ReleaseObject(pixnnmAttributes);
    ReleaseStr(sczDependency);

    return hr;
}
