//-------------------------------------------------------------------------------------------------
// <copyright file="msiwork.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//
// <summary>
// MSI Package operations for setup stub executable.
// </summary>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"


// structs

struct INSTALL_CONTEXT
{
    SETUP_ENGINE_INFORMATION* pEngineInfo;
    DWORD dwPackageIndex;
    ACTION_STATE Action;
};


// prototypes

static INT CALLBACK InstallEngineCallback(
    __in LPVOID pvContext,
    __in UINT uiMessage,
    __in LPCWSTR wzMessage
    );
static INT CALLBACK InstallEngineRecordCallback(
    __in LPVOID pvContext,
    __in UINT uiMessage,
    __in MSIHANDLE hRecord
    );
static void HandleInstallMessage(
    __in INSTALL_CONTEXT* pInstallContext,
    __in INSTALLMESSAGE mt,
    __in UINT uiFlags,
    __in LPCWSTR wzMessage,
    __in_opt MSIHANDLE hRecord,
    __out INT* piResult
    );
static void HandleInstallMessageProgress(
    __in INSTALL_CONTEXT* pInstallContext,
    __in_opt LPCWSTR wzMessage,
    __in_opt MSIHANDLE hRecord,
    __out INT* piResult
    );
static HRESULT SendProgressUpdate(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo
    );


// function definitions

HRESULT ExtractMsi(
    __in SETUP_INSTALL_PACKAGE* pPackage,
    __out LPWSTR* ppwzTempMsiPath
    )
{
    HRESULT hr = S_OK;
    DWORD cchTempMsiPath = MAX_PATH;

    // Drop the package into the temp directory.
    hr = StrAlloc(ppwzTempMsiPath, cchTempMsiPath);
    ExitOnFailure(hr, "Failed to allocate string for temp path.");

    hr = DirCreateTempPath(L"CT", *ppwzTempMsiPath, cchTempMsiPath);
    ExitOnFailure(hr, "Failed to generate a temporary filename.");

    hr = ResExportDataToFile(pPackage->szSource, *ppwzTempMsiPath, CREATE_ALWAYS);
    ExitOnFailure1(hr, "Failed to write MSI package to %S.", *ppwzTempMsiPath);

LExit:
    return hr;
}


HRESULT DetectMsi(
    __in REFGUID guidProductCode,
    __in DWORD64 qwVersion,
    __in BOOL fRepair,
    __in PACKAGE_STATE requested,
    __in PACKAGE_STATE expected,
    __out PACKAGE_STATE* pCurrent,
    __out ACTION_STATE* pExecuteAction,
    __out ACTION_STATE* pRollbackAction
    )
{
    Assert(pCurrent && pExecuteAction && pRollbackAction);

    HRESULT hr = S_OK;
    UINT er = ERROR_SUCCESS;
    DWORD dwMajorVersion = static_cast<DWORD>(qwVersion >> 32);
    DWORD dwMinorVersion = static_cast<DWORD>(qwVersion & 0xFFFFFFFF);
    WCHAR wzInstalledVersion[36] = { 0 };
    DWORD cchInstalledVersion = countof(wzInstalledVersion);
    DWORD dwInstalledMajorVersion = 0;
    DWORD dwInstalledMinorVersion = 0;
    WCHAR wzProductCode[39] = { 0 };

    if (0 == ::StringFromGUID2(guidProductCode, wzProductCode, countof(wzProductCode)))
    {
        hr = HRESULT_FROM_WIN32(ERROR_INSUFFICIENT_BUFFER);
        ExitOnFailure(hr, "Failed to convert ProductCode GUID to string.");
    }

    er = HOOKABLE_MsiGetProductInfoW(wzProductCode, L"VersionString", wzInstalledVersion, &cchInstalledVersion);
    if (ERROR_SUCCESS == er)
    {
        // package pressent
        *pCurrent = PACKAGE_STATE_PRESENT; // well, we know that some version of this package is present.

        hr = FileVersionFromString(wzInstalledVersion, &dwInstalledMajorVersion, &dwInstalledMinorVersion);
        ExitOnFailure2(hr, "Failed to convert version: %S to DWORDs for ProductCode: %S", wzInstalledVersion, wzProductCode);

        if (PACKAGE_STATE_PRESENT == requested)
        {
            //
            // Take a look at the version and determine if this is a potential
            // minor upgrade, a major upgrade (just install) otherwise, there
            // is a newer version so no work necessary.
            //
            // minor upgrade "10.1.X.X" > "10.0.X.X" or "10.3.2.0" > "10.3.1.0"
            if (((dwMajorVersion >> 16) == (dwInstalledMajorVersion >> 16) && dwMajorVersion > dwInstalledMajorVersion) || 
                (dwMajorVersion == dwInstalledMajorVersion && dwMinorVersion > dwInstalledMinorVersion))
            {
                *pExecuteAction = ACTION_STATE_MINOR_UPGRADE;
            }
            else if (dwMajorVersion > dwInstalledMajorVersion) // major upgrade "11.X.X.X" > "10.X.X.X"
            {
                *pExecuteAction = ACTION_STATE_MAJOR_UPGRADE;
            }
            else if (dwMajorVersion == dwInstalledMajorVersion) // maintenance install "10.X.X.X" = "10.X.X.X"
            {
                *pExecuteAction = fRepair ? ACTION_STATE_RECACHE : ACTION_STATE_MAINTENANCE;
            }
            else // newer version present "14.X.X.X" < "15.X.X.X", skip
            {
                *pExecuteAction = ACTION_STATE_NONE; // TODO: inform about this state to enable warnings
            }
        }
        else if (PACKAGE_STATE_ABSENT == requested)
        {
            if (dwMajorVersion == dwInstalledMajorVersion) // install "10.X.X.X" = "10.X.X.X"
            {
                *pExecuteAction = ACTION_STATE_UNINSTALL;
            }
            else
            {
                *pExecuteAction = ACTION_STATE_NONE; // TODO: inform about this state to enable warnings
            }
        }

        hr = S_OK;
    }
    else if (ERROR_UNKNOWN_PRODUCT == er || ERROR_UNKNOWN_PROPERTY == er)
    {
        // package not pressent
        *pCurrent = PACKAGE_STATE_ABSENT;

        *pExecuteAction = PACKAGE_STATE_PRESENT == requested ? ACTION_STATE_INSTALL : ACTION_STATE_NONE;

        hr = S_OK;
    }
    else
    {
        hr = HRESULT_FROM_WIN32(er);
        ExitOnFailure1(hr, "Failed to get product information for ProductCode: %S", wzProductCode);
    }

    // rollback action
    switch (PACKAGE_STATE_UNKNOWN != expected ? expected : *pCurrent)
    {
    case PACKAGE_STATE_PRESENT:
        switch (requested)
        {
        case PACKAGE_STATE_PRESENT:
            *pRollbackAction = fRepair ? ACTION_STATE_RECACHE : ACTION_STATE_MAINTENANCE;
            break;
        case PACKAGE_STATE_ABSENT:
            *pRollbackAction = ACTION_STATE_INSTALL;
            break;
        default:
            ExitFunction1(hr = E_UNEXPECTED);
        }
        break;
    case PACKAGE_STATE_ABSENT:
        switch (requested)
        {
        case PACKAGE_STATE_PRESENT:
            *pRollbackAction = ACTION_STATE_UNINSTALL;
            break;
        case PACKAGE_STATE_ABSENT:
            *pRollbackAction = ACTION_STATE_NONE;
            break;
        default:
            ExitFunction1(hr = E_UNEXPECTED);
        }
        break;
    default:
        ExitFunction1(hr = E_UNEXPECTED);
    }

LExit:
    return hr;
}


HRESULT DetectMsiFeature(
    __in REFGUID guidProductCode,
    __in PACKAGE_STATE packageState,
    __in LPCWSTR wzFeature,
    __in BOOL fRepair,
    __in FEATURE_STATE requested,
    __in FEATURE_STATE expected,
    __out FEATURE_STATE* pCurrent,
    __out FEATURE_ACTION* pExecuteAction,
    __out FEATURE_ACTION* pRollbackAction
    )
{
    HRESULT hr = S_OK;
    INSTALLSTATE state = INSTALLSTATE_UNKNOWN;
    WCHAR wzProductCode[39] = { 0 };

    // convert product code to string
    if (0 == ::StringFromGUID2(guidProductCode, wzProductCode, countof(wzProductCode)))
    {
        hr = HRESULT_FROM_WIN32(ERROR_INSUFFICIENT_BUFFER);
        ExitOnFailure(hr, "Failed to convert ProductCode GUID to string.");
    }

    // get current state
    if (PACKAGE_STATE_PRESENT == packageState) // only try to detect features if the product is installed
    {
        state = HOOKABLE_MsiQueryFeatureStateW(wzProductCode, wzFeature);
        if (INSTALLSTATE_INVALIDARG == state)
        {
            ExitOnFailure(hr = E_FAIL, "Failed to query feature state.");
        }
        if (INSTALLSTATE_UNKNOWN == state) // in case of an upgrade this could happen
        {
            state = INSTALLSTATE_ABSENT;
        }
    }
    else
    {
        state = INSTALLSTATE_ABSENT;
    }

    // set current state
    switch (state)
    {
    case INSTALLSTATE_ABSENT:     *pCurrent = FEATURE_STATE_ABSENT;     break;
    case INSTALLSTATE_ADVERTISED: *pCurrent = FEATURE_STATE_ADVERTISED; break;
    case INSTALLSTATE_LOCAL:      *pCurrent = FEATURE_STATE_LOCAL;      break;
    case INSTALLSTATE_SOURCE:     *pCurrent = FEATURE_STATE_SOURCE;     break;
    default:
        ExitFunction1(hr = E_UNEXPECTED);
    } 

    // set action state
    *pExecuteAction = FEATURE_ACTION_NONE;
    switch (requested)
    {
    case FEATURE_STATE_ABSENT:
        if (INSTALLSTATE_ABSENT != state)
        {
            *pExecuteAction = FEATURE_ACTION_REMOVE;
        }
        break;
    case FEATURE_STATE_ADVERTISED:
        if (INSTALLSTATE_ADVERTISED != state)
        {
            *pExecuteAction = FEATURE_ACTION_ADVERTISE;
        }
        else if (fRepair)
        {
            *pExecuteAction = FEATURE_ACTION_REINSTALL;
        }
        break;
    case FEATURE_STATE_LOCAL:
        if (INSTALLSTATE_LOCAL != state)
        {
            *pExecuteAction = FEATURE_ACTION_ADDLOCAL;
        }
        else if (fRepair)
        {
            *pExecuteAction = FEATURE_ACTION_REINSTALL;
        }
        break;
    case FEATURE_STATE_SOURCE:
        if (INSTALLSTATE_SOURCE != state)
        {
            *pExecuteAction = FEATURE_ACTION_ADDSOURCE;
        }
        else if (fRepair)
        {
            *pExecuteAction = FEATURE_ACTION_REINSTALL;
        }
        break;
    default:
        ExitFunction1(hr = E_UNEXPECTED);
    }

    // rollback action
    if (FEATURE_ACTION_REINSTALL == *pExecuteAction)
    {
        *pRollbackAction = FEATURE_ACTION_REINSTALL;
    }
    else if (FEATURE_ACTION_NONE != *pExecuteAction)
    {
        // if we have an expected state, rollback to the expected state else rollback to actuall state
        switch (FEATURE_STATE_UNKNOWN != expected ? expected : *pCurrent)
        {
        case FEATURE_STATE_ABSENT:     *pRollbackAction = FEATURE_ACTION_REMOVE;    break;
        case FEATURE_STATE_ADVERTISED: *pRollbackAction = FEATURE_ACTION_ADVERTISE; break;
        case FEATURE_STATE_LOCAL:      *pRollbackAction = FEATURE_ACTION_ADDLOCAL;  break;
        case FEATURE_STATE_SOURCE:     *pRollbackAction = FEATURE_ACTION_ADDSOURCE; break;
        default:
            ExitFunction1(hr = E_UNEXPECTED);
        }
    }
    else
    {
        *pRollbackAction = FEATURE_ACTION_NONE;
    }

LExit:
    return hr;
}


HRESULT CacheMsi(
    __in SETUP_INSTALL_PACKAGE* pPackage,
    __in LPCWSTR wzTempMsiPath,
    __in BOOL bAllUsers,
    __out LPWSTR* ppwzCachedMsiPath
    )
{
    Assert(ppwzCachedMsiPath);

    HRESULT hr = S_OK;
    DWORD dwAppDataFlags = bAllUsers ? CSIDL_COMMON_APPDATA : CSIDL_LOCAL_APPDATA;
    WCHAR wzLocalAppPath[MAX_PATH];
    WCHAR wzProductCode[39];
    WCHAR wzVersion[36];

    // actually cache the MSI now
    hr = ::SHGetFolderPathW(NULL, dwAppDataFlags | CSIDL_FLAG_CREATE, NULL, SHGFP_TYPE_CURRENT, wzLocalAppPath);
    ExitOnFailure(hr, "Failed to get local app data folder.");

    hr = ::StringFromGUID2(pPackage->guidProductCode, wzProductCode, countof(wzProductCode));
    ExitOnFailure(hr, "Failed to convert ProductCode GUID to string.");

    hr = ::StringCchPrintfW(wzVersion, countof(wzVersion), L"%d.%d.%d.%d", pPackage->dwVersionMajor >> 16, pPackage->dwVersionMajor & 0xFFFF,
                                                                           pPackage->dwVersionMinor >> 16, pPackage->dwVersionMinor & 0xFFFF);
    ExitOnFailure(hr, "Failed to format version string.");

    hr = StrAllocFormatted(ppwzCachedMsiPath, L"%ls\\Applications\\Cache\\%lsv%ls.msi", wzLocalAppPath, wzProductCode, wzVersion);
    ExitOnFailure(hr, "Failed to allocate formatted.");

    hr = FileEnsureMove(wzTempMsiPath, *ppwzCachedMsiPath, TRUE, FALSE);
    ExitOnFailure(hr, "Failed to move MSI to cache.");

LExit:
    return hr;
}


HRESULT ConfigureMsi(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in DWORD dwPackageIndex,
    __in REFGUID guidProductCode,
    __in_opt LPCWSTR wzMsiPath,
    __in LPCWSTR wzMsiInstallArgs,
    __in ACTION_STATE action,
    __in MSI_UI_LEVEL uiLevel,
    __out LPDWORD pdwReturnCode
    )
{
    Trace2(REPORT_STANDARD, "Executing MSI: Path = '%S', Arguments = '%S'", wzMsiPath, wzMsiInstallArgs);

    HRESULT hr = S_OK;
    DWORD er = ERROR_SUCCESS;

    INSTALL_CONTEXT ic = { 0 };
    DWORD dwUiLevel;// = INSTALLUILEVEL_NONE | INSTALLUILEVEL_SOURCERESONLY; //INSTALLUILEVEL_BASIC; // use at least basic UI so the user knows something is going on.
    DWORD dwMessageFilter = 0;
    INSTALLUI_HANDLERW pfnExternalUI = NULL;
    INSTALLUI_HANDLER_RECORD pfnExternalUIRecord = NULL;
    LPWSTR pwzLogFile = NULL;
    LPWSTR pwzProperties = NULL;
    WCHAR wzProductCode[39] = { 0 };

    if (0 == ::StringFromGUID2(guidProductCode, wzProductCode, countof(wzProductCode)))
    {
        hr = HRESULT_FROM_WIN32(ERROR_INSUFFICIENT_BUFFER);
        ExitOnFailure(hr, "Failed to convert ProductCode GUID to string.");
    }

    switch (uiLevel)
    {
    case MSI_UI_LEVEL_DEFAULT:
        dwUiLevel = INSTALLUILEVEL_DEFAULT;
        break;
    case MSI_UI_LEVEL_NONE:
        dwUiLevel = INSTALLUILEVEL_NONE;
        if (pEngineInfo->fEnableSourceResPrompt)
        {
            dwUiLevel |= INSTALLUILEVEL_SOURCERESONLY;
        }
        break;
    case MSI_UI_LEVEL_BASIC:
        dwUiLevel = INSTALLUILEVEL_BASIC;
        break;
    case MSI_UI_LEVEL_REDUCED:
        dwUiLevel = INSTALLUILEVEL_REDUCED;
        break;
    case MSI_UI_LEVEL_FULL:
        dwUiLevel = INSTALLUILEVEL_FULL;
        break;
    default:
        ExitFunction1(hr = E_UNEXPECTED);
    }

    //
    // clear progress information
    //

    pEngineInfo->dwMsiProgressTotal = 0;
    pEngineInfo->dwMsiProgressStep = 0;
    pEngineInfo->dwMsiProgressComplete = 0;
    pEngineInfo->fMsiProgressForward = FALSE;
    pEngineInfo->fMsiProgressScriptInProgress = FALSE;
    pEngineInfo->fMsiProgressEnableActionData = FALSE;
    pEngineInfo->fMsiProgressFinished = FALSE;


    //
    // setup the external UI handler and logging
    //
    //if (fShowUI)
    //{
    //    dwUiLevel = INSTALLUILEVEL_FULL; // use Full UI
    //}
    //else if (fIgnoreFailures)
    //{
    //    dwUiLevel |= INSTALLUILEVEL_PROGRESSONLY; // don't throw any scary error dialogs
    //}

    HOOKABLE_MsiSetInternalUI(static_cast<INSTALLUILEVEL>(dwUiLevel), NULL);

    ZeroMemory(&ic, sizeof(ic));
    ic.pEngineInfo = pEngineInfo;
    ic.dwPackageIndex = dwPackageIndex;
    ic.Action = action;

    dwMessageFilter = INSTALLLOGMODE_PROGRESS | INSTALLLOGMODE_FATALEXIT | INSTALLLOGMODE_ERROR
        | INSTALLLOGMODE_WARNING | INSTALLLOGMODE_USER | INSTALLLOGMODE_INFO
        | INSTALLLOGMODE_RESOLVESOURCE | INSTALLLOGMODE_OUTOFDISKSPACE
        | INSTALLLOGMODE_ACTIONSTART | INSTALLLOGMODE_ACTIONDATA
        | INSTALLLOGMODE_COMMONDATA | INSTALLLOGMODE_PROGRESS | INSTALLLOGMODE_INITIALIZE
        | INSTALLLOGMODE_TERMINATE | INSTALLLOGMODE_SHOWDIALOG;

    if (pEngineInfo->pfnMsiSetExternalUIRecord) // set if MSI version >= 3.1
    {
        er = HOOKABLE_MsiSetExternalUIRecord(InstallEngineRecordCallback, dwMessageFilter, static_cast<LPVOID>(&ic), &pfnExternalUIRecord);
        hr = HRESULT_FROM_WIN32(er);
        ExitOnFailure(hr, "Failed to set external UI callback.");
    }
    else
    {
        pfnExternalUI = HOOKABLE_MsiSetExternalUIW(InstallEngineCallback, dwMessageFilter, static_cast<LPVOID>(&ic));
    }

    hr = PathCreateTempFile(NULL, L"setup.wix.msi.%00000x.log", SHRT_MAX, FILE_ATTRIBUTE_NORMAL, &pwzLogFile, NULL);
    if (SUCCEEDED(hr))
    {
        er = HOOKABLE_MsiEnableLogW(INSTALLLOGMODE_FATALEXIT|INSTALLLOGMODE_ERROR|INSTALLLOGMODE_WARNING
            |INSTALLLOGMODE_USER|INSTALLLOGMODE_INFO|INSTALLLOGMODE_RESOLVESOURCE
            |INSTALLLOGMODE_OUTOFDISKSPACE|INSTALLLOGMODE_ACTIONSTART|INSTALLLOGMODE_ACTIONDATA
            |INSTALLLOGMODE_COMMONDATA|INSTALLLOGMODE_PROPERTYDUMP, // |INSTALLLOGMODE_VERBOSE
            pwzLogFile, 0);
        hr = HRESULT_FROM_WIN32(er);
        if (SUCCEEDED(hr))
        {
            LogStringLine(REPORT_STANDARD, "MSI log file set to: %S", pwzLogFile);
        }
        //TraceOnFailure1(hr, "failed to set logging: %S", pwzLogFile);
    }

    // Ignore all failures setting up the UI handler and 
    // logging, they aren't important enough to abort the
    // install attempt.
    hr = S_OK;

    // set up our properties
    hr = StrAllocString(&pwzProperties, wzMsiInstallArgs ? wzMsiInstallArgs : L"", 0);
    ExitOnFailure1(hr, "Failed to allocate string for MSI install arguments: %S", wzMsiInstallArgs);

    //
    // Do the actual action.
    //
    switch (action)
    {
    case ACTION_STATE_ADMIN_INSTALL: __fallthrough;
        hr = StrAllocConcat(&pwzProperties, L" ACTION=ADMIN", 0);
        ExitOnFailure(hr, "Failed to format property: ADMIN");
    case ACTION_STATE_MAJOR_UPGRADE: __fallthrough;
    case ACTION_STATE_INSTALL:
        er = HOOKABLE_MsiInstallProductW(wzMsiPath, pwzProperties);
        //hr = HRESULT_FROM_WIN32(er);
        break;
    case ACTION_STATE_MAINTENANCE: __fallthrough;
    case ACTION_STATE_RECACHE:
        hr = StrAllocConcat(&pwzProperties, L" REINSTALLMODE=\"vdmus\"", 0);
        ExitOnFailure(hr, "failed to format property: REINSTALLMODE");

        er = HOOKABLE_MsiConfigureProductExW(wzProductCode, INSTALLLEVEL_DEFAULT, INSTALLSTATE_DEFAULT, pwzProperties);
        //hr = HRESULT_FROM_WIN32(er);
        break;
    case ACTION_STATE_MINOR_UPGRADE:
        hr = StrAllocConcat(&pwzProperties, L" REINSTALL=ALL REINSTALLMODE=\"vomus\"", 0);
        ExitOnFailure(hr, "failed to format properties: REINSTALL and REINSTALLMODE");

        er = HOOKABLE_MsiInstallProductW(wzMsiPath, pwzProperties);
        //hr = HRESULT_FROM_WIN32(er);
        break;
    case ACTION_STATE_UNINSTALL:
        er = HOOKABLE_MsiConfigureProductExW(wzProductCode, INSTALLLEVEL_DEFAULT, INSTALLSTATE_ABSENT, pwzProperties);
        //hr = HRESULT_FROM_WIN32(er);
        break;
    }

    *pdwReturnCode = er;

LExit:
    if (pfnExternalUI)  // unset the UI handler
    {
        HOOKABLE_MsiSetExternalUIW(pfnExternalUI, 0, NULL);
    }
    if (pfnExternalUIRecord)  // unset the UI record handler
    {
        HOOKABLE_MsiSetExternalUIRecord(pfnExternalUIRecord, 0, NULL, NULL);
    }

    ReleaseStr(pwzProperties);
    ReleaseStr(pwzLogFile);

    return hr;
}


HRESULT ExtractMst(
    __in SETUP_INSTALL_TRANSFORM* pTransform,
    __out LPWSTR* ppwzTempMstPath
    )
{
    HRESULT hr = S_OK;
    DWORD cchTempMsiPath = MAX_PATH;

    // Drop the package into the temp directory.
    hr = StrAlloc(ppwzTempMstPath, cchTempMsiPath);
    ExitOnFailure(hr, "Failed to allocate string for temp path.");

    hr = DirCreateTempPath(L"MST", *ppwzTempMstPath, cchTempMsiPath);
    ExitOnFailure(hr, "Failed to generate a temporary filename.");

    hr = ResExportDataToFile(pTransform->szTransform, *ppwzTempMstPath, CREATE_ALWAYS);
    ExitOnFailure1(hr, "Failed to write MSI transform to %S.", *ppwzTempMstPath);

LExit:
    return hr;
}


// privat helper function definitions

static INT CALLBACK InstallEngineCallback(
    __in LPVOID pvContext,
    __in UINT uiMessage,
    __in LPCWSTR wzMessage
    )
{
    INT iResult = IDOK;
    INSTALLMESSAGE mt = static_cast<INSTALLMESSAGE>(0xFF000000 & uiMessage);
    UINT uiFlags = 0x00FFFFFF & uiMessage;
    INSTALL_CONTEXT* pInstallContext = static_cast<INSTALL_CONTEXT*>(pvContext);

    // bail if we weren't passed valid context data
    if (!pInstallContext)
    {
        Trace(REPORT_STANDARD, "Windows Installer installation callback not passed valid state");
        ExitFunction();
    }

    if (INSTALLMESSAGE_PROGRESS == mt)
    {
        HandleInstallMessageProgress(pInstallContext, wzMessage, NULL, &iResult);
    }
    else
    {
        HandleInstallMessage(pInstallContext, mt, uiFlags, wzMessage, NULL, &iResult);
    }


LExit:
    return iResult;
}

static INT CALLBACK InstallEngineRecordCallback(
    __in LPVOID pvContext,
    __in UINT uiMessage,
    __in MSIHANDLE hRecord
    )
{
    HRESULT hr = S_OK;
    DWORD er = ERROR_SUCCESS;
    INT iResult = IDOK;
    INSTALLMESSAGE mt = static_cast<INSTALLMESSAGE>(0xFF000000 & uiMessage);
    UINT uiFlags = 0x00FFFFFF & uiMessage;
    INSTALL_CONTEXT* pInstallContext = static_cast<INSTALL_CONTEXT*>(pvContext);
    LPWSTR pwzMessage = NULL;
    DWORD cchMessage = 0;

    // bail if we weren't passed valid context data
    if (!pInstallContext)
    {
        Trace(REPORT_STANDARD, "Windows Installer installation callback not passed valid state");
        ExitFunction();
    }

    if (INSTALLMESSAGE_PROGRESS == mt)
    {
        HandleInstallMessageProgress(pInstallContext, NULL, hRecord, &iResult);
    }
    else
    {
        // create formated message string
        if (hRecord)
        {
            er = ::MsiFormatRecordW(NULL, hRecord, L"", &cchMessage);
            if (ERROR_MORE_DATA == er || ERROR_SUCCESS == er)
            {
                hr = StrAlloc(&pwzMessage, ++cchMessage);
            }
            else
            {
                hr = HRESULT_FROM_WIN32(er);
            }
            ExitOnFailure(hr, "Failed to allocate string for formated message.");

            er = ::MsiFormatRecordW(NULL, hRecord, pwzMessage, &cchMessage);
            hr = HRESULT_FROM_WIN32(er);
            ExitOnFailure(hr, "Failed to format message record.");
        }

        // pass to handler including both the formated message and the original record
        HandleInstallMessage(pInstallContext, mt, uiFlags, pwzMessage, hRecord, &iResult);
    }

LExit:
    ReleaseStr(pwzMessage);
    return iResult;
}

static void HandleInstallMessage(
    __in INSTALL_CONTEXT* pInstallContext,
    __in INSTALLMESSAGE mt,
    __in UINT uiFlags,
    __in LPCWSTR wzMessage,
    __in_opt MSIHANDLE hRecord,
    __out INT* piResult
    )
{
    HRESULT hr = S_OK;
    SETUP_ENGINE_INFORMATION* pEngineInfo = pInstallContext->pEngineInfo;
    DWORD dwPackageIndex = pInstallContext->dwPackageIndex;

    switch (mt)
    {
    case INSTALLMESSAGE_FATALEXIT: // Premature termination
        LogStringLine(REPORT_ERROR, "Windows Installer: FatalExit: %S", wzMessage);
        hr = SendSuxExecMsiMessage(pEngineInfo, SRM_EXEC_MSI_FATAL_ERROR, dwPackageIndex, uiFlags, wzMessage, hRecord, piResult);
        ExitOnFailure(hr, "Failed to send SRM_EXEC_MSI_FATAL_ERROR message.");
        break;

    case INSTALLMESSAGE_ERROR: // Error message to display
        LogStringLine(REPORT_ERROR, "Windows Installer: Error: %S", wzMessage);
        hr = SendSuxExecMsiMessage(pEngineInfo, SRM_EXEC_MSI_ERROR, dwPackageIndex, uiFlags, wzMessage, hRecord, piResult);
        ExitOnFailure(hr, "Failed to send SRM_EXEC_MSI_ERROR message.");
        break;

    case INSTALLMESSAGE_WARNING:
        LogStringLine(REPORT_ERROR, "Windows Installer: Warning: %S", wzMessage);
        hr = SendSuxExecMsiMessage(pEngineInfo, SRM_EXEC_MSI_WARNING, dwPackageIndex, uiFlags, wzMessage, hRecord, piResult);
        ExitOnFailure(hr, "Failed to send SRM_EXEC_MSI_WARNING message.");
        break;

    case INSTALLMESSAGE_USER:
        LogStringLine(REPORT_STANDARD, "Windows Installer: User: %S", wzMessage);
        hr = SendSuxExecMsiMessage(pEngineInfo, SRM_EXEC_MSI_USER, dwPackageIndex, uiFlags, wzMessage, hRecord, piResult);
        ExitOnFailure(hr, "Failed to send SRM_EXEC_MSI_USER message.");
        break;

    case INSTALLMESSAGE_INFO:
        LogStringLine(REPORT_VERBOSE, "Windows Installer: Info: %S", wzMessage);
        hr = SendSuxExecMsiMessage(pEngineInfo, SRM_EXEC_MSI_INFO, dwPackageIndex, uiFlags, wzMessage, hRecord, piResult);
        ExitOnFailure(hr, "Failed to send SRM_EXEC_MSI_INFO message.");
        break;

    case INSTALLMESSAGE_ACTIONSTART:
        hr = SendSuxExecMsiMessage(pEngineInfo, SRM_EXEC_MSI_ACTION_START, dwPackageIndex, uiFlags, wzMessage, hRecord, piResult);
        ExitOnFailure(hr, "Failed to send SRM_EXEC_MSI_ACTION_START message.");
        break;

    case INSTALLMESSAGE_ACTIONDATA:
        hr = SendSuxExecMsiMessage(pEngineInfo, SRM_EXEC_MSI_ACTION_DATA, dwPackageIndex, uiFlags, wzMessage, hRecord, piResult);
        ExitOnFailure(hr, "Failed to send SRM_EXEC_MSI_ACTION_DATA message.");
        break;

#if 0
    case INSTALLMESSAGE_COMMONDATA:
        if (L'1' == wzMessage[0] && L':' == wzMessage[1] && L' ' == wzMessage[2])
        {
            if (L'0' == wzMessage[3])
            {
                // TODO: handle the language common data message.
                lres = IDOK;
                return lres;
            }
            else if (L'1' == wzMessage[3])
            {
                // TODO: really handle sending the caption.
                lres = ::SendSuxMessage(pInstallContext->pSetupUXInformation, SRM_EXEC_SET_CAPTION, uiFlags, reinterpret_cast<LPARAM>(wzMessage + 3));
                return lres;
            }
            else if (L'2' == wzMessage[3])
            {
                // TODO: really handle sending the cancel button status.
                lres = ::SendSuxMessage(pInstallContext->pSetupUXInformation, SRM_EXEC_SET_CANCEL, uiFlags, reinterpret_cast<LPARAM>(wzMessage + 3));
                return lres;
            }
        }
        break;

/*
    case INSTALLMESSAGE_RMFILESINUSE:
        lres = ::SendSuxMessage(pInstallContext->pSetupUXInformation, SRM_EXEC_RESTART_MANAGER_FILE_IN_USE, uiFlags, wzMessage);
        return lres;
*/

    case INSTALLMESSAGE_FILESINUSE:
        lres = ::SendSuxMessage(pInstallContext->pSetupUXInformation, SRM_EXEC_FILE_IN_USE, uiFlags, reinterpret_cast<LPARAM>(wzMessage));
        break;

    case INSTALLMESSAGE_OUTOFDISKSPACE:
        lres = ::SendSuxMessage(pInstallContext->pSetupUXInformation, SRM_EXEC_OUT_OF_DISK_SPACE, uiFlags, reinterpret_cast<LPARAM>(wzMessage));
        break;

    case INSTALLMESSAGE_INITIALIZE: // this message is received prior to internal UI initialization, no string data
        lres = ::SendSuxMessage(pInstallContext->pSetupUXInformation, SRM_EXEC_INITIALIZE, 0, 0);
        break;

    case INSTALLMESSAGE_TERMINATE: // sent after UI termination, no string data
        lres = ::SendSuxMessage(pInstallContext->pSetupUXInformation, SRM_EXEC_TERMINATE, 0, 0);
        break;

    case INSTALLMESSAGE_SHOWDIALOG: // sent prior to display of authored dialog or wizard
        lres = ::SendSuxMessage(pInstallContext->pSetupUXInformation, SRM_EXEC_SHOW_DIALOG, 0, 0);
        break;

    case INSTALLMESSAGE_RESOLVESOURCE:
        lres = ::SendSuxMessage(pInstallContext->pSetupUXInformation, SRM_EXEC_RESOLVE_SOURCE, 0, 0);
        lres = 0; // always return 0 for resolve source.
        break;
#endif

    default:
        Trace3(REPORT_VERBOSE, "InstallEngineCallback() - mt: %d, uiFlags: %d, wzMessage: %S", mt, uiFlags, wzMessage);
        break;
    }

LExit:
    return;
}

static void HandleInstallMessageProgress(
    __in INSTALL_CONTEXT* pInstallContext,
    __in_opt LPCWSTR wzMessage,
    __in_opt MSIHANDLE hRecord,
    __out INT* piResult
    )
{
    HRESULT hr = S_OK;
    SETUP_ENGINE_INFORMATION* pEngineInfo = pInstallContext->pEngineInfo;
    int iFields[4] = { 0 };
    int cFields = 0;
    int iProgressStep = 0;
    BOOL fAbort = FALSE;

    // get field values
    if (wzMessage)
    {
        // parse message string
        cFields = swscanf_s(wzMessage, L"1: %d 2: %d 3: %d 4: %d", &iFields[0], &iFields[1], &iFields[2], &iFields[3]);
    }
    else if (hRecord)
    {
        cFields = ::MsiRecordGetFieldCount(hRecord);
        cFields = min(cFields, countof(iFields)); // no buffer overrun if there are more fields than our buffer can hold
        for (int i = 0; i < cFields; ++i)
        {
            iFields[i] = ::MsiRecordGetInteger(hRecord, i + 1);
        }
    }

    // verify that we have enought field values
    if (1 > cFields)
    {
        ExitFunction(); // unknown message, bail
    }

    // hande based on message type
    switch (iFields[0])
    {
    case 0: // master progress reset
        if (4 > cFields)
        {
            Trace2(REPORT_STANDARD, "INSTALLMESSAGE_PROGRESS - Invalid field count %d, '%S'", cFields, wzMessage);
            break;
        }
        //Trace3(REPORT_STANDARD, "INSTALLMESSAGE_PROGRESS - MASTER RESET - %d, %d, %d", iFields[1], iFields[2], iFields[3]);

        // we only care about the first stage after script execution has started
        if (!pEngineInfo->fMsiProgressScriptInProgress && 1 != iFields[3])
        {
            pEngineInfo->fMsiProgressFinished = TRUE;
        }

        pEngineInfo->dwMsiProgressTotal = iFields[1];
        pEngineInfo->dwMsiProgressComplete = 0 == iFields[2] ? 0 : iFields[1]; // if forward start at 0, if backwards start at max
        pEngineInfo->fMsiProgressForward = 0 == iFields[2];
        pEngineInfo->fMsiProgressEnableActionData = FALSE;
        pEngineInfo->fMsiProgressScriptInProgress = 1 == iFields[3];

        // send progress update messages
        if (!pEngineInfo->fMsiProgressScriptInProgress && !pEngineInfo->fMsiProgressFinished)
        {
            SendProgressUpdate(pEngineInfo);
        }

        break;
    case 1: // action info
        if (3 > cFields)
        {
            Trace2(REPORT_STANDARD, "INSTALLMESSAGE_PROGRESS - Invalid field count %d, '%S'", cFields, wzMessage);
            break;
        }
        //Trace3(REPORT_STANDARD, "INSTALLMESSAGE_PROGRESS - ACTION INFO - %d, %d, %d", iFields[1], iFields[2], iFields[3]);

        pEngineInfo->dwMsiProgressStep = iFields[1];
        pEngineInfo->fMsiProgressEnableActionData = 0 != iFields[2];

        break;
    case 2: // progress report
        if (2 > cFields)
        {
            Trace2(REPORT_STANDARD, "INSTALLMESSAGE_PROGRESS - Invalid field count %d, '%S'", cFields, wzMessage);
            break;
        }
        //Trace3(REPORT_STANDARD, "INSTALLMESSAGE_PROGRESS - PROGRESS REPORT - %d, %d, %d", iFields[1], iFields[2], iFields[3]);

        if (!pEngineInfo->dwMsiProgressTotal)
        {
            break;
        }

        // pick step from either action info or message payload
        if (pEngineInfo->fMsiProgressEnableActionData)
        {
            iProgressStep = pEngineInfo->dwMsiProgressStep;
        }
        else
        {
            iProgressStep = iFields[1];
        }

        // update progress
        if (pEngineInfo->fMsiProgressForward)
        {
            pEngineInfo->dwMsiProgressComplete += iProgressStep;
        }
        else
        {
            pEngineInfo->dwMsiProgressComplete -= iProgressStep;
        }

        // send progress update messages
        if (!pEngineInfo->fMsiProgressScriptInProgress && !pEngineInfo->fMsiProgressFinished)
        {
            SendProgressUpdate(pEngineInfo);
        }

        break;

    default:
        ExitFunction(); // unknown message, bail
    }

    // check abort
    if (!pEngineInfo->fUserAbort && !pEngineInfo->fRollback) // only if the user has not already triggered an abort and we are not in rollback
    {
        hr = SendSuxQueryAbort(pEngineInfo, &fAbort);
        ExitOnFailure(hr, "Failed to send query abort message.");

        if (fAbort)
        {
            pEngineInfo->fUserAbort = TRUE;
            *piResult = IDCANCEL;
        }
    }

LExit:
    return;
}


static HRESULT SendProgressUpdate(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo
    )
{
    DWORD dwMsiProgressTotal = pEngineInfo->dwMsiProgressTotal;
    DWORD dwMsiProgressComplete = pEngineInfo->dwMsiProgressComplete; //min(dwMsiProgressTotal, pEngineInfo->dwMsiProgressComplete);
    double dProgressGauge = 0;
    double dProgressStageTotal = (double)pEngineInfo->qwProgressStageTotal;
    DWORD64 qwProgressComplete = 0;

    // calculate progress
    if (dwMsiProgressTotal) // avoid "divide by zero" if the MSI range is blank
    {
        // calculate gauge
        dProgressGauge = (double)dwMsiProgressComplete / (double)dwMsiProgressTotal;
        dProgressGauge = (1.0 / (1.0 + exp(3.7 - dProgressGauge * 7.5)) - 0.024127021417669196) / 0.975872978582330804;
        dwMsiProgressComplete = (DWORD)(dProgressGauge * dwMsiProgressTotal);

        // calculate progress within range
        qwProgressComplete = (DWORD64)(dwMsiProgressComplete * (dProgressStageTotal / dwMsiProgressTotal));
        qwProgressComplete = min(qwProgressComplete, pEngineInfo->qwProgressStageTotal);
    }

    // convert it to master progress
    if (!pEngineInfo->fRollback)
    {
        // for normal the stage progress is beyond the master complete point
        qwProgressComplete = pEngineInfo->qwProgressComplete + qwProgressComplete;
    }
    else
    {
        // for rollback the stage progress is before the master complete point
        qwProgressComplete = pEngineInfo->qwProgressComplete - qwProgressComplete;
    }

    // send progress update
    return SendSuxMasterProgressUpdateMessage(pEngineInfo, FALSE, pEngineInfo->qwProgressTotal, qwProgressComplete);
}


#if 0
HRESULT ApplyTransform(
    __in LPCWSTR wzMsiPath,
    __in LPCWSTR wzMstPath
    )
{
    HRESULT hr = S_OK;
    DWORD er = ERROR_SUCCESS;
    PMSIHANDLE hMsi;

    er = ::MsiOpenDatabaseW(wzMsiPath, reinterpret_cast<LPCWSTR>(MSIDBOPEN_TRANSACT), &hMsi);
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to open MSI database");

    er = ::MsiDatabaseApplyTransformW(hMsi, wzMstPath, MSITRANSFORM_ERROR_CHANGECODEPAGE);
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to apply transform");

    er = ::MsiDatabaseCommit(hMsi);
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to apply transform");

LExit:
    return hr;
}
#endif
