//-------------------------------------------------------------------------------------------------
// <copyright file="property.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//
// <summary>
//    Property managing functions for Burn.
// </summary>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"


// constants

const DWORD GROW_PROPERTY_ARRAY = 3;


// internal function declarations

static HRESULT FindPropertyIndexByName(
    __in BURN_PROPERTIES* pProperties,
    __in_z LPCWSTR wzProperty,
    __out DWORD* piProperty
    );
static HRESULT InsertProperty(
    __in BURN_PROPERTIES* pProperties,
    __in_z LPCWSTR wzProperty,
    __in DWORD iPosition
    );


// function definitions

extern "C" void PropertiesUninitialize(
    __in BURN_PROPERTIES* pProperties
    )
{
    if (pProperties->rgProperties)
    {
        for (DWORD i = 0; i < pProperties->cProperties; ++i)
        {
            BURN_PROPERTY* pProperty = &pProperties->rgProperties[i];
            if (pProperty)
            {
                ReleaseStr(pProperty->wzName);
                BVariantUninitialize(&pProperty->Value);
            }
        }
        MemFree(pProperties->rgProperties);
    }
}

extern "C" HRESULT PropertyGetNumeric(
    __in BURN_PROPERTIES* pProperties,
    __in_z LPCWSTR wzProperty,
    __out LONGLONG* pllValue
    )
{
    HRESULT hr = S_OK;
    DWORD iProperty = 0;

    hr = FindPropertyIndexByName(pProperties, wzProperty, &iProperty);
    ExitOnFailure(hr, "Failed to find property value.");

    if (S_FALSE == hr)
    {
        ExitFunction1(hr = E_INVALIDARG);
    }

    hr = BVariantGetNumeric(&pProperties->rgProperties[iProperty].Value, pllValue);
    ExitOnFailure(hr, "Failed to get value as numeric.");

LExit:
    return hr;
}

extern "C" HRESULT PropertyGetString(
    __in BURN_PROPERTIES* pProperties,
    __in_z LPCWSTR wzProperty,
    __out_z LPWSTR* psczValue
    )
{
    HRESULT hr = S_OK;
    DWORD iProperty = 0;

    hr = FindPropertyIndexByName(pProperties, wzProperty, &iProperty);
    ExitOnFailure(hr, "Failed to find property value.");

    if (S_FALSE == hr)
    {
        ExitFunction1(hr = E_INVALIDARG);
    }

    hr = BVariantGetString(&pProperties->rgProperties[iProperty].Value, psczValue);
    ExitOnFailure(hr, "Failed to get value as string.");

LExit:
    return hr;
}

extern "C" HRESULT PropertyGetVersion(
    __in BURN_PROPERTIES* pProperties,
    __in_z LPCWSTR wzProperty,
    __in DWORD64* pqwValue
    )
{
    HRESULT hr = S_OK;
    DWORD iProperty = 0;

    hr = FindPropertyIndexByName(pProperties, wzProperty, &iProperty);
    ExitOnFailure(hr, "Failed to find property value.");

    if (S_FALSE == hr)
    {
        ExitFunction1(hr = E_INVALIDARG);
    }

    hr = BVariantGetVersion(&pProperties->rgProperties[iProperty].Value, pqwValue);
    ExitOnFailure(hr, "Failed to get value as version.");

LExit:
    return hr;
}

extern "C" HRESULT PropertyGetVariant(
    __in BURN_PROPERTIES* pProperties,
    __in_z LPCWSTR wzProperty,
    __in BURN_VARIANT* pValue
    )
{
    HRESULT hr = S_OK;
    DWORD iProperty = 0;

    hr = FindPropertyIndexByName(pProperties, wzProperty, &iProperty);
    ExitOnFailure(hr, "Failed to find property value.");

    if (S_FALSE == hr)
    {
        ExitFunction1(hr = E_INVALIDARG);
    }

    hr = BVariantCopy(&pProperties->rgProperties[iProperty].Value, pValue);
    ExitOnFailure(hr, "Failed to copy property value.");

LExit:
    return hr;
}

extern "C" HRESULT PropertyGetFormatted(
    __in BURN_PROPERTIES* pProperties,
    __in_z LPCWSTR wzProperty,
    __out_z LPWSTR* psczValue
    )
{
    HRESULT hr = S_OK;
    DWORD iProperty = 0;
    BURN_PROPERTY* pProperty = NULL;

    hr = FindPropertyIndexByName(pProperties, wzProperty, &iProperty);
    ExitOnFailure(hr, "Failed to find property value.");

    if (S_FALSE == hr)
    {
        ExitFunction1(hr = E_INVALIDARG);
    }

    pProperty = &pProperties->rgProperties[iProperty];
    if (BURN_VARIANT_TYPE_STRING == pProperty->Value.Type)
    {
        hr = PropertyFormatString(pProperties, pProperty->Value.wzValue, psczValue);
        ExitOnFailure(hr, "Failed to format property value.");
    }
    else
    {
        hr = BVariantGetString(&pProperty->Value, psczValue);
        ExitOnFailure(hr, "Failed to get value as string.");
    }

LExit:
    return hr;
}

extern "C" HRESULT PropertySetNumeric(
    __in BURN_PROPERTIES* pProperties,
    __in_z LPCWSTR wzProperty,
    __in LONGLONG llValue
    )
{
    BURN_VARIANT variant = { 0 };

    variant.llValue = llValue;
    variant.Type = BURN_VARIANT_TYPE_NUMERIC;

    return PropertySetVariant(pProperties, wzProperty, &variant);
}

extern "C" HRESULT PropertySetString(
    __in BURN_PROPERTIES* pProperties,
    __in_z LPCWSTR wzProperty,
    __in_z_opt LPCWSTR wzValue
    )
{
    HRESULT hr = S_OK;
    DWORD iProperty = 0;

    hr = FindPropertyIndexByName(pProperties, wzProperty, &iProperty);
    ExitOnFailure(hr, "Failed to find property value.");

    // insert element if not found
    if (S_FALSE == hr)
    {
        hr = InsertProperty(pProperties, wzProperty, iProperty);
        ExitOnFailure(hr, "Failed to insert property.");
    }

    // update property value
    hr = BVariantSetString(&pProperties->rgProperties[iProperty].Value, wzValue, 0);
    ExitOnFailure(hr, "Failed to set property value.");

LExit:
    return hr;
}

extern "C" HRESULT PropertySetVersion(
    __in BURN_PROPERTIES* pProperties,
    __in_z LPCWSTR wzProperty,
    __in DWORD64 qwValue
    )
{
    BURN_VARIANT variant = { 0 };

    variant.qwValue = qwValue;
    variant.Type = BURN_VARIANT_TYPE_VERSION;

    return PropertySetVariant(pProperties, wzProperty, &variant);
}

extern "C" HRESULT PropertySetVariant(
    __in BURN_PROPERTIES* pProperties,
    __in_z LPCWSTR wzProperty,
    __in BURN_VARIANT* pVariant
    )
{
    HRESULT hr = S_OK;
    DWORD iProperty = 0;

    hr = FindPropertyIndexByName(pProperties, wzProperty, &iProperty);
    ExitOnFailure(hr, "Failed to find property value.");

    // insert element if not found
    if (S_FALSE == hr)
    {
        hr = InsertProperty(pProperties, wzProperty, iProperty);
        ExitOnFailure(hr, "Failed to insert property.");
    }

    // update property value
    hr = BVariantCopy(pVariant, &pProperties->rgProperties[iProperty].Value);
    ExitOnFailure(hr, "Failed to set property value.");

LExit:
    return hr;
}

extern "C" HRESULT PropertyFormatString(
    __in BURN_PROPERTIES* pProperties,
    __in_z LPCWSTR wzIn,
    __out_z LPWSTR* ppwzOut
    )
{
    HRESULT hr = S_OK;
    DWORD er = ERROR_SUCCESS;
    LPWSTR pwzUnformatted = NULL;
    LPWSTR pwzFormat = NULL;
    LPCWSTR wzRead = NULL;
    LPCWSTR wzOpen = NULL;
    LPCWSTR wzClose = NULL;
    LPWSTR pwz = NULL;
    LPWSTR* rgProperties = NULL;
    DWORD cProperties = 0;
    DWORD cch = 0;
    MSIHANDLE hRecord = NULL;

    // allocate buffer for format string
    hr = StrAlloc(&pwzFormat, lstrlenW(wzIn) + 1);
    ExitOnFailure(hr, "Failed to allocate buffer for format string.");

    // read out properties from the unformatted string and build a format string
    wzRead = wzIn;
    for (;;)
    {
        // scan for opening '['
        wzOpen = wcschr(wzRead, L'[');
        if (!wzOpen)
        {
            // end reached, append the remainder of the string and end loop
            hr = StrAllocConcat(&pwzFormat, wzRead, 0);
            ExitOnFailure(hr, "Failed to append string.");
            break;
        }

        // scan for closing ']'
        wzClose = wcschr(wzOpen + 1, L']');
        if (!wzClose)
        {
            // end reached, treat unterminated expander as literal
            hr = StrAllocConcat(&pwzFormat, wzRead, 0);
            ExitOnFailure(hr, "Failed to append string.");
            break;
        }
        cch = wzClose - wzOpen - 1;

        if (0 == cch)
        {
            // blank, copy all text including the terminator
            hr = StrAllocConcat(&pwzFormat, wzRead, (DWORD_PTR)(wzClose - wzRead) + 1);
            ExitOnFailure(hr, "Failed to append string.");
        }
        else
        {
            // append text preceding expander
            if (wzOpen > wzRead)
            {
                hr = StrAllocConcat(&pwzFormat, wzRead, (DWORD_PTR)(wzOpen - wzRead));
                ExitOnFailure(hr, "Failed to append string.");
            }

            // get property name
            hr = StrAllocString(&pwz, wzOpen + 1, cch);
            ExitOnFailure(hr, "Failed to get property name.");

            // allocate space in property array
            if (rgProperties)
            {
                LPVOID pv = MemReAlloc(rgProperties, sizeof(LPWSTR) * (cProperties + 1), TRUE);
                ExitOnNull(pv, hr, E_OUTOFMEMORY, "Failed to reallocate property array.");
                rgProperties = (LPWSTR*)pv;
            }
            else
            {
                rgProperties = (LPWSTR*)MemAlloc(sizeof(LPWSTR) * (cProperties + 1), TRUE);
                ExitOnNull(rgProperties, hr, E_OUTOFMEMORY, "Failed to allocate property array.");
            }

            // set property value
            if (2 <= cch && L'\\' == wzOpen[1])
            {
                // escape sequence, copy character
                hr = StrAllocString(&rgProperties[cProperties], &wzOpen[2], 1);
            }
            else
            {
                // get formatted property value
                hr = PropertyGetFormatted(pProperties, pwz, &rgProperties[cProperties]);
                if (E_INVALIDARG == hr) // property not found
                {
                    hr = StrAllocString(&rgProperties[cProperties], L"", 0);
                }
            }
            ExitOnFailure(hr, "Failed to set property value.");
            ++cProperties;

            // append placeholder to format string
            hr = StrAllocFormatted(&pwz, L"[%d]", cProperties);
            ExitOnFailure(hr, "Failed to format placeholder string.");

            hr = StrAllocConcat(&pwzFormat, pwz, 0);
            ExitOnFailure(hr, "Failed to append placeholder.");
        }

        // update read pointer
        wzRead = wzClose + 1;
    }

    // create record
    hRecord = ::MsiCreateRecord(cProperties);
    ExitOnNull(hRecord, hr, E_OUTOFMEMORY, "Failed to allocate record.");

    // set format string
    er = ::MsiRecordSetStringW(hRecord, 0, pwzFormat);
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to set record format string.");

    // copy record fields
    for (DWORD i = 0; i < cProperties; ++i)
    {
        if (*rgProperties[i]) // not setting if blank
        {
            er = ::MsiRecordSetStringW(hRecord, i + 1, rgProperties[i]);
            hr = HRESULT_FROM_WIN32(er);
            ExitOnFailure(hr, "Failed to set record string.");
        }
    }

    // format record
    cch = 0;
    er = ::MsiFormatRecordW(NULL, hRecord, L"", &cch);
    if (ERROR_MORE_DATA == er || ERROR_SUCCESS == er)
    {
        hr = StrAlloc(&pwz, ++cch);
    }
    else
    {
        hr = HRESULT_FROM_WIN32(er);
    }
    ExitOnFailure(hr, "Failed to allocate string.");

    er = ::MsiFormatRecordW(NULL, hRecord, pwz, &cch);
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to format record.");

    // return value
    hr = StrAllocString(ppwzOut, pwz, 0);
    ExitOnFailure(hr, "Failed to copy string.");

LExit:
    if (rgProperties)
    {
        for (DWORD i = 0; i < cProperties; ++i)
        {
            ReleaseStr(rgProperties[i]);
        }
        MemFree(rgProperties);
    }

    if (hRecord)
    {
        ::MsiCloseHandle(hRecord);
    }

    ReleaseStr(pwzUnformatted);
    ReleaseStr(pwzFormat);
    ReleaseStr(pwz);

    return hr;
}

extern "C" HRESULT PropertyEscapeString(
    __in_z LPCWSTR wzIn,
    __out_z LPWSTR* ppwzOut
    )
{
    HRESULT hr = S_OK;
    LPCWSTR wzRead = NULL;
    LPWSTR pwzEscaped = NULL;
    LPWSTR pwz = NULL;
    SIZE_T i = 0;

    // allocate buffer for escaped string
    hr = StrAlloc(&pwzEscaped, lstrlenW(wzIn) + 1);
    ExitOnFailure(hr, "Failed to allocate buffer for escaped string.");

    // read trough string and move characters, inserting escapes as needed
    wzRead = wzIn;
    for (;;)
    {
        // find next character needing escaping
        i = wcscspn(wzRead, L"[]{}");

        // copy skipped characters
        if (0 < i)
        {
            hr = StrAllocConcat(&pwzEscaped, wzRead, i);
            ExitOnFailure(hr, "Failed to append characters.");
        }

        if (L'\0' == wzRead[i])
        {
            break; // end reached
        }

        // escape character
        hr = StrAllocFormatted(&pwz, L"[\\%c]", wzRead[i]);
        ExitOnFailure(hr, "Failed to format escape sequence.");

        hr = StrAllocConcat(&pwzEscaped, pwz, 0);
        ExitOnFailure(hr, "Failed to append escape sequence.");

        // update read pointer
        wzRead += i + 1;
    }

    // return value
    hr = StrAllocString(ppwzOut, pwzEscaped, 0);
    ExitOnFailure(hr, "Failed to copy string.");

LExit:
    ReleaseStr(pwzEscaped);
    ReleaseStr(pwz);
    return hr;
}


// internal function definitions

static HRESULT FindPropertyIndexByName(
    __in BURN_PROPERTIES* pProperties,
    __in_z LPCWSTR wzProperty,
    __out DWORD* piProperty
    )
{
    HRESULT hr = S_OK;
    DWORD iRangeFirst = 0;
    DWORD cRangeLength = pProperties->cProperties;
    int cchProperty = lstrlenW(wzProperty);

    while (cRangeLength)
    {
        // get property in middle of range
        DWORD iPosition = (cRangeLength / 2);
        BURN_PROPERTY* pProperty = &pProperties->rgProperties[iRangeFirst + iPosition];

        switch (::CompareStringW(LOCALE_INVARIANT, SORT_STRINGSORT, wzProperty, cchProperty, pProperty->wzName, -1))
        {
        case CSTR_LESS_THAN:
            // restrict range to elements before the current
            cRangeLength = iPosition;
            break;
        case CSTR_EQUAL:
            // property found
            *piProperty = iRangeFirst + iPosition;
            ExitFunction1(hr = S_OK);
        case CSTR_GREATER_THAN:
            // restrict range to elements after the current
            iRangeFirst += iPosition + 1;
            cRangeLength -= iPosition + 1;
            break;
        default:
            ExitWithLastError(hr, "Failed to compare strings.");
        }
    }

    *piProperty = iRangeFirst;
    hr = S_FALSE; // property not found

LExit:
    return hr;
}

static HRESULT InsertProperty(
    __in BURN_PROPERTIES* pProperties,
    __in_z LPCWSTR wzProperty,
    __in DWORD iPosition
    )
{
    HRESULT hr = S_OK;

    // ensure there is room in the property array
    if (pProperties->cProperties == pProperties->dwMaxProperties)
    {
        pProperties->dwMaxProperties += GROW_PROPERTY_ARRAY;

        if (pProperties->rgProperties)
        {
            LPVOID pv = MemReAlloc(pProperties->rgProperties, sizeof(BURN_PROPERTY) * pProperties->dwMaxProperties, FALSE);
            ExitOnNull(pv, hr, E_OUTOFMEMORY, "Failed to allocate room for more properties.");

            pProperties->rgProperties = (BURN_PROPERTY*)pv;
            memset(&pProperties->rgProperties[pProperties->cProperties], 0, sizeof(BURN_PROPERTY) * (pProperties->dwMaxProperties - pProperties->cProperties));
        }
        else
        {
            pProperties->rgProperties = (BURN_PROPERTY*)MemAlloc(sizeof(BURN_PROPERTY) * pProperties->dwMaxProperties, TRUE);
            ExitOnNull(pProperties->rgProperties, hr, E_OUTOFMEMORY, "Failed to allocate room for properties.");
        }
    }

    // move properties
    if (0 < pProperties->cProperties - iPosition)
    {
        memmove(&pProperties->rgProperties[iPosition + 1], &pProperties->rgProperties[iPosition], sizeof(BURN_PROPERTY) * (pProperties->cProperties - iPosition));
        memset(&pProperties->rgProperties[iPosition], 0, sizeof(BURN_PROPERTY));
    }

    ++pProperties->cProperties;

    // allocate name
    hr = StrAllocString(&pProperties->rgProperties[iPosition].wzName, wzProperty, 0);
    ExitOnFailure(hr, "Failed to copy property name.");

LExit:
    return hr;
}
