//-------------------------------------------------------------------------------------------------
// <copyright file="worker.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//
// <summary>
// Worker thread for setup stub.
// </summary>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"


// structs

typedef struct
{
    WCHAR wzProductCode[39];

    WCHAR wzTempPath[MAX_PATH];
    WCHAR wzCachedPath[MAX_PATH];

    BOOL bCached;
    BOOL bInstalled;
} SETUP_ROLLBACK_PACKAGE;


// globals

#ifdef DEBUG
SETUP_DEBUG_HOOKS EngineDebugHooks = { 0 };
#endif


// prototypes

static HRESULT InitializePipe(
    __in LPCWSTR wzPipeName,
    __out HANDLE *phPipe
    );
static HRESULT ListenWorkerMessagePipe(
    __in HANDLE hPipe
    );
static HRESULT DoRegistrySearches(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo
    );
static HRESULT RegistrySearchExists(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in HKEY hRoot,
    __in LPCWSTR wzKey,
    __in LPCWSTR wzProperty
    );
static HRESULT RegistrySearchValue(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in HKEY hRoot,
    __in LPCWSTR wzKey,
    __in LPCWSTR wzValue,
    __in LPCWSTR wzProperty
    );
static HRESULT PrepareSetupPackages(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in SETUP_PACKAGES *pSetupPackages
    );
static HRESULT PrepareSetupPackage(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in SETUP_PACKAGE *pSetupPackage,
    __in DWORD dwPackageIndex,
    __inout DWORD64 *pqwProgressTotal,
    __inout DWORD* pdwVolumes,
    __inout SETUP_DISK_COST_INFO** prgVolumes
    );
static HRESULT AddDiskCost(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in SETUP_DISK_COST* pDiskCost,
    __inout DWORD* pdwVolumes,
    __inout SETUP_DISK_COST_INFO** prgVolumes
    );
static HRESULT CacheSetupPackages(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in SETUP_PACKAGES *pSetupPackages
    );
static HRESULT CacheSetupPackage(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in SETUP_PACKAGE *pSetupPackage,
    __in DWORD dwPackageIndex
    );
static HRESULT ExtractPackage(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in DWORD dwIndex,
    __in LPCWSTR wzSource,
    __in LPCWSTR wzFileName,
    __out LPWSTR* ppwzPackagePath
    );
static HRESULT DownloadPackage(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in DWORD dwIndex,
    __in LPCWSTR wzSource,
    __in LPCWSTR wzFileName,
    __out LPWSTR* ppwzPackagePath
    );
static HRESULT CreatePackageTempPath(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in DWORD dwIndex,
    __out LPWSTR* ppwzPackageTempPath
    );
static HRESULT GetWorkerMessage(
    __in HANDLE hPipe,
    __in SETUP_COMM_MESSAGE *pmsg
    );
static HRESULT ProcessMessage(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in SETUP_COMM_MESSAGE *pMsg
    );
static void FreeWorkerMessage(
    __in SETUP_COMM_MESSAGE *pmsg
    );
static HRESULT ExecutePackages(
    __in SETUP_PACKAGES *pSetupPackages,
    __in SETUP_ENGINE_INFORMATION *pSetupUXInformation
    );
static HRESULT ExecutePackage(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in SETUP_PACKAGE *pSetupPackage,
    __in DWORD dwPackageIndex,
    __out LPDWORD pdwCode,
    __out LPBOOL pfAbort
    );
static HRESULT RollbackPackages(
    __in SETUP_PACKAGES *pSetupPackages,
    __in SETUP_ENGINE_INFORMATION *pEngineInfo
    );
static HRESULT RollbackPackage(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in SETUP_PACKAGE *pSetupPackage,
    __in DWORD dwPackageIndex,
    __out LPDWORD pdwCode
    );
static HRESULT ResolveRelativePath(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __inout SETUP_PACKAGE* pSetupPackage
    );
static HRESULT ValidateSourcePath(
    __in SETUP_PACKAGE* pSetupPackage,
    __in LPCWSTR wzPath
    );
static HRESULT GetMsiProductCodeAndVersion(
    __in LPCWSTR wzMsiPath,
    __out LPGUID pguidProductCode,
    __out DWORD64* pqwProductVersion
    );
static HRESULT GetBasePathForRelativePath(
    __in LPCWSTR wzFullPath,
    __in LPCWSTR wzRelativePath,
    __out LPWSTR* ppwzBasePath
    );
static HRESULT ConcatMsiFeatureActionArguments(
    __in SETUP_PACKAGE_MSI* pMsiPackage,
    __in BOOL fRollback,
    __inout LPWSTR* ppwzArguments
    );
static HRESULT ConcatMsiPropertyArguments(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SETUP_PACKAGE_MSI* pMsiPackage,
    __in BOOL fRollback,
    __inout LPWSTR* ppwzArguments
    );


// function definitions

DWORD WINAPI WorkerThread(
    __in LPVOID pvContext
    )
{
    HRESULT hr = S_OK;
    BOOL fComInitialized = FALSE;
    SETUP_WORKER_INFORMATION* pWorkerInfo = static_cast<SETUP_WORKER_INFORMATION*>(pvContext);
    SETUP_ENGINE_INFORMATION* pEngineInfo = pWorkerInfo->pEngineInfo;
    HANDLE hPipe = INVALID_HANDLE_VALUE;
    SETUP_COMM_MESSAGE msg = { 0 };

    // initialize COM
    hr = ::CoInitialize(NULL);
    ExitOnFailure(hr, "Failed to initialize COM.");
    fComInitialized = TRUE;

    // initialize message pipe
    hr = InitializePipe(pEngineInfo->pwzPipe, &hPipe);
    ExitOnFailure(hr, "Failed to initialize message pipe.");

    // signal that the thread was initialized successfully
    if (!::SetEvent(pWorkerInfo->hInitEvent))
    {
        ExitWithLastError(hr, "Failed to set init event.");
    }
    pWorkerInfo = NULL; // after this point pWorkerInfo is no longer valid

    // wait for client connection
    hr = ListenWorkerMessagePipe(hPipe);
    ExitOnFailure(hr, "Failed to connect to extension's pipe.");

    while (S_OK == (hr = GetWorkerMessage(hPipe, &msg)))
    {
        if (SWM_QUIT == msg.dwMessage)
        {
            Trace(REPORT_STANDARD, "Worker thread received quit message.");
            break;
        }

        hr = ProcessMessage(pEngineInfo, &msg);
        ExitOnFailure(hr, "Failed to process message.");

        FreeWorkerMessage(&msg);
    }

LExit:
    if (INVALID_HANDLE_VALUE != hPipe)
    {
        ::CloseHandle(hPipe);
    }

    FreeWorkerMessage(&msg);

    if (fComInitialized)
    {
        ::CoUninitialize();
    }

    return hr;
}


LRESULT SendSuxMessage(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in UINT Msg,
    __in LPVOID pvData
    )
{
    return pEngineInfo->pfnMessageHandler((LPVOID)pEngineInfo->pvHandlerContext, Msg, pvData);
}


// privat helper function definitions

static HRESULT InitializePipe(
    __in LPCWSTR wzPipeName,
    __out HANDLE *phPipe
    )
{
    HRESULT hr = S_OK;
    HANDLE hPipe = INVALID_HANDLE_VALUE;

    // TODO: use a different security descriptor to lock down this named pipe even more than it is now.
    hPipe = ::CreateNamedPipeW(wzPipeName, PIPE_ACCESS_DUPLEX | FILE_FLAG_FIRST_PIPE_INSTANCE, PIPE_TYPE_BYTE | PIPE_READMODE_BYTE | PIPE_WAIT, 1, 64 * 1024, 64 * 1024, 1, NULL);
    if (INVALID_HANDLE_VALUE == hPipe)
    {
        ExitWithLastError1(hr, "Failed to create setup communication pipe: %S", wzPipeName);
    }

    *phPipe = hPipe;
    hPipe = INVALID_HANDLE_VALUE;

LExit:
    if (INVALID_HANDLE_VALUE != hPipe)
    {
        ::CloseHandle(hPipe);
    }

    return hr;
}


static HRESULT ListenWorkerMessagePipe(
    __in HANDLE hPipe
    )
{
    HRESULT hr = S_OK;

    if (!::ConnectNamedPipe(hPipe, NULL))
    {
        DWORD er = ::GetLastError();
        if (ERROR_PIPE_CONNECTED != er)
        {
            hr = HRESULT_FROM_WIN32(er);
            ExitOnFailure(hr, "Failed to connect to worker message pipe.");
        }
    }

LExit:
    return hr;
}


static HRESULT DoRegistrySearches(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo
    )
{
    HRESULT hr = S_OK;

    for (DWORD i = 0; i < pEngineInfo->pSetupPackages->cRegistrySearches; ++i)
    {
        SETUP_REGISTRY_SEARCH* pRegistrySearch = &pEngineInfo->pSetupPackages->rgRegistrySearches[i];

        switch (pRegistrySearch->Type)
        {
        case REGISTRY_SEARCH_TYPE_EXISTS:
            LogStringLine(REPORT_STANDARD, "RegistrySearch: Checking registry key existence. Id = '%S'", pRegistrySearch->wzId);
            hr = RegistrySearchExists(pEngineInfo, pRegistrySearch->hRoot, pRegistrySearch->wzKey, pRegistrySearch->wzProperty);
            break;
        case REGISTRY_SEARCH_TYPE_VALUE:
            LogStringLine(REPORT_STANDARD, "RegistrySearch: Searching for registry key value. Id = '%S'", pRegistrySearch->wzId);
            hr = RegistrySearchValue(pEngineInfo, pRegistrySearch->hRoot, pRegistrySearch->wzKey, pRegistrySearch->wzValue, pRegistrySearch->wzProperty);
            break;
        default:
            ExitFunction1(hr = E_UNEXPECTED);
        }
        if (FAILED(hr))
        {
            TraceError1(hr, "Registry search failed. Id = '%S'", pRegistrySearch->wzId);
        }
    }

    hr = S_OK;

LExit:
    return hr;
}


static HRESULT RegistrySearchExists(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in HKEY hRoot,
    __in LPCWSTR wzKey,
    __in LPCWSTR wzProperty
    )
{
    HRESULT hr = S_OK;
    DWORD er = ERROR_SUCCESS;
    HKEY hKey = NULL;

    er = ::RegOpenKeyExW(hRoot, wzKey, 0, KEY_QUERY_VALUE, &hKey);
    switch (er)
    {
    case ERROR_SUCCESS:
        hr = SetIntProperty(pEngineInfo, wzProperty, 1);
        break;
    case ERROR_FILE_NOT_FOUND:
        hr = SetIntProperty(pEngineInfo, wzProperty, 0);
        break;
    default:
        hr = HRESULT_FROM_WIN32(er);
        ExitOnFailure(hr, "Failed to open registry key.");
    }
    ExitOnFailure(hr, "Failed to set property.");

LExit:
    if (hKey)
    {
        ::RegCloseKey(hKey);
    }
    return hr;
}


static HRESULT RegistrySearchValue(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in HKEY hRoot,
    __in LPCWSTR wzKey,
    __in LPCWSTR wzValue,
    __in LPCWSTR wzProperty
    )
{
    HRESULT hr = S_OK;
    DWORD er = ERROR_SUCCESS;
    HKEY hKey = NULL;
    DWORD dwType = 0;
    DWORD cbData = 0;
    LPBYTE pData = NULL;

    er = ::RegOpenKeyExW(hRoot, wzKey, 0, KEY_QUERY_VALUE, &hKey);
    if (ERROR_FILE_NOT_FOUND == er)
    {
        LogStringLine(REPORT_STANDARD, "Registry key not found. Key = '%S'", wzKey);
        ExitFunction1(hr = S_OK);
    }
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to open registry key.");

    er = ::RegQueryValueExW(hKey, wzValue, NULL, &dwType, NULL, &cbData);
    if (ERROR_FILE_NOT_FOUND == er)
    {
        LogStringLine(REPORT_STANDARD, "Registry value not found. Key = '%S', Value = '%S'", wzKey, wzValue);
        ExitFunction1(hr = S_OK);
    }
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to query registry key value size.");

    pData = (LPBYTE)MemAlloc(cbData + sizeof(WCHAR), TRUE); // + sizeof(WCHAR) here to ensure that we always have a null terminator for REG_SZ
    ExitOnNull(pData, hr, E_OUTOFMEMORY, "Failed to allocate memory registry value.");

    er = ::RegQueryValueExW(hKey, wzValue, NULL, &dwType, pData, &cbData);
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to query registry key value.");

    switch (dwType)
    {
    case REG_DWORD:
        if (sizeof(LONG) != cbData)
        {
            ExitFunction1(hr = E_UNEXPECTED);
        }
        hr = SetIntProperty(pEngineInfo, wzProperty, *((LONG*)pData));
        break;
    case REG_QWORD:
        if (sizeof(LONGLONG) != cbData)
        {
            ExitFunction1(hr = E_UNEXPECTED);
        }
        hr = SetIntProperty(pEngineInfo, wzProperty, *((LONGLONG*)pData));
        break;
    case REG_SZ:
        hr = SetStrProperty(pEngineInfo, wzProperty, (LPCWSTR)pData);
        break;
    default:
        ExitOnFailure1(hr = E_FAIL, "Unsuported registry key value type. Type = '%d'", dwType);
    }
    ExitOnFailure(hr, "Failed to set property.");

LExit:
    ReleaseMem(pData);
    if (hKey)
    {
        ::RegCloseKey(hKey);
    }
    return hr;
}


static HRESULT PrepareSetupPackages(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in SETUP_PACKAGES *pSetupPackages
    )
{
    HRESULT hr = S_OK;
    DWORD64 qwProgressTotal = 0;
    DWORD dwVolumes = 0;
    SETUP_DISK_COST_INFO* rgVolumes = NULL;;

    for (DWORD i = 0; i < pSetupPackages->cPackages; ++i)
    {
        SETUP_PACKAGE *pSetupPackage = pSetupPackages->rgPackages + i;

        //
        // should we process the package?
        //

        if (PACKAGE_STATUS_PENDING_PREPARE != pSetupPackage->Status)
        {
            continue; // package is not pending prepare
        }


        //
        // send package begin message
        //

        hr = SendSuxPreparePackageBeginMessage(pEngineInfo, i);
        ExitOnFailure(hr, "Failed to send package begin message.");


        //
        // process the package
        //

        hr = PrepareSetupPackage(pEngineInfo, pSetupPackage, i, &qwProgressTotal, &dwVolumes, &rgVolumes);
        ExitOnFailure(hr, "Failed to prepare setup package.");


        //
        // send package complete message
        //

        hr = SendSuxPreparePackageCompleteMessage(pEngineInfo, i, pSetupPackage->CurrentState, pSetupPackage->ActionState, pSetupPackage->Status);
        ExitOnFailure(hr, "Failed to send package complete message.");
    }

    // set total progress
    pEngineInfo->qwProgressTotal = qwProgressTotal;


    //
    // send disk cost message
    //

    hr = SendSuxPrepareDiskCostMessage(pEngineInfo, rgVolumes, dwVolumes);
    ExitOnFailure(hr, "Failed to send disk cost message.");

LExit:
    if (rgVolumes)
    {
        for (DWORD i = 0; i < dwVolumes; ++i)
        {
            ReleaseStr(rgVolumes[i].wzVolumeRootPath);
        }
        MemFree(rgVolumes);
    }
    return hr;
}


static HRESULT PrepareSetupPackage(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SETUP_PACKAGE* pSetupPackage,
    __in DWORD dwPackageIndex,
    __inout DWORD64 *pqwProgressTotal,
    __inout DWORD* pdwVolumes,
    __inout SETUP_DISK_COST_INFO** prgVolumes
    )
{
    HRESULT hr = S_OK;
    DWORD64 qwProgressTicks = 0;
    ACTION_STATE executeAction = ACTION_STATE_NONE;
    ACTION_STATE rollbackAction = ACTION_STATE_NONE;
    BOOL fHasFeatureAction = FALSE;
    BOOL fCondition = FALSE;
    BOOL fPrereqCheck = FALSE;
    BOOL fPrereqChecks = TRUE;


    //
    // evaluate package condition
    //

    if (pSetupPackage->pwzCondition && *pSetupPackage->pwzCondition)
    {
        hr = EvaluateCondition(pEngineInfo, pSetupPackage->pwzCondition, &fCondition);
        ExitOnFailure1(hr, "Failed to evaluate package condition: '%S'", pSetupPackage->pwzCondition);

        if (!fCondition)
        {
            // conditionalized out, set status accordingly
            pSetupPackage->Status = PACKAGE_STATUS_IGNORED;
            ExitFunction();
        }
    }

    if (pSetupPackage->fPermanent && PACKAGE_STATE_ABSENT == pSetupPackage->RequestedState)
    {
        // the package is permanent, but we are trying to uninstall it
        pSetupPackage->Status = PACKAGE_STATUS_IGNORED;
        ExitFunction();
    }


    //
    // detect the current state
    //

    switch (pSetupPackage->PackageType)
    {
    case SETUP_PACKAGE_TYPE_MSI:
        hr = DetectMsi(pSetupPackage->Msi.guidProductCode, pSetupPackage->Msi.qwVersion, pSetupPackage->fRepair,
            pSetupPackage->RequestedState, pSetupPackage->ExpectedState, &pSetupPackage->CurrentState, &executeAction, &rollbackAction);
        ExitOnFailure(hr, "Failed to calculate current MSI state.");

        // detect features
        for (DWORD i = 0; i < pSetupPackage->Msi.cFeatures; i++)
        {
            SETUP_PACKAGE_MSI_FEATURE* pFeature = &pSetupPackage->Msi.rgFeatures[i];

            // send feature detection begin message
            hr = SendSuxPrepareFeatureBeginMessage(pEngineInfo, dwPackageIndex, i);
            ExitOnFailure(hr, "Failed to send feature begin message.");

            hr = DetectMsiFeature(pSetupPackage->Msi.guidProductCode, pSetupPackage->CurrentState, pFeature->wzId, pFeature->fRepair,
                pFeature->RequestedState, pFeature->ExpectedState, &pFeature->CurrentState, &pFeature->ExecuteAction, &pFeature->RollbackAction);
            ExitOnFailure1(hr, "Failed to calculate current feature state. Feature = '%S'", pFeature->wzId);

            // add features that has an action state unequal to none to progress
            if (FEATURE_ACTION_NONE != pFeature->ExecuteAction)
            {
                fHasFeatureAction = TRUE;
            }

            // send feature detection complete message
            hr = SendSuxPrepareFeatureCompleteMessage(pEngineInfo, dwPackageIndex, i, pFeature->CurrentState, pFeature->ExecuteAction);
            ExitOnFailure(hr, "Failed to send feature complete message.");
        }

        // we only want to do maintenance if there are feature actions
        if (ACTION_STATE_MAINTENANCE == executeAction && !fHasFeatureAction)
        {
            pSetupPackage->ActionState = ACTION_STATE_NONE;
            pSetupPackage->RollbackAction = ACTION_STATE_NONE;
        }
        else
        {
            pSetupPackage->ActionState = executeAction;
            pSetupPackage->RollbackAction = rollbackAction;
        }
        break;
    case SETUP_PACKAGE_TYPE_EXE:
        hr = DetectExe(pEngineInfo, &pSetupPackage->Exe, pSetupPackage->fRepair,
            pSetupPackage->RequestedState, pSetupPackage->ExpectedState, &pSetupPackage->CurrentState, &pSetupPackage->ActionState, &pSetupPackage->RollbackAction);
        ExitOnFailure(hr, "Failed to calculate current EXE state.");
        break;
    default:
        hr = E_NOTIMPL;
        ExitOnFailure(hr, "Have not implemented other package type support yet");
    }

    if (ACTION_STATE_NONE == pSetupPackage->ActionState)
    {
        // no action needed for the package
        pSetupPackage->Status = PACKAGE_STATUS_IGNORED;
        ExitFunction();
    }


    //
    // prerequisite checks
    //

    // if we are installing
    if (ACTION_STATE_UNINSTALL != pSetupPackage->ActionState)
    {
        for (DWORD i = 0; i < pSetupPackage->cPrereqChecks; ++i)
        {
            SETUP_PREREQUISITE_CHECK* pPrereqCheck = &pSetupPackage->rgPrereqChecks[i];

            // do prerequisite check
            fPrereqCheck = TRUE; // a blank prerequisite check should always pass
            if (pPrereqCheck->pwzCondition && *pPrereqCheck->pwzCondition)
            {
                hr = EvaluateCondition(pEngineInfo, pPrereqCheck->pwzCondition, &fPrereqCheck);
                ExitOnFailure1(hr, "Failed to evaluate prerequisite check: '%S'", pPrereqCheck->pwzCondition);
            }

            // send prerequisite check message
            hr = SendSuxPrerequisiteCheckMessage(pEngineInfo, dwPackageIndex, i, fPrereqCheck);
            ExitOnFailure(hr, "Failed to send prerequisite check message.");

            // if the individal check faild, indicate that package has failed prerequisit check
            if (!fPrereqCheck)
            {
                fPrereqChecks = FALSE;
            }
        }

        if (!fPrereqChecks)
        {
            LogStringLine(REPORT_STANDARD, "Package prerequisite check failed. The package will not be installed. Id = '%S'", pSetupPackage->wzId);

            // prerequisite check failed, set status accordingly
            pSetupPackage->Status = PACKAGE_STATUS_PREREQUISITE_CHECK_FAILED;
            ExitFunction();
        }
    }


    //
    // calculate cost
    //

    qwProgressTicks += pSetupPackage->qwCost;

    // add disk cost
    for (DWORD dci = 0; dci < pSetupPackage->cDiskCosts; ++dci)
    {
        SETUP_DISK_COST* pDiskCost = &pSetupPackage->rgDiskCosts[dci];
        qwProgressTicks += pDiskCost->qwCost; // add to progress always

        if (ACTION_STATE_INSTALL == pSetupPackage->ActionState ||
            ACTION_STATE_ADMIN_INSTALL == pSetupPackage->ActionState)
        {
            hr = AddDiskCost(pEngineInfo, pDiskCost, pdwVolumes, prgVolumes);
            if (FAILED(hr))
            {
                // bad, but not catastrophic
                TraceError(hr, "Failed to add disk cost for package.");
            }
        }
    }

    if (SETUP_PACKAGE_TYPE_MSI == pSetupPackage->PackageType)
    {
        for (DWORD i = 0; i < pSetupPackage->Msi.cFeatures; i++)
        {
            SETUP_PACKAGE_MSI_FEATURE* pFeature = &pSetupPackage->Msi.rgFeatures[i];

            if (FEATURE_ACTION_NONE != pFeature->ExecuteAction)
            {
                qwProgressTicks += pFeature->qwCost;

                // add disk cost
                for (DWORD dci = 0; dci < pFeature->cDiskCosts; ++dci)
                {
                    SETUP_DISK_COST* pDiskCost = &pFeature->rgDiskCosts[dci];
                    qwProgressTicks += pDiskCost->qwCost; // add to progress always

                    if (FEATURE_ACTION_ADDLOCAL == pFeature->ExecuteAction ||
                        FEATURE_ACTION_ADDSOURCE == pFeature->ExecuteAction ||
                        FEATURE_ACTION_ADDDEFAULT == pFeature->ExecuteAction)
                    {
                        hr = AddDiskCost(pEngineInfo, pDiskCost, pdwVolumes, prgVolumes);
                        if (FAILED(hr))
                        {
                            // bad, but not catastrophic
                            TraceError1(hr, "Failed to add disk cost for feature. Id = '%S'", pFeature->wzId);
                        }
                    }
                }
            }
        }
    }

    // update progress
    pSetupPackage->qwProgressTicks = qwProgressTicks;
    pSetupPackage->qwProgressOffset = *pqwProgressTotal;
    *pqwProgressTotal += qwProgressTicks;

    // successful, set status to pending cache
    pSetupPackage->Status = PACKAGE_STATUS_PENDING_CACHE;

LExit:
    return hr;
}


static HRESULT AddDiskCost(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in SETUP_DISK_COST* pDiskCost,
    __inout DWORD* pdwVolumes,
    __inout SETUP_DISK_COST_INFO** prgVolumes
    )
{
    HRESULT hr = S_OK;
    LPWSTR pwzPath = NULL;
    LPWSTR pwzVolumePath = NULL;
    DWORD cchVolumePath = 0;
    DWORD dwVolumeSerialNumber = 0;
    SETUP_DISK_COST_INFO* pDiskCostInfo = NULL;
    ULARGE_INTEGER uliTotal = { 0 };
    ULARGE_INTEGER uliFree = { 0 };

    // format path string
    hr = FormatPropertyString(pEngineInfo, pDiskCost->wzPath, &pwzPath);
    ExitOnFailure(hr, "Failed to format path.");

    // get volume path
    cchVolumePath = ::GetFullPathNameW(pwzPath, 0, NULL, NULL);
    if (0 == cchVolumePath)
    {
        ExitWithLastError(hr, "Failed to get full path name.");
    }

    hr = StrAlloc(&pwzVolumePath, cchVolumePath);
    ExitOnFailure(hr, "Failed to allocate volume path buffer.");

    if (!::GetVolumePathNameW(pwzPath, pwzVolumePath, cchVolumePath))
    {
        ExitWithLastError(hr, "Failed to get volume path name.");
    }

    if (!::GetVolumeInformationW(pwzVolumePath, NULL, 0, &dwVolumeSerialNumber, NULL, NULL, NULL, 0))
    {
        ExitWithLastError(hr, "Failed to get volume information.");
    }

    // find existing volume in list
    for (DWORD i = 0; i < *pdwVolumes; ++i)
    {
        if ((*prgVolumes + i)->dwVolumeSerialNumber == dwVolumeSerialNumber)
        {
            pDiskCostInfo = (*prgVolumes + i);
            break;
        }
    }

    if (!pDiskCostInfo)
    {
        // get disk space
        if (!::GetDiskFreeSpaceExW(pwzVolumePath, NULL, &uliTotal, &uliFree))
        {
            ExitWithLastError(hr, "Failed to disk free space.");
        }

        // make room in the volume array
        if (*prgVolumes)
        {
            LPVOID pv = MemReAlloc(*prgVolumes, sizeof(SETUP_DISK_COST_INFO) * (*pdwVolumes + 1), TRUE);
            ExitOnNull(pv, hr, E_OUTOFMEMORY, "Failed to reallocate disk cost info array.");
            *prgVolumes = (SETUP_DISK_COST_INFO*)pv;
        }
        else
        {
            *prgVolumes = (SETUP_DISK_COST_INFO*)MemAlloc(sizeof(SETUP_DISK_COST_INFO) * (*pdwVolumes + 1), TRUE);
            ExitOnNull(*prgVolumes, hr, E_OUTOFMEMORY, "Failed to allocate disk cost info array.");
        }

        // insert the volume
        for (DWORD i = 0; i < *pdwVolumes; ++i)
        {
            if (CSTR_GREATER_THAN == ::CompareStringW(LOCALE_INVARIANT, 0, (*prgVolumes)[i].wzVolumeRootPath, -1, pwzVolumePath, -1))
            {
                ::MoveMemory((*prgVolumes + i + 1), (*prgVolumes + i), sizeof(SETUP_DISK_COST_INFO) * (*pdwVolumes - i));
                pDiskCostInfo = (*prgVolumes + i);
                break;
            }
        }
        if (!pDiskCostInfo)
        {
            pDiskCostInfo = (*prgVolumes + *pdwVolumes);
        }
        ++*pdwVolumes;

        // set values
        pDiskCostInfo->wzVolumeRootPath = pwzVolumePath;
        pwzVolumePath = NULL;
        pDiskCostInfo->dwVolumeSerialNumber = dwVolumeSerialNumber;
        pDiskCostInfo->qwTotal = uliTotal.QuadPart;
        pDiskCostInfo->qwFree = uliFree.QuadPart;
    }

    // add cost
    pDiskCostInfo->qwCost += pDiskCost->qwCost;

LExit:
    ReleaseStr(pwzPath);
    ReleaseStr(pwzVolumePath);
    return hr;
}


static HRESULT CacheSetupPackages(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in SETUP_PACKAGES *pSetupPackages
    )
{
    HRESULT hr = S_OK;

    for (DWORD i = 0; i < pSetupPackages->cPackages; ++i)
    {
        SETUP_PACKAGE *pSetupPackage = pSetupPackages->rgPackages + i;


        //
        // should we process the package?
        //

        if (PACKAGE_STATUS_PENDING_CACHE != pSetupPackage->Status)
        {
            continue; // package is not pending preparation
        }


        //
        // send package begin message
        //

        hr = SendSuxCachePackageBeginMessage(pEngineInfo, i);
        ExitOnFailure(hr, "Failed to send package begin message.");


        //
        // process the package
        //

        hr = CacheSetupPackage(pEngineInfo, pSetupPackage, i);
        ExitOnFailure1(hr, "Failed to cache setup package. %d", i);

        //
        // send package complete message
        //

        hr = SendSuxCachePackageCompleteMessage(pEngineInfo, i, pSetupPackage->Status);
        ExitOnFailure(hr, "Failed to send package complete message.");
    }

LExit:
    return hr;
}


static HRESULT CacheSetupPackage(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in SETUP_PACKAGE *pSetupPackage,
    __in DWORD dwPackageIndex
    )
{
    HRESULT hr = S_OK;

    switch (pSetupPackage->SourceType)
    {
    //case PACKAGE_SOURCE_TYPE_RELATIVE:
    //    hr = ResolveRelativePath();
    //    break;
    case PACKAGE_SOURCE_TYPE_EMBEDDED:
        hr = ExtractPackage(pEngineInfo, dwPackageIndex, pSetupPackage->wzSource, pSetupPackage->wzFileName, &pSetupPackage->wzExecutePath);
        if (FAILED(hr))
        {
            TraceError(hr, "Failed to extract package.");

            pSetupPackage->Status = PACKAGE_STATUS_EXTRACTION_FAILED;
            ExitFunction1(hr = S_OK);
        }
        break;
    case PACKAGE_SOURCE_TYPE_DOWNLOAD:
        hr = DownloadPackage(pEngineInfo, dwPackageIndex, pSetupPackage->wzSource, pSetupPackage->wzFileName, &pSetupPackage->wzExecutePath);
        if (FAILED(hr))
        {
            TraceError(hr, "Failed to download package.");

            pSetupPackage->Status = PACKAGE_STATUS_DOWNLOAD_FAILED;
            ExitFunction1(hr = S_OK);
        }
        break;
    }

    // update status
    pSetupPackage->Status = PACKAGE_STATUS_PENDING_EXECUTE;

LExit:
    return hr;
}


static HRESULT ExtractPackage(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in DWORD dwIndex,
    __in LPCWSTR wzSource,
    __in LPCWSTR wzFileName,
    __out LPWSTR* ppwzPackagePath
    )
{
    HRESULT hr = S_OK;
    LPWSTR pwzPackageTempPath = NULL;
    LPSTR pszSource = NULL;

    // create package temp directory path
    hr = CreatePackageTempPath(pEngineInfo, dwIndex, &pwzPackageTempPath);
    ExitOnFailure(hr, "Failed to create package temp derectory path.");

    // create package file path
    hr = StrAllocFormatted(ppwzPackagePath, L"%s%s", pwzPackageTempPath, wzFileName);
    ExitOnFailure(hr, "Failed to create package file path.");

    // create package temp directory
    hr = DirEnsureExists(pwzPackageTempPath, NULL);
    ExitOnFailure(hr, "Failed to create package temp directory.");

    hr = StrAnsiAllocString(&pszSource, wzSource, 0, CP_UTF8);
    ExitOnFailure(hr, "Failed to covert source path to ANSI.");

    // export resource
    hr = ResExportDataToFile(pszSource, *ppwzPackagePath, CREATE_ALWAYS);
    ExitOnFailure1(hr, "Failed to write package to %S.", *ppwzPackagePath);

LExit:
    ReleaseStr(pwzPackageTempPath);
    ReleaseStr(pszSource);
    return hr;
}


static HRESULT DownloadPackage(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in DWORD dwIndex,
    __in LPCWSTR wzSource,
    __in LPCWSTR wzFileName,
    __out LPWSTR* ppwzPackagePath
    )
{
    HRESULT hr = S_OK;

    ExitOnFailure(hr = E_NOTIMPL, "Not yet implemented");

LExit:
    return hr;
}


static HRESULT CreatePackageTempPath(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in DWORD dwIndex,
    __out LPWSTR* ppwzPackageTempPath
    )
{
    HRESULT hr = S_OK;
    LPWSTR pwzTempPath = NULL;
    DWORD cch = 0;
    GUID guid = { 0 };

    if (!pEngineInfo->pwzCachePath)
    {
        hr = StrAlloc(&pwzTempPath, MAX_PATH);
        ExitOnFailure(hr, "Failed to allocate string for temp path.");

        // get temp path
        cch = ::GetTempPathW(MAX_PATH, pwzTempPath);
        if (!cch || cch >= MAX_PATH)
        {
            ExitWithLastError(hr, "Failed to GetTempPath.");
        }

        // create guid to use as directory name
        hr = ::CoCreateGuid(&guid);
        ExitOnFailure(hr, "Failed to create GUID.");

        // format full cache directory name
        hr = StrAllocFormatted(&pEngineInfo->pwzCachePath,
            L"%s%08lx%04x%04x%02x%02x%02x%02x%02x%02x%02x%02x\\",
            pwzTempPath,
            guid.Data1,
            guid.Data2,
            guid.Data3,
            guid.Data4[0],
            guid.Data4[1],
            guid.Data4[2],
            guid.Data4[3],
            guid.Data4[4],
            guid.Data4[5],
            guid.Data4[6],
            guid.Data4[7]
        );
        ExitOnFailure(hr, "Failed to format cache path.");
    }

    // <cache path>\<package index>
    hr = StrAllocFormatted(ppwzPackageTempPath, L"%s%u\\", pEngineInfo->pwzCachePath, dwIndex);
    ExitOnFailure(hr, "Failed to format package temp path.");

LExit:
    ReleaseStr(pwzTempPath);
    return hr;
}


static HRESULT ExecutePackages(
    __in SETUP_PACKAGES *pSetupPackages,
    __in SETUP_ENGINE_INFORMATION *pEngineInfo
    )
{
    HRESULT hr = S_OK;
    DWORD dwCode = 0;
    BOOL fAbort = FALSE;


    //
    // initialize progress
    //

    hr = SendSuxMasterProgressUpdateMessage(pEngineInfo, TRUE, pEngineInfo->qwProgressTotal, 0);
    ExitOnFailure(hr, "Failed to send master progress reset message.");


    //
    // execute packages
    //

    for (DWORD i = 0; i < pSetupPackages->cPackages; ++i)
    {
        SETUP_PACKAGE *pSetupPackage = pSetupPackages->rgPackages + i;


        //
        // should we process the package?
        //

        if (PACKAGE_STATUS_PENDING_EXECUTE != pSetupPackage->Status)
        {
            continue; // package is not pending execution
        }


        //
        // check if we should abort
        //

        hr = SendSuxQueryAbort(pEngineInfo, &fAbort);
        ExitOnFailure(hr, "Failed to send query abort message.");

        if (fAbort)
        {
            pEngineInfo->fUserAbort = TRUE;
            ExitFunction1(hr = S_FALSE);
        }


        //
        // update progress stage
        //

        pEngineInfo->qwProgressStageTotal = pSetupPackage->qwProgressTicks;
        //Trace1(REPORT_STANDARD, "PACKAGE PROGRESS STAGE: %d", (DWORD)pSetupPackage->qwProgressTicks);


        //
        // send package begin message
        //

        hr = SendSuxExecPackageBeginMessage(pEngineInfo, i, pSetupPackage->ActionState, FALSE);
        ExitOnFailure(hr, "Failed to send package begin message.");


        //
        // process the package
        //

        hr = ExecutePackage(pEngineInfo, pSetupPackage, i, &dwCode, &fAbort);
        ExitOnFailure1(hr, "Failed to execute package. %d", i);


        //
        // send package complete message
        //

        hr = SendSuxExecPackageCompleteMessage(pEngineInfo, i, dwCode, pSetupPackage->Status);
        ExitOnFailure(hr, "Failed to send package complete message.");


        //
        // abort here if rollback was triggered
        //

        if (fAbort)
        {
            ExitFunction1(hr = S_FALSE);
        }


        //
        // update progress
        //

        pEngineInfo->qwProgressComplete += pSetupPackage->qwProgressTicks;

        hr = SendSuxMasterProgressUpdateMessage(pEngineInfo, FALSE, pEngineInfo->qwProgressTotal, pEngineInfo->qwProgressComplete);
        ExitOnFailure(hr, "Failed to send master progress update message.");
    }

    hr = S_OK;

LExit:
    return hr;
}


static HRESULT ExecutePackage(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SETUP_PACKAGE *pSetupPackage,
    __in DWORD dwPackageIndex,
    __out LPDWORD pdwCode,
    __out LPBOOL pfAbort
    )
{
    HRESULT hr = S_OK;
    LPWSTR pwzArguments = NULL;
    EXIT_CODE_TYPE exitCodeType = EXIT_CODE_TYPE_NONE;

    // resolve source for relative paths
    if (PACKAGE_SOURCE_TYPE_RELATIVE == pSetupPackage->SourceType)
    {
        // resolve source only if the package is not an MSI package that is being uninstalled
        if (SETUP_PACKAGE_TYPE_MSI != pSetupPackage->PackageType || ACTION_STATE_UNINSTALL != pSetupPackage->ActionState)
        {
            hr = ResolveRelativePath(pEngineInfo, pSetupPackage);
            ExitOnFailure(hr, "Failed to resolve relative source path.");

            if (S_FALSE == hr)
            {
                pSetupPackage->Status = PACKAGE_STATUS_SOURCE_RESOLUTION_FAILED;

                if (pSetupPackage->fVital)
                {
                    // if the package is vital kick off a rollback
                    *pfAbort = TRUE;
                }
                ExitFunction1(hr = S_OK);
            }
        }
    }

    // set global arguments
    //if (wzGlobalArguments)
    //{
    //    hr = StrAllocString(&pwzArguments, wzGlobalArguments, 0);
    //    ExitOnFailure(hr, "Failed to allocate string for global arguments.");
    //}

    switch (pSetupPackage->PackageType)
    {
    case SETUP_PACKAGE_TYPE_EXE:
        LogStringLine(REPORT_STANDARD, "Configuring EXE package. Id = '%S'", pSetupPackage->wzId);

        switch (pSetupPackage->ActionState)
        {
        case ACTION_STATE_INSTALL:
            hr = FormatPropertyString(pEngineInfo, pSetupPackage->Exe.wzArguments, &pwzArguments);
            break;
        case ACTION_STATE_UNINSTALL:
            hr = FormatPropertyString(pEngineInfo, pSetupPackage->Exe.wzUninstallArguments, &pwzArguments);
            break;
        case ACTION_STATE_RECACHE:
            hr = FormatPropertyString(pEngineInfo, pSetupPackage->Exe.wzRepairArguments, &pwzArguments);
            break;
        default:
            ExitFunction1(hr = E_UNEXPECTED);
        }
        ExitOnFailure(hr, "Failed to format exe arguments.");

        hr = ConfigureExe(pSetupPackage->wzExecutePath, pwzArguments, &pSetupPackage->Exe.dwExitCode);
        ExitOnFailure(hr, "Failed to configure EXE package.");

        // log
        LogStringLine(REPORT_STANDARD, "Done configuring EXE package. ReturnCode = %d", pSetupPackage->Exe.dwExitCode);

        // handle exit code
        *pdwCode = pSetupPackage->Exe.dwExitCode;

        for (DWORD i = 0; i < pSetupPackage->Exe.cExitCodes; ++i)
        {
            if (pSetupPackage->Exe.rgExitCodes[i].fWildcard || pSetupPackage->Exe.dwExitCode == pSetupPackage->Exe.rgExitCodes[i].dwExitCode)
            {
                exitCodeType = pSetupPackage->Exe.rgExitCodes[i].Type;
                break;
            }
        }

        switch (exitCodeType)
        {
        case EXIT_CODE_TYPE_SUCCESS:
            pSetupPackage->Status = PACKAGE_STATUS_SUCCEEDED;
            break;
        case EXIT_CODE_TYPE_WARNING:
            pSetupPackage->Status = PACKAGE_STATUS_WARNING;
            break;
        case EXIT_CODE_TYPE_NONE:
            // if no exit code mapping was found above, treate unequal to zero as error
            if (0 == pSetupPackage->Exe.dwExitCode)
            {
                pSetupPackage->Status = PACKAGE_STATUS_SUCCEEDED;
                break;
            }
            __fallthrough;
        case EXIT_CODE_TYPE_ERROR:
            pSetupPackage->Status = PACKAGE_STATUS_FAILED;

            if (pSetupPackage->fVital)
            {
                // if the package is vital kick of a rollback
                *pfAbort = TRUE;
            }
            break;
        default:
            ExitFunction1(hr = E_UNEXPECTED);
        }

        break;

    case SETUP_PACKAGE_TYPE_MSI:
        LogStringLine(REPORT_STANDARD, "Configuring MSI package. Id = '%S'", pSetupPackage->wzId);

        if (ACTION_STATE_UNINSTALL != pSetupPackage->ActionState)
        {
            hr = ConcatMsiFeatureActionArguments(&pSetupPackage->Msi, FALSE, &pwzArguments);
            ExitOnFailure(hr, "Failed to concat feature action arguments.");
        }

        hr = ConcatMsiPropertyArguments(pEngineInfo, &pSetupPackage->Msi, FALSE, &pwzArguments);
        ExitOnFailure(hr, "Failed to concat property arguments.");

        hr = ConfigureMsi(pEngineInfo, dwPackageIndex, pSetupPackage->Msi.guidProductCode, pSetupPackage->wzExecutePath, pwzArguments, pSetupPackage->ActionState, pSetupPackage->Msi.uiLevel, &pSetupPackage->Msi.dwReturnCode);
        ExitOnFailure(hr, "Failed to configure MSI package.");

        // log
        LogStringLine(REPORT_STANDARD, "Done configuring MSI package. ReturnCode = %d", pSetupPackage->Msi.dwReturnCode);

        // handle return code
        *pdwCode = pSetupPackage->Msi.dwReturnCode;
        if (ERROR_INSTALL_USEREXIT == pSetupPackage->Msi.dwReturnCode)
        {
            pSetupPackage->Status = PACKAGE_STATUS_ABORTED;

            // kick of a rollback
            *pfAbort = TRUE;
            ExitFunction1(hr = S_OK);
        }
        else if (ERROR_SUCCESS != pSetupPackage->Msi.dwReturnCode)
        {
            pSetupPackage->Status = PACKAGE_STATUS_FAILED;

            if (pSetupPackage->fVital)
            {
                // if the package is vital kick of a rollback
                *pfAbort = TRUE;
            }
            ExitFunction1(hr = S_OK);
        }

        // update package status
        pSetupPackage->Status = PACKAGE_STATUS_SUCCEEDED;

        // check if we had a user abort request that was not honored by the package
        if (pEngineInfo->fUserAbort)
        {
            // kick of a rollback
            *pfAbort = TRUE;
            ExitFunction1(hr = S_OK);
        }

        break;

    default:
        hr = E_NOTIMPL;
        ExitOnFailure(hr, "Have not implemented other package type support yet");
    }

LExit:
    ReleaseStr(pwzArguments);
    return hr;
}


static HRESULT RollbackPackages(
    __in SETUP_PACKAGES *pSetupPackages,
    __in SETUP_ENGINE_INFORMATION *pEngineInfo
    )
{
    HRESULT hr = S_OK;
    DWORD dwCode = 0;


    //
    // rollback packages
    //

    // TODO: start with the last executed package
    for (int i = pSetupPackages->cPackages - 1; i >= 0; --i)
    {
        SETUP_PACKAGE *pSetupPackage = pSetupPackages->rgPackages + i;

        // only packages that are in the succeeded state - other packages has not been changed
        if (PACKAGE_STATUS_SUCCEEDED != pSetupPackage->Status)
        {
            continue;
        }

        // there is never any case where we want to roll back a permantent package
        if (pSetupPackage->fPermanent)
        {
            continue;
        }


        //
        // update progress stage
        //

        pEngineInfo->qwProgressComplete = pSetupPackage->qwProgressOffset + pSetupPackage->qwProgressTicks;
        pEngineInfo->qwProgressStageTotal = pSetupPackage->qwProgressTicks;


        //
        // send package begin message
        //

        hr = SendSuxExecPackageBeginMessage(pEngineInfo, i, pSetupPackage->RollbackAction, TRUE);
        ExitOnFailure(hr, "Failed to send package begin message.");


        //
        // rollback package
        //

        hr = RollbackPackage(pEngineInfo, pSetupPackage, i, &dwCode);
        ExitOnFailure1(hr, "Failed to rollback package. %d", i);


        //
        // update progress
        //

        pEngineInfo->qwProgressComplete = pSetupPackage->qwProgressOffset;

        //hr = SendProgressUpdate(pEngineInfo, pEngineInfo->qwProgressTotal, pEngineInfo->qwProgressComplete);
        hr = SendSuxMasterProgressUpdateMessage(pEngineInfo, FALSE, pEngineInfo->qwProgressTotal, pEngineInfo->qwProgressComplete);
        ExitOnFailure(hr, "Failed to send master progress update message.");


        //
        // send package complete message
        //

        hr = SendSuxExecPackageCompleteMessage(pEngineInfo, i, dwCode, pSetupPackage->Status);
        ExitOnFailure(hr, "Failed to send package complete message.");
    }

LExit:
    return hr;
}


static HRESULT RollbackPackage(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in SETUP_PACKAGE *pSetupPackage,
    __in DWORD dwPackageIndex,
    __out LPDWORD pdwCode
    )
{
    HRESULT hr = S_OK;
    LPWSTR pwzArguments = NULL;

    // resolve source for relative paths
    if (PACKAGE_SOURCE_TYPE_RELATIVE == pSetupPackage->SourceType)
    {
        hr = ResolveRelativePath(pEngineInfo, pSetupPackage);
        ExitOnFailure(hr, "Failed to resolve relative source path.");

        if (S_FALSE == hr)
        {
            pSetupPackage->Status = PACKAGE_STATUS_ROLLBACK_FAILED; // TODO: should we have higher granularity here?
            ExitFunction1(hr = S_OK);
        }
    }

    switch (pSetupPackage->PackageType)
    {
    case SETUP_PACKAGE_TYPE_MSI:
        LogStringLine(REPORT_STANDARD, "Rolling back MSI package. Id = '%S'", pSetupPackage->wzId);

        // if we don't have a rollback action it is because we did an upgrade of the package
        if (ACTION_STATE_NONE != pSetupPackage->RollbackAction)
        {
            if (ACTION_STATE_UNINSTALL != pSetupPackage->RollbackAction)
            {
                hr = ConcatMsiFeatureActionArguments(&pSetupPackage->Msi, TRUE, &pwzArguments);
                ExitOnFailure(hr, "Failed to concat feature action arguments.");
            }

            hr = ConcatMsiPropertyArguments(pEngineInfo, &pSetupPackage->Msi, TRUE, &pwzArguments);
            ExitOnFailure(hr, "Failed to concat property arguments.");

            hr = ConfigureMsi(pEngineInfo, dwPackageIndex, pSetupPackage->Msi.guidProductCode, pSetupPackage->wzExecutePath, pwzArguments, pSetupPackage->RollbackAction, pSetupPackage->Msi.uiLevel, &pSetupPackage->Msi.dwReturnCode);
            ExitOnFailure(hr, "Failed to configure MSI package.");

            *pdwCode = pSetupPackage->Msi.dwReturnCode;
            if (ERROR_SUCCESS != pSetupPackage->Msi.dwReturnCode)
            {
                pSetupPackage->Status = PACKAGE_STATUS_ROLLBACK_FAILED;
            }
            else
            {
                pSetupPackage->Status = PACKAGE_STATUS_ROLLBACK_SUCCEEDED;
            }
        }
        else
        {
            *pdwCode = ERROR_SUCCESS;
            pSetupPackage->Status = PACKAGE_STATUS_ROLLBACK_NODOWNGRADE;
        }

        LogStringLine(REPORT_STANDARD, "Done rolling back MSI package. ReturnCode = %d", *pdwCode);
        break;
    case SETUP_PACKAGE_TYPE_EXE:
        LogStringLine(REPORT_STANDARD, "Rolling back EXE package. Id = '%S'", pSetupPackage->wzId);

        switch (pSetupPackage->RollbackAction)
        {
        case ACTION_STATE_INSTALL:
            hr = StrAllocString(&pwzArguments, pSetupPackage->Exe.wzArguments, 0);
            ExitOnFailure(hr, "Failed to copy exe arguments.");
            break;
        case ACTION_STATE_UNINSTALL:
            hr = StrAllocString(&pwzArguments, pSetupPackage->Exe.wzUninstallArguments, 0);
            ExitOnFailure(hr, "Failed to copy exe arguments.");
            break;
        default:
            ExitFunction1(hr = E_UNEXPECTED);
        }

        hr = ConfigureExe(pSetupPackage->wzExecutePath, pwzArguments, &pSetupPackage->Exe.dwExitCode);
        ExitOnFailure(hr, "Failed to configure MSI package.");

        *pdwCode = pSetupPackage->Exe.dwExitCode;
        if (0 != pSetupPackage->Exe.dwExitCode)
        {
            pSetupPackage->Status = PACKAGE_STATUS_ROLLBACK_FAILED;
        }
        else
        {
            pSetupPackage->Status = PACKAGE_STATUS_ROLLBACK_SUCCEEDED;
        }

        LogStringLine(REPORT_STANDARD, "Done rolling back EXE package. ReturnCode = %d", *pdwCode);
        break;
    default:
        hr = E_NOTIMPL;
        ExitOnFailure(hr, "Have not implemented other package type support yet");
    }

LExit:
    ReleaseStr(pwzArguments);
    return hr;
}


static HRESULT ResolveRelativePath(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __inout SETUP_PACKAGE* pSetupPackage
    )
{
    HRESULT hr = S_OK;
    LPWSTR pwzPath = NULL;
    LPWSTR pwzFileName = NULL;
    LPWSTR pwzBasePath = NULL;
    BOOL fPathProvided = FALSE;

    // if the package already has a path, test it
    if (pSetupPackage->wzExecutePath && *pSetupPackage->wzExecutePath)
    {
        hr = ValidateSourcePath(pSetupPackage, pSetupPackage->wzExecutePath);
        if (S_OK == hr)
        {
            ExitFunction1(hr = S_OK);
        }
        if (FAILED(hr))
        {
            TraceError1(hr, "Failed to validate source path. '%S'", pwzPath);
        }
    }

    // test known paths
    for (DWORD i = 0; i < pEngineInfo->cSourcePaths; i++)
    {
        SETUP_SOURCE_PATH* pSourcePath = &pEngineInfo->rgSourcePaths[i];

        // create absolute path
        hr = StrAllocFormatted(&pwzPath, L"%s%s", pSourcePath->wzPath, pSetupPackage->wzSource);
        ExitOnFailure(hr, "Failed to format path name.");

        // validate source path
        hr = ValidateSourcePath(pSetupPackage, pwzPath);
        if (S_OK == hr)
        {
            hr = StrAllocString(&pSetupPackage->wzExecutePath, pwzPath, 0);
            ExitOnFailure(hr, "Failed to allocate copy path string.");

            ExitFunction1(hr = S_OK);
        }
        if (FAILED(hr))
        {
            TraceError1(hr, "Failed to validate source path. '%S'", pwzPath);
        }
    }

    // query user for source path
    do {
        // get only the file name partition of the relative path
        pwzFileName = FileFromPath(pSetupPackage->wzSource);

        // send source resolution message
        ReleaseNullStr(pwzPath);
        fPathProvided = FALSE;
        hr = SendSuxSourceResolutionMessage(pEngineInfo, pwzFileName, &pwzPath, &fPathProvided);
        ExitOnFailure(hr, "Failed to send source resolution message.");

        if (!fPathProvided)
        {
            // a source path was not provided, unable to resolve source
            ExitFunction1(hr = S_FALSE);
        }

        // validate source path
        hr = ValidateSourcePath(pSetupPackage, pwzPath);
        if (FAILED(hr))
        {
            TraceError(hr, "Failed to validate source path.");
            hr = S_FALSE;
        }
    } while (S_FALSE == hr); // leave loop only if hr == S_OK

    // if the path ends with the whole relative path, add the base to the source path list
    hr = GetBasePathForRelativePath(pwzPath, pSetupPackage->wzSource, &pwzBasePath);
    if (FAILED(hr))
    {
        TraceError(hr, "Failed to get base for source path.");
    }
    else if (S_OK == hr)
    {
        // add to source path list
        hr = AddSourcePath(pEngineInfo, pwzBasePath);
        if (FAILED(hr))
        {
            TraceError(hr, "Failed to get base for source path.");
        }
    }

    hr = StrAllocString(&pSetupPackage->wzExecutePath, pwzPath, 0);
    ExitOnFailure(hr, "Failed to allocate copy path string.");

    hr = S_OK;

LExit:
    ReleaseStr(pwzPath);
    ReleaseStr(pwzBasePath);
    return hr;
}


static HRESULT ValidateSourcePath(
    __in SETUP_PACKAGE* pSetupPackage,
    __in LPCWSTR wzPath
    )
{
    HRESULT hr = S_OK;
    GUID guidProductCode = { 0 };
    DWORD64 qwProductVersion = 0;
    HANDLE hFile = NULL;
    LARGE_INTEGER liFileSize = { 0 };

    // test if the file exists
    if (!FileExistsEx(wzPath, NULL))
    {
        ExitFunction1(hr = S_FALSE);
    }

    switch (pSetupPackage->PackageType)
    {
    case SETUP_PACKAGE_TYPE_MSI:
        // check product code and version
        hr = GetMsiProductCodeAndVersion(wzPath, &guidProductCode, &qwProductVersion);
        ExitOnFailure(hr, "Failed to get MSI product code and version.");

        if (!::IsEqualGUID(guidProductCode, pSetupPackage->Msi.guidProductCode) || qwProductVersion != pSetupPackage->Msi.qwVersion)
        {
            ExitFunction1(hr = S_FALSE);
        }

        break;
    case SETUP_PACKAGE_TYPE_EXE:
        // check file size
        if (pSetupPackage->cbFileSize)
        {
            hFile = ::CreateFileW(wzPath, GENERIC_READ, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
            ExitOnNullWithLastError1(hFile, hr, "Failed to open file. '%S'", wzPath);

            if (!::GetFileSizeEx(hFile, &liFileSize))
            {
                ExitWithLastError(hr, "Failed to get file size.");
            }

            if (liFileSize.QuadPart != (LONGLONG)pSetupPackage->cbFileSize)
            {
                ExitFunction1(hr = S_FALSE);
            }
        }
        break;
    default:
        ExitFunction1(hr = E_UNEXPECTED);
    }

    hr = S_OK;

LExit:
    if (hFile)
    {
        ::CloseHandle(hFile);
    }
    return hr;
}


static HRESULT GetMsiProductCodeAndVersion(
    __in LPCWSTR wzMsiPath,
    __out LPGUID pguidProductCode,
    __out DWORD64* pqwProductVersion
    )
{
    HRESULT hr = S_OK;
    DWORD er = ERROR_SUCCESS;
    MSIHANDLE hDatabase = NULL;
    MSIHANDLE hView = NULL;
    MSIHANDLE hRec = NULL;
    WCHAR wzProductCode[39] = { 0 };    // '{00000000-0000-0000-0000-000000000000}\0'
    WCHAR wzProductVersion[24] = { 0 }; // '65535.65535.65535.65535\0'
    DWORD cch = 0;
    ULARGE_INTEGER uliProductVersion = { 0 };

    // open a view on the Property table.
    er = ::MsiOpenDatabaseW(wzMsiPath, (LPCWSTR)MSIDBOPEN_READONLY, &hDatabase);
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to open MSI database");

    er = ::MsiDatabaseOpenViewW(hDatabase, L"SELECT `Value` FROM `Property` WHERE `Property`=?", &hView);
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to open view on Property table.");

    hRec = ::MsiCreateRecord(1);
    ExitOnNull(hRec, hr, E_OUTOFMEMORY, "Failed to create record.");

    // Get the ProductCode.
    er = ::MsiRecordSetStringW(hRec, 1, L"ProductCode");
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to set string 'ProductCode' into record.");

    er = ::MsiViewExecute(hView, hRec);
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to execute view to get ProductCode.");

    er = ::MsiViewFetch(hView, &hRec);
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to fetch 'ProductCode' from view.");

    cch = countof(wzProductCode);
    er = ::MsiRecordGetStringW(hRec, 1, wzProductCode, &cch);
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to fetch 'ProductCode' from record.");

    hr = ::IIDFromString(wzProductCode, pguidProductCode);
    ExitOnFailure1(hr, "Failed to parse product code. '%S'", wzProductCode);

    // Get the ProductVersion.
    er = ::MsiRecordSetStringW(hRec, 1, L"ProductVersion");
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to set string 'ProductVersion' into record.");

    er = ::MsiViewExecute(hView, hRec);
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to execute view to get ProductVersion");

    er = ::MsiViewFetch(hView, &hRec);
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to fetch 'ProductVersion' from view.");

    cch = countof(wzProductVersion);
    er = ::MsiRecordGetStringW(hRec, 1, wzProductVersion, &cch);
    hr = HRESULT_FROM_WIN32(er);
    ExitOnFailure(hr, "Failed to fetch 'ProductVersion' from record.");

    hr = FileVersionFromString(wzProductVersion, &uliProductVersion.HighPart, &uliProductVersion.LowPart);
    ExitOnFailure1(hr, "Failed to parse product version. '%S'", wzProductVersion);
    *pqwProductVersion = uliProductVersion.QuadPart;

LExit:
    if (hRec)
    {
        ::MsiCloseHandle(hRec);
    }
    if (hView)
    {
        ::MsiCloseHandle(hView);
    }
    if (hDatabase)
    {
        ::MsiCloseHandle(hDatabase);
    }
    return hr;
}


static HRESULT GetBasePathForRelativePath(
    __in LPCWSTR wzFullPath,
    __in LPCWSTR wzRelativePath,
    __out LPWSTR* ppwzBasePath
    )
{
    HRESULT hr = S_OK;

    int cchFullPath = lstrlenW(wzFullPath);
    int cchRelativePath = lstrlenW(wzRelativePath);
    int cchBasePath = cchFullPath - cchRelativePath;

    if (0 >= cchBasePath)
    {
        ExitFunction1(hr = S_FALSE);
    }

    // TODO: what if source path is '..\product.msi' or 'dir\.\product.msi' or 'dir/product.msi'? we shouldn't work directly with strings here...
    if (CSTR_EQUAL != ::CompareStringW(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE, wzFullPath + cchBasePath, cchRelativePath, wzRelativePath, cchRelativePath))
    {
        ExitFunction1(hr = S_FALSE);
    }

    hr = StrAllocString(ppwzBasePath, wzFullPath, cchBasePath);
    ExitOnFailure(hr, "Failed to allocate base path string.");

    hr = S_OK;

LExit:
    return hr;
}


static HRESULT GetWorkerMessage(
    __in HANDLE hPipe,
    __in SETUP_COMM_MESSAGE *pmsg
    )
{
    HRESULT hr = S_OK;
    DWORD cbRead = 0;
    LPVOID pv = NULL;

    // Blank out the message in case anything goes wrong below.
    memset(pmsg, 0, sizeof(*pmsg));

    // Read the pmsg->dwMessage (first DWORD) and pmsg->cbData (second DWORD)
    // for the message out of the pipe.
    if (!::ReadFile(hPipe, pmsg, sizeof(DWORD) * 2, &cbRead, NULL))
    {
        ExitWithLastError(hr, "Failed to read message from pipe.");
    }

    // Finally, if there is data for this message then read all that.
    if (pmsg->cbData)
    {
        pv = MemAlloc(pmsg->cbData, TRUE);
        ExitOnNull(pv, hr, E_OUTOFMEMORY, "Failed to allocate memory for data in pipe.");

        if (!::ReadFile(hPipe, pv, pmsg->cbData, &cbRead, NULL))
        {
            ExitWithLastError(hr, "Failed to read extra data from pipe.");
        }

        pmsg->fAllocatedData = TRUE;
        pmsg->pvData = pv;
        pv = NULL;
    }

LExit:
    ReleaseMem(pv);

    return hr;
}


static HRESULT ProcessMessage(
    __in SETUP_ENGINE_INFORMATION *pEngineInfo,
    __in SETUP_COMM_MESSAGE *pMsg
    )
{
    HRESULT hr = S_OK;
    CHAIN_COMPLETE_STRUCT ChainComplete = { 0 };

    switch (pMsg->dwMessage)
    {
    case SWM_PREPARE:
        SendSuxMessage(pEngineInfo, SRM_PREPARE_CHAIN_BEGIN, NULL);

        hr = DoRegistrySearches(pEngineInfo);
        ExitOnFailure(hr, "Failed to do registry searches.");

        hr = PrepareSetupPackages(pEngineInfo, pEngineInfo->pSetupPackages);
        ChainComplete.hResult = hr;

        SendSuxMessage(pEngineInfo, SRM_PREPARE_CHAIN_COMPLETE, (PVOID)&ChainComplete);
        ExitOnFailure(hr, "Failed to detect packages.");
        break;

    case SWM_CACHE:
        SendSuxMessage(pEngineInfo, SRM_CACHE_CHAIN_BEGIN, NULL);

        hr = CacheSetupPackages(pEngineInfo, pEngineInfo->pSetupPackages);
        ChainComplete.hResult = hr;
        SendSuxMessage(pEngineInfo, SRM_CACHE_CHAIN_COMPLETE, (PVOID)&ChainComplete);
        ExitOnFailure(hr, "Failed to prepare packages.");
        break;

    case SWM_EXECUTE:
        SendSuxMessage(pEngineInfo, SRM_EXEC_CHAIN_BEGIN, NULL);

        pEngineInfo->fUserAbort = FALSE;
        pEngineInfo->fRollback = FALSE;

        hr = ExecutePackages(pEngineInfo->pSetupPackages, pEngineInfo);
        if (S_FALSE == hr) // TODO: no rollback on repair
        {
            pEngineInfo->fRollback = TRUE;
            hr = RollbackPackages(pEngineInfo->pSetupPackages, pEngineInfo);
        }
        ChainComplete.hResult = hr;
        SendSuxMessage(pEngineInfo, SRM_EXEC_CHAIN_COMPLETE, (PVOID)&ChainComplete);
        ExitOnFailure(hr, "Failed to process packages.");

        if (pEngineInfo->fLogInitialized)
        {
            LogUninitialize(TRUE);
        }
        pEngineInfo->fLogInitialized = FALSE;
        break;

    default:
        //LRESULT lResult = SendSuxMessage(pEngineInfo, SRM_ECHO, (WPARAM)pMsg->pvData, 0); // just send the message data back.
        //Trace1(REPORT_STANDARD, "Result from the special message is: 0x%x", lResult);
        break;
    }

LExit:
    return hr;
}


static void FreeWorkerMessage(
    __in SETUP_COMM_MESSAGE *pmsg
    )
{
    if (pmsg->fAllocatedData && pmsg->pvData)
    {
        ReleaseNullMem(pmsg->pvData);
        pmsg->fAllocatedData = FALSE;
    }
}


static HRESULT ConcatMsiFeatureActionArguments(
    __in SETUP_PACKAGE_MSI* pMsiPackage,
    __in BOOL fRollback,
    __inout LPWSTR* ppwzArguments
    )
{
    HRESULT hr = S_OK;
    LPWSTR pwz = NULL;
    LPWSTR pwzAddLocal = NULL;
    LPWSTR pwzAddSource = NULL;
    LPWSTR pwzAddDefault = NULL;
    LPWSTR pwzReinstall = NULL;
    LPWSTR pwzAdvertise = NULL;
    LPWSTR pwzRemove = NULL;

    // features
    for (DWORD i = 0; i < pMsiPackage->cFeatures; ++i)
    {
        SETUP_PACKAGE_MSI_FEATURE* pFeature = &pMsiPackage->rgFeatures[i];

        switch (fRollback ? pFeature->RollbackAction : pFeature->ExecuteAction)
        {
        case FEATURE_ACTION_ADDLOCAL:
            if (pwzAddLocal)
            {
                hr = StrAllocConcat(&pwzAddLocal, L",", 0);
                ExitOnFailure(hr, "Failed to concat separator.");
            }
            hr = StrAllocConcat(&pwzAddLocal, pFeature->wzId, 0);
            ExitOnFailure(hr, "Failed to concat feature.");
            break;
        case FEATURE_ACTION_ADDSOURCE:
            if (pwzAddSource)
            {
                hr = StrAllocConcat(&pwzAddSource, L",", 0);
                ExitOnFailure(hr, "Failed to concat separator.");
            }
            hr = StrAllocConcat(&pwzAddSource, pFeature->wzId, 0);
            ExitOnFailure(hr, "Failed to concat feature.");
            break;
        case FEATURE_ACTION_ADDDEFAULT:
            if (pwzAddDefault)
            {
                hr = StrAllocConcat(&pwzAddDefault, L",", 0);
                ExitOnFailure(hr, "Failed to concat separator.");
            }
            hr = StrAllocConcat(&pwzAddDefault, pFeature->wzId, 0);
            ExitOnFailure(hr, "Failed to concat feature.");
            break;
        case FEATURE_ACTION_REINSTALL:
            if (pwzReinstall)
            {
                hr = StrAllocConcat(&pwzReinstall, L",", 0);
                ExitOnFailure(hr, "Failed to concat separator.");
            }
            hr = StrAllocConcat(&pwzReinstall, pFeature->wzId, 0);
            ExitOnFailure(hr, "Failed to concat feature.");
            break;
        case FEATURE_ACTION_ADVERTISE:
            if (pwzAdvertise)
            {
                hr = StrAllocConcat(&pwzAdvertise, L",", 0);
                ExitOnFailure(hr, "Failed to concat separator.");
            }
            hr = StrAllocConcat(&pwzAdvertise, pFeature->wzId, 0);
            ExitOnFailure(hr, "Failed to concat feature.");
            break;
        case FEATURE_ACTION_REMOVE:
            if (pwzRemove)
            {
                hr = StrAllocConcat(&pwzRemove, L",", 0);
                ExitOnFailure(hr, "Failed to concat separator.");
            }
            hr = StrAllocConcat(&pwzRemove, pFeature->wzId, 0);
            ExitOnFailure(hr, "Failed to concat feature.");
            break;
        }
    }

    if (pwzAddLocal)
    {
        hr = StrAllocFormatted(&pwz, L" ADDLOCAL=\"%s\"", pwzAddLocal, 0);
        ExitOnFailure(hr, "Failed to format ADDLOCAL string.");

        hr = StrAllocConcat(ppwzArguments, pwz, 0);
        ExitOnFailure(hr, "Failed to concat argument string.");
    }
    if (pwzAddSource)
    {
        hr = StrAllocFormatted(&pwz, L" ADDSOURCE=\"%s\"", pwzAddSource, 0);
        ExitOnFailure(hr, "Failed to format ADDSOURCE string.");

        hr = StrAllocConcat(ppwzArguments, pwz, 0);
        ExitOnFailure(hr, "Failed to concat argument string.");
    }
    if (pwzAddDefault)
    {
        hr = StrAllocFormatted(&pwz, L" ADDDEFAULT=\"%s\"", pwzAddDefault, 0);
        ExitOnFailure(hr, "Failed to format ADDDEFAULT string.");

        hr = StrAllocConcat(ppwzArguments, pwz, 0);
        ExitOnFailure(hr, "Failed to concat argument string.");
    }
    if (pwzReinstall)
    {
        hr = StrAllocFormatted(&pwz, L" REINSTALL=\"%s\"", pwzReinstall, 0);
        ExitOnFailure(hr, "Failed to format REINSTALL string.");

        hr = StrAllocConcat(ppwzArguments, pwz, 0);
        ExitOnFailure(hr, "Failed to concat argument string.");
    }
    if (pwzAdvertise)
    {
        hr = StrAllocFormatted(&pwz, L" ADVERTISE=\"%s\"", pwzAdvertise, 0);
        ExitOnFailure(hr, "Failed to format ADVERTISE string.");

        hr = StrAllocConcat(ppwzArguments, pwz, 0);
        ExitOnFailure(hr, "Failed to concat argument string.");
    }
    if (pwzRemove)
    {
        hr = StrAllocFormatted(&pwz, L" REMOVE=\"%s\"", pwzRemove, 0);
        ExitOnFailure(hr, "Failed to format REMOVE string.");

        hr = StrAllocConcat(ppwzArguments, pwz, 0);
        ExitOnFailure(hr, "Failed to concat argument string.");
    }

LExit:
    ReleaseStr(pwz);
    ReleaseStr(pwzAddLocal);
    ReleaseStr(pwzAddSource);
    ReleaseStr(pwzAddDefault);
    ReleaseStr(pwzReinstall);
    ReleaseStr(pwzAdvertise);
    ReleaseStr(pwzRemove);
    return hr;
}


static HRESULT ConcatMsiPropertyArguments(
    __in SETUP_ENGINE_INFORMATION* pEngineInfo,
    __in SETUP_PACKAGE_MSI* pMsiPackage,
    __in BOOL fRollback,
    __inout LPWSTR* ppwzArguments
    )
{
    HRESULT hr = S_OK;
    LPWSTR pwzFormattedValue = NULL;
    LPWSTR pwz = NULL;

    for (DWORD i = 0; i < pMsiPackage->cProperties; ++i)
    {
        SETUP_PACKAGE_MSI_PROPERTY* pProperty = &pMsiPackage->rgProperties[i];
        LPCWSTR wzValue = fRollback ? pProperty->wzExpectedValue : pProperty->wzRequestedValue;

        if (wzValue)
        {
            // format property value
            hr = FormatPropertyString(pEngineInfo, wzValue, &pwzFormattedValue);
            ExitOnFailure(hr, "Failed to format property value.");

            // append to argument string
            // TODO: escape quotation marks
            hr = StrAllocFormatted(&pwz, L" %s=\"%s\"", pProperty->wzId, pwzFormattedValue, 0);
            ExitOnFailure(hr, "Failed to format parameter string.");

            hr = StrAllocConcat(ppwzArguments, pwz, 0);
            ExitOnFailure(hr, "Failed to concat parameter to parameters string.");
        }
    }

LExit:
    ReleaseStr(pwzFormattedValue);
    ReleaseStr(pwz);
    return hr;
}


#if 0
HRESULT ProcessSetupChain(
    __in_opt LPCWSTR wzMsiArgs,
    __in_opt DWORD dwLocaleId
    )
{
    HRESULT hr = S_OK;

    SETUP_INSTALL_CHAIN* pSetupChain = NULL;
    DWORD cbSetupChain = 0;

    SETUP_ROLLBACK_PACKAGE* prgRollbackPackages = NULL;
    BYTE cRollbackPackages = 0;

    LPWSTR pwzTempMsiPath = NULL;
    LPWSTR pwzTempMstPath = NULL;
    LPWSTR pwzCachedMsiPath = NULL;
    LPCWSTR wzMsiPath;
    LPVOID pvErrorMessage = NULL;

    hr = ResReadData("MANIFEST", (LPVOID*)&pSetupChain, &cbSetupChain);
    ExitOnFailure(hr, "Failed to get manifest for setup chain.");

    // Create enough space to track rollback information for all possible packages in the chain.
    prgRollbackPackages = static_cast<SETUP_ROLLBACK_PACKAGE*>(MemAlloc(sizeof(SETUP_ROLLBACK_PACKAGE) * pSetupChain->cPackages, TRUE));
    ExitOnNull(prgRollbackPackages, hr, E_OUTOFMEMORY, "Failed to allocate rollback chain.");

    // If we have transforms, extract them to temp path
    for(BYTE j = 0; j < pSetupChain->cTransforms; ++j)
    {
        SETUP_INSTALL_TRANSFORM* pTransform = pSetupChain->rgTransforms + j;

        // Compare Locales
        if(MatchLanguage(dwLocaleId , pTransform->dwLocaleId))
        {
            // Get the MST out of the setup.exe.
            hr = ExtractMst(pTransform, &pwzTempMstPath);
            ExitOnFailure(hr, "Failed to extract transform to temp path.");
        }
    }

    for (BYTE i = 0; i < pSetupChain->cPackages; ++i)
    {
        CONFIGURE_ACTION action = CONFIGURE_ACTION_INSTALL; // assume this will be a simple install
        BOOL bDeleteTempPath = FALSE;

        SETUP_INSTALL_PACKAGE* pPackage = pSetupChain->rgPackages + i;
        SETUP_ROLLBACK_PACKAGE* pRollbackPackage = prgRollbackPackages + cRollbackPackages;

        hr = ::StringFromGUID2(pPackage->guidProductCode, pRollbackPackage->wzProductCode, countof(pRollbackPackage->wzProductCode));
        ExitOnFailure(hr, "Failed to convert ProductCode GUID to string.");

        //
        // Try to detect the MSI to determine what action is necessary.
        //
        hr = DetectMsi(pRollbackPackage->wzProductCode, pPackage->dwVersionMajor, pPackage->dwVersionMinor, &action);
        ExitOnFailure(hr, "Failed during MSI detection.");

        //
        // If there is no action, move on to the next package.  If we need to do a minor upgrade
        // but the package isn't configured to allow minor upgrades, bail
        //
        if (CONFIGURE_ACTION_NONE == action)
        {
            continue;
        }
        else if (CONFIGURE_ACTION_MINOR_UPGRADE == action && !(pPackage->dwAttributes & SETUP_INSTALL_CHAIN_MINOR_UPGRADE_ALLOWED))
        {
            hr = HRESULT_FROM_WIN32(ERROR_PRODUCT_VERSION);
            DWORD cchErrorMessage = ::FormatMessageW(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM, NULL, hr, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPWSTR)&pvErrorMessage, 0, NULL);
            if (0 < cchErrorMessage)
            {
                ::MessageBoxExW(NULL, (LPWSTR)pvErrorMessage, L"Install Error", MB_ICONERROR | MB_OK, 0);
            }
            ExitOnFailure(hr, "Cannot upgrade previously installed package.");
        }

        ++cRollbackPackages;

        // Get the MSI out of the setup.exe.
        hr = ExtractMsi(pPackage, &pwzTempMsiPath);
        ExitOnFailure(hr, "Failed to extract MSI to temp path.");

        bDeleteTempPath = TRUE; // extracted temp MSI path, so we'll need to clean it up at the end of this loop

        hr = ::StringCchCopyW(pRollbackPackage->wzTempPath, countof(pRollbackPackage->wzTempPath), pwzTempMsiPath);
        ExitOnFailure(hr, "Failed to copy temp path into rollback chain.");

        // If use transform is set for this package, and transform exists use transform
        if(pwzTempMstPath && (pPackage->dwAttributes & SETUP_INSTALL_CHAIN_USE_TRANSFORM))
        {
            // Apply transform to database
            hr = ApplyTransform(pwzTempMsiPath, pwzTempMstPath);
        }

        if (pPackage->dwAttributes & SETUP_INSTALL_CHAIN_CACHE)
        {
            // Cache the package to our application cache.
            hr = CacheMsi(pPackage, pwzTempMsiPath, pPackage->dwAttributes & SETUP_INSTALL_CHAIN_ALLUSERS, &pwzCachedMsiPath);
            ExitOnFailure(hr, "Failed to cache MSI file.");

            bDeleteTempPath = FALSE; // temp MSI has been cached (moved), so we shouldn't delete the temp file any longer.

            wzMsiPath = pwzCachedMsiPath;

            hr = ::StringCchCopyW(pRollbackPackage->wzCachedPath, countof(pRollbackPackage->wzCachedPath), pwzCachedMsiPath);
            ExitOnFailure(hr, "Failed to copy cached path into rollback chain.");

            pRollbackPackage->bCached = TRUE;
        }
        else
        {
            wzMsiPath = pwzTempMsiPath;
        }

        // Install the cached package.
        hr = ConfigureMsi(pRollbackPackage->wzProductCode, wzMsiPath, wzMsiArgs, action, pPackage->dwAttributes & SETUP_INSTALL_CHAIN_SHOW_UI, pPackage->dwAttributes & SETUP_INSTALL_CHAIN_IGNORE_FAILURES);
        ExitOnFailure1(hr, "Failed to launch and wait for cached MSI: %ls", wzMsiPath);

        if (S_FALSE == hr)
        {
            if (pRollbackPackage->bCached)
            {
                ::DeleteFileW(pRollbackPackage->wzCachedPath);

                *pRollbackPackage->wzCachedPath = L'\0';
                pRollbackPackage->bCached = FALSE;
            }
        }
        else
        {
            pRollbackPackage->bInstalled = TRUE;
        }

        if (bDeleteTempPath)
        {
            ::DeleteFileW(pwzTempMsiPath);
        }
    }

LExit:
    if (FAILED(hr) && 
        HRESULT_FROM_WIN32(ERROR_SUCCESS_REBOOT_REQUIRED) != hr &&
        HRESULT_FROM_WIN32(ERROR_SUCCESS_RESTART_REQUIRED) != hr &&
        HRESULT_FROM_WIN32(ERROR_SUCCESS_REBOOT_INITIATED) != hr)
    {
        for (BYTE i = 0; i < cRollbackPackages; ++i)
        {
            SETUP_ROLLBACK_PACKAGE* pRollbackPackage = prgRollbackPackages + i;

            if (pRollbackPackage->bInstalled)
            {
                ConfigureMsi(pRollbackPackage->wzProductCode, NULL, NULL, CONFIGURE_ACTION_UNINSTALL, FALSE, TRUE);
            }

            if (pRollbackPackage->wzCachedPath && *pRollbackPackage->wzCachedPath)
            {
                ::DeleteFileW(pRollbackPackage->wzCachedPath);
            }

            if (pRollbackPackage->wzTempPath && *pRollbackPackage->wzTempPath)
            {
                ::DeleteFileW(pRollbackPackage->wzTempPath);
            }
        }
    }

    if (prgRollbackPackages)
    {
        MemFree(prgRollbackPackages);
    }

    if (pvErrorMessage)
    {
        ::LocalFree(pvErrorMessage);
    }
    ReleaseStr(pwzTempMsiPath);
    ReleaseStr(pwzCachedMsiPath);

    //UISetupComplete(hr);

    return hr;
}
#endif

BOOL MatchLanguage(
    __in DWORD dwLang,
    __in DWORD dwTransformLang
    )
{
    if (INVALID_LCID == dwLang)
    {
        LANGID sysLangID = ::GetSystemDefaultUILanguage();
        dwLang = MAKELCID(sysLangID, SORT_DEFAULT);
    }

    // Convert Languages as appropriate
    switch(PRIMARYLANGID(dwLang))
    {
        case 0x04:
            switch(dwLang)
            {
                // Match Simplified Chinese
                case 0x0804: __fallthrough;
                case 0x1004: __fallthrough;
                case 0x1404:
                    dwTransformLang = 0x0804;
                    break;

                // Match Traditional Chinese
                case 0x0404: __fallthrough;
                case 0x0c04:
                    dwTransformLang = 0x0404;
                    break;
            }
            return dwLang == dwTransformLang;
            break;

        default:
            dwLang = PRIMARYLANGID(dwLang);
            break;
    }

    return dwLang == PRIMARYLANGID(dwTransformLang);
}
