﻿//-------------------------------------------------------------------------------------------------
// <copyright file="BurnBundleBuilder.cs" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
// 
// <summary>
// WiX Burn bootstrapper build task generates a chainer manifest from items and properties
// defined in a .burnproj project and runs the Burn bundle bunder.
// </summary>
//-------------------------------------------------------------------------------------------------

namespace Microsoft.Tools.WindowsInstallerXml.Build.Tasks
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Xml;
    using Microsoft.Build.Framework;
    using Microsoft.Build.Utilities;

    /// <summary>
    /// An MSBuild task to generate chainer manifests and run the Burn.exe bundle builder.
    /// </summary>
    public class BurnBundleBuilder : Task
    {
        // these are metadata names that are written as-is as attributes into the source manifest
        // (rather than being handled specially as child elements like Log).
        private static Dictionary<string, XmlNodeType> ResourceMetadataNames = new Dictionary<string, XmlNodeType>
        {
            { "Packaging", XmlNodeType.Attribute },
            { "RelativeDirectory", XmlNodeType.Attribute },
        };

        private static Dictionary<string, XmlNodeType> CommonPackageMetadataNames = new Dictionary<string, XmlNodeType>
        {
            { "Id", XmlNodeType.Attribute }, 
            { "Permanent", XmlNodeType.Attribute }, 
        };

        private static Dictionary<string, XmlNodeType> ExePackageMetadataNames = new Dictionary<string, XmlNodeType>
        {
            { "DetectionCondition", XmlNodeType.Element },
            { "InstallArguments", XmlNodeType.Attribute },
            { "RepairArguments", XmlNodeType.Attribute },
            { "UninstallArguments", XmlNodeType.Attribute },
        };

        private static Dictionary<string, XmlNodeType> MsiPackageMetadataNames = new Dictionary<string, XmlNodeType>
        {
            { "UILevel", XmlNodeType.Attribute }, 
        };

        #region Manifest detail properties
        /// <summary>
        /// Gets or sets the installation chain.
        /// </summary>
        /// <value>A list of packages with metadata.</value>
        public ITaskItem[] Chain { get; set; }

        /// <summary>
        /// Gets or sets the chain-level log level.
        /// </summary>
        /// <value>One of: normal(default), verbose, debug.</value>
        public string LogLevel { get; set; }

        /// <summary>
        /// Gets or sets the chain-level log mode.
        /// </summary>
        /// <value>One of: append, overwrite, unique(default).</value>
        public string LogMode { get; set; }

        /// <summary>
        /// Gets or sets the chain-level log path.
        /// </summary>
        /// <value>Absolute path for a log file. Environment variables [%VAR%] are resolved.</value>
        public string LogPath { get; set; }

        /// <summary>
        /// Gets or sets the path to the Burn manifest to generate.
        /// </summary>
        /// <value>Path to the generated Burn manifest.</value>
        [Required]
        [Output]
        public ITaskItem Manifest { get; set; }

        /// <summary>
        /// Gets or sets the path to a Burn-compatible stub.
        /// </summary>
        /// <value>Path to a stub .exe. Default: burnstub.exe.</value>
        public string StubPath { get; set; }

        /// <summary>
        /// Gets or sets the UX DLL and resources to be included .
        /// </summary>
        /// <value>A list of UX resources, starting with the UX DLL. Default: stdux.dll;thm.xml;stdux.png;tou.htm.</value>
        public ITaskItem[] UX { get; set; }
        #endregion

        public override bool Execute()
        {
            return this.GenerateManifest();
        }

        private bool GenerateManifest()
        {
            bool success = true;

            using (XmlWriter writer = XmlWriter.Create(this.Manifest.ItemSpec))
            {
                writer.WriteStartDocument();
                writer.WriteStartElement("BurnManifest", "http://schemas.microsoft.com/wix/2008/Burn");

                WriteLogElement(writer, this.LogLevel, this.LogMode, this.LogPath);

                if (!String.IsNullOrEmpty(this.StubPath))
                {
                    writer.WriteStartElement("Stub");
                    writer.WriteAttributeString("SourceFile", this.StubPath);
                    writer.WriteEndElement();
                }

                if (null != this.UX && 0 < this.UX.Length)
                {
                    writer.WriteStartElement("UX");
                    WriteResourceAttributes(writer, this.UX[0]);

                    if (1 < this.UX.Length)
                    {
                        for (int i = 1; i < this.UX.Length; ++i)
                        {
                            writer.WriteStartElement("Resource");
                            WriteResourceAttributes(writer, this.UX[i]);
                            writer.WriteEndElement();
                        }
                    }

                    writer.WriteEndElement();
                }

                if (null != this.Chain)
                {
                    writer.WriteStartElement("Chain");

                    foreach (ITaskItem package in this.Chain)
                    {
                        string elementName;
                        Dictionary<string, XmlNodeType> metadataNames;

                        if (package.ItemSpec.EndsWith(".exe"))
                        {
                            elementName = "ExePackage";
                            metadataNames = ExePackageMetadataNames;
                        }
                        else if (package.ItemSpec.EndsWith(".msi"))
                        {
                            elementName = "MsiPackage";
                            metadataNames = MsiPackageMetadataNames;
                        }
                        else
                        {
                            continue;
                        }

                        writer.WriteStartElement(elementName);
                        WriteResourceAttributes(writer, package);
                        WriteCommonPackageMetadataAttributes(writer, package);
                        WriteMetadataAsAttributes(writer, package, metadataNames);
                        WriteMetadataAsElements(writer, package, metadataNames);
                        WriteCommonPackageMetadataElements(writer, package);
                        writer.WriteEndElement();
                    }

                    writer.WriteEndElement();
                }

                writer.WriteEndElement();
                writer.WriteEndDocument();
            }

            return success;
        }

        private static void WriteCommonPackageMetadataAttributes(XmlWriter writer, ITaskItem package)
        {
            // default id to the name of the package, munging it to id-ness as appropriate
            string id = package.GetMetadata("Id");
            if (String.IsNullOrEmpty(id))
            {
                id = Path.GetFileNameWithoutExtension(package.ItemSpec).Replace(' ', '_');
                package.SetMetadata("Id", id);
            }

            // some metadata goes straight into the manifest as attributes
            WriteMetadataAsAttributes(writer, package, CommonPackageMetadataNames);

            // default Vital to "yes" (because, really, if you don't care about it, why install it?)
            bool vital = // allow the MSBuild-typical true/false boolean values plus the WiX-typical yes/no
                !String.Equals(package.GetMetadata("Vital"), "no", StringComparison.OrdinalIgnoreCase) &&
                !String.Equals(package.GetMetadata("Vital"), "false", StringComparison.OrdinalIgnoreCase);
            writer.WriteAttributeString("Vital", vital ? "yes" : "no");
        }

        private static void WriteCommonPackageMetadataElements(XmlWriter writer, ITaskItem package)
        {
            // these turn into a Log child element
            string logLevel = package.GetMetadata("LogLevel");
            string logMode = package.GetMetadata("LogMode");
            string logPath = package.GetMetadata("LogPath");
            WriteLogElement(writer, logLevel, logMode, logPath);

            // these turn into SetProperty child elements
            string properties = package.GetMetadata("Properties");
            if (!String.IsNullOrEmpty(properties))
            {
                // turn name1=value1;name2=value2 into []=name1,value1,name2,value2
                string[] pairs = properties.Trim().Trim(';', '=').Split(';', '=');
                if (null != pairs && 0 < pairs.Length)
                {
                    for (int i = 0; i < pairs.Length; i += 2)
                    {
                        writer.WriteStartElement("SetProperty");
                        writer.WriteAttributeString("Id", pairs[i]);
                        writer.WriteAttributeString("Global", pairs[i + 1]);
                        writer.WriteEndElement();
                    }
                }
            }
        }

        private static void WriteResourceAttributes(XmlWriter writer, ITaskItem item)
        {
            writer.WriteAttributeString("SourceFile", item.ItemSpec);
            WriteMetadataAsAttributes(writer, item, ResourceMetadataNames);
        }

        private static void WriteMetadataAsAttributes(XmlWriter writer, ITaskItem item, Dictionary<string, XmlNodeType> legalMetadataNames)
        {
            foreach (string metadataName in item.MetadataNames)
            {
                XmlNodeType type;
                if (legalMetadataNames.TryGetValue(metadataName, out type) && XmlNodeType.Attribute == type)
                {
                    string metadataValue = item.GetMetadata(metadataName);
                    if (!String.IsNullOrEmpty(metadataValue))
                    {
                        writer.WriteAttributeString(metadataName, metadataValue);
                    }
                }
            }
        }

        private static void WriteMetadataAsElements(XmlWriter writer, ITaskItem item, Dictionary<string, XmlNodeType> legalMetadataNames)
        {
            foreach (string metadataName in item.MetadataNames)
            {
                XmlNodeType type;
                if (legalMetadataNames.TryGetValue(metadataName, out type) && XmlNodeType.Element == type)
                {
                    string metadataValue = item.GetMetadata(metadataName);
                    if (!String.IsNullOrEmpty(metadataValue))
                    {
                        writer.WriteElementString(metadataName, metadataValue);
                    }
                }
            }
        }

        private static void WriteLogElement(XmlWriter writer, string logLevel, string logMode, string logPath)
        {
            if (!String.IsNullOrEmpty(logLevel) || !String.IsNullOrEmpty(logMode) || !String.IsNullOrEmpty(logPath))
            {
                writer.WriteStartElement("Log");

                if (!String.IsNullOrEmpty(logLevel))
                {
                    // TODO: Validate
                    writer.WriteAttributeString("Level", logLevel);
                }
                if (!String.IsNullOrEmpty(logMode))
                {
                    // TODO: Validate
                    writer.WriteAttributeString("Mode", logMode);
                }
                if (!String.IsNullOrEmpty(logPath))
                {
                    // TODO: Validate
                    writer.WriteAttributeString("Path", logPath);
                }

                writer.WriteEndElement();
            }
        }
    }
}
