//-------------------------------------------------------------------------------------------------
// <copyright file="StandardUserExperience.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"

static const LPCWSTR STDUX_WINDOW_CLASS = L"WixBurnStdux";

enum STDUX_STATE
{
    STDUX_STATE_INITIALIZING,
    STDUX_STATE_PREPARING,
    STDUX_STATE_PREPARED,
    STDUX_STATE_CACHING,
    STDUX_STATE_CACHED,
    STDUX_STATE_EXECUTING,
    STDUX_STATE_EXECUTED,
    STDUX_STATE_RESTART,
    STDUX_STATE_ERROR,
};


enum STDUX_CONTROL
{
    STDUX_CONTROL_TOU_LINK,
    STDUX_CONTROL_INSTALL_BUTTON,
    STDUX_CONTROL_REPAIR_BUTTON,
    STDUX_CONTROL_UNINSTALL_BUTTON,
    STDUX_CONTROL_CLOSE_BUTTON,
    STDUX_CONTROL_OK_BUTTON,
    STDUX_CONTROL_CANCEL_BUTTON,
    STDUX_CONTROL_RETRY_BUTTON,
    STDUX_CONTROL_ABORT_BUTTON,
    STDUX_CONTROL_IGNORE_BUTTON,
    STDUX_CONTROL_YES_BUTTON,
    STDUX_CONTROL_NO_BUTTON,
    STDUX_CONTROL_PROGRESS_BAR,
    STDUX_CONTROL_MESSAGE_TEXT,
};

enum WM_STDUX
{
    WM_STDUX_PREPARE_PACKAGES = WM_APP + 1,
    WM_STDUX_CACHE_PACKAGES,
    WM_STDUX_EXECUTE_PACKAGES,
    WM_STDUX_MODAL_ERROR,
};


class CStandardUserExperience : public IBurnUserExperience
{
public: // IUnknown
    virtual HRESULT __stdcall QueryInterface(
        __in const IID& riid,
        __out void** ppvObject
        )
    {
        HRESULT hr = S_OK;

        ExitOnNull(ppvObject, hr, E_INVALIDARG, "Invalid argument ppvObject");
        *ppvObject = NULL;

        if (::IsEqualIID(__uuidof(IBurnUserExperience), riid))
        {
            *ppvObject = static_cast<IBurnUserExperience*>(this);
        }
        else if (::IsEqualIID(IID_IUnknown, riid))
        {
            *ppvObject = static_cast<IUnknown*>(this);
        }
        else // no interface for requested iid
        {
            ExitFunction1(hr = E_NOINTERFACE);
        }

        AddRef();
    LExit:
        return hr;
    }


    virtual ULONG __stdcall AddRef()
    {
        return ::InterlockedIncrement(&this->m_cReferences);
    }


    virtual ULONG __stdcall Release()
    {
        long l = ::InterlockedDecrement(&this->m_cReferences);
        if (0 < l)
        {
            return l;
        }

        delete this;
        return 0;
    }


public: // IBurnUserExperience
    //
    // Initialize - ensure all the necessary objects are created/initialized.
    //
    STDMETHODIMP Initialize(
        __in IBurnCore* pCore,
        __in int nCmdShow
        )
    {
        HRESULT hr = S_OK;
        INITCOMMONCONTROLSEX icc = { 0 };

        m_hModalWait = ::CreateEventW(NULL, TRUE, FALSE, NULL);
        ExitOnNullWithLastError(m_hModalWait, hr, "Failed to create modal event.");

        // Initialize common controls.
        icc.dwSize = sizeof(icc);
        icc.dwICC = ICC_PROGRESS_CLASS /* | ICC_STANDARD_CLASSES (requires 0x0501 == _WIN32_WINNT*/;
        ::InitCommonControlsEx(&icc);

        hr = ThemeInitialize(m_hModule);
        ExitOnFailure(hr, "Failed to initialize theme manager.");

        pCore->AddRef();
        m_pCore = pCore;

        hr = CreateMainWindow();
        ExitOnFailure(hr, "Failed to create main window.");

    LExit:
        return hr;
    }


    //
    // Run - start pumping messages.
    //
    STDMETHODIMP Run()
    {
        HRESULT hr = S_OK;
        BOOL fRet = FALSE;
        MSG msg;

        while (0 != (fRet = ::GetMessage(&msg, NULL, 0, 0)))
        {
            if (-1 == fRet)
            {
                hr = E_UNEXPECTED;
                ExitOnFailure(hr, "Unexpected return value from message pump.");
            }
            else if (!::IsDialogMessage(m_hWnd, &msg))
            {
                ::TranslateMessage(&msg);
                ::DispatchMessage(&msg);
            }
        }

    LExit:
        return hr;
    }

    STDMETHODIMP_(void) Uninitialize()
    {
    }

    STDMETHODIMP_(BOOL) OnPrepareBegin(
        __in DWORD cPackages
        )
    {
        return TRUE;
    }


    STDMETHODIMP_(BOOL) OnPreparePackageBegin(
        __in_z LPCWSTR wzPackageId
        )
    {
        Trace(REPORT_STANDARD, "CStandardUserExperience::OnPreparePackageBegin()");
        return TRUE; // be a good Boy Scout, always prepare.
    }


    STDMETHODIMP_(void) OnPreparePackageComplete(
        __in LPCWSTR wzPackageId,
        __in HRESULT hrStatus,
        __in PACKAGE_STATE state,
        __in REQUEST_STATE requested,
        __in ACTION_STATE execute,
        __in ACTION_STATE rollback
        )
    {
        Trace(REPORT_STANDARD, "CStandardUserExperience::OnPreparePackageComplete()");
    }


    STDMETHODIMP_(void) OnPrepareComplete(
        __in HRESULT hrStatus
        )
    {
        Trace(REPORT_STANDARD, "CStandardUserExperience::OnPrepareComplete()");
        if (SUCCEEDED(hrStatus))
        {
            SetState(STDUX_STATE_PREPARED);

            // Quiet or passive display just goes straight into caching (if installing) or
            // executing (if uninstalling).
            if (BURN_DISPLAY_NONE == m_command.display || BURN_DISPLAY_PASSIVE == m_command.display)
            {
                ::PostMessageW(m_hWnd, BURN_ACTION_INSTALL == m_command.action ? WM_STDUX_CACHE_PACKAGES : WM_STDUX_EXECUTE_PACKAGES, 0, 0);
            }
        }
    }


    STDMETHODIMP_(void) OnCacheComplete(
        __in HRESULT hrStatus
        )
    {
        if (SUCCEEDED(hrStatus))
        {
            SetState(STDUX_STATE_CACHED);

            // After caching the packages, always execute them.
            ::PostMessageW(m_hWnd, WM_STDUX_EXECUTE_PACKAGES, 0, 0);
        }
    }


    STDMETHODIMP_(BOOL) OnExecuteBegin(
        __in DWORD cExecutingPackages
        )
    {
        return TRUE;
    }


    STDMETHODIMP_(BOOL) OnExecutePackageBegin(
        __in LPCWSTR wzPackageId,
        __in BOOL fExecute
        )
    {
        Trace(REPORT_STANDARD, "CStandardUserExperience::OnExecutePackageBegin()");
        return TRUE; // always execute.
    }


    STDMETHODIMP_(INT) OnError(
        __in DWORD dwCode,
        __in_z LPCWSTR wzError,
        __in DWORD dwUIHint
        )
    {
        HRESULT hr = S_OK;

        if (BURN_DISPLAY_FULL == m_command.display)
        {
            ModalPrompt(STDUX_STATE_ERROR, dwUIHint, wzError);
        }
        else // just cancel when quiet or passive.
        {
            m_nModalResult = IDABORT;
        }

        return m_nModalResult;
    }


    STDMETHODIMP_(INT) OnProgress(
        __in DWORD dwProgressPercentage,
        __in DWORD dwOverallProgressPercentage
        )
    {
        HRESULT hr = S_OK;

        m_nModalResult = IDOK;

        hr = ThemeSetProgressControl(m_pTheme, STDUX_CONTROL_PROGRESS_BAR, dwOverallProgressPercentage);
        ExitOnFailure(hr, "Failed to set progress.");

    LExit:
        if (FAILED(hr) && IDNOACTION == m_nModalResult)
        {
            m_nModalResult = IDERROR;
        }

        return m_nModalResult;
    }


    STDMETHODIMP_(void) OnExecutePackageComplete(
        __in LPCWSTR wzPackageId,
        __in HRESULT hrExitCode
        )
    {
        Trace(REPORT_STANDARD, "CStandardUserExperience::OnExecutePackageComplete()");
    }


    STDMETHODIMP_(void) OnExecuteComplete(
        __in HRESULT hrStatus
        )
    {
        // Failure or quiet or passive display just close at the end, no questions asked.
        if (BURN_DISPLAY_NONE == m_command.display || BURN_DISPLAY_PASSIVE == m_command.display)
        {
            ::PostMessageW(m_hWnd, WM_CLOSE, 0, 0);
        }
        else
        {
            if (SUCCEEDED(hrStatus))
            {
                OnProgress(100, 100); // ensure we're 100% complete.
            }

            SetState(STDUX_STATE_EXECUTED);
        }
    }


    STDMETHODIMP_(BOOL) OnRestartRequired()
    {
        // If the user already said "OK", don't ask again.
        if (!m_fAllowRestart)
        {
            if (BURN_DISPLAY_FULL == m_command.display && BURN_RESTART_PROMPT == m_command.restart)
            {
                // TODO: Get restart message from somewhere

                ModalPrompt(STDUX_STATE_RESTART, MB_OKCANCEL, L"Restart required.  Allow restart?");
                m_fAllowRestart = (IDOK == m_nModalResult); // only allow restart if the user clicked OK.
            }
            else
            {
                m_fAllowRestart = (BURN_RESTART_ALWAYS == m_command.restart || BURN_RESTART_AUTOMATIC == m_command.restart || BURN_RESTART_PROMPT == m_command.restart);
            }
        }

        return m_fAllowRestart;
    }


    STDMETHODIMP GetControlText(
        __in DWORD dwControlID,
        __out LPWSTR* psczText
        )
    {
        HRESULT hr = ThemeGetTextControl(m_pTheme, dwControlID, psczText);

        return hr;
    }


private: // privates
    //
    // CreateMainWindow - creates the main install window.
    //
    HRESULT CreateMainWindow()
    {
        HRESULT hr = S_OK;
        WNDCLASSW wc = { 0 };
        DWORD dwWindowStyle = 0;
        LPWSTR sczThemePath = NULL;
        LPWSTR sczCommandLine = NULL;
        
        // load theme relative to stdux.dll.
        hr = PathRelativeToModule(&sczThemePath, L"thm.xml", m_hModule);
        ExitOnFailure(hr, "Failed to combine module path with thm.xml.");

        hr = ThemeLoadFromFile(sczThemePath, &m_pTheme);
        ExitOnFailure(hr, "Failed to load theme from thm.xml.");

        hr = m_pCore->GetCommandLineParameters(&sczCommandLine);

        if (S_OK == hr)
        {
            hr = ProcessCommandLine(sczCommandLine);
            ExitOnFailure(hr, "Unknown commandline parameters.");
        }

        if (NULL != m_sczLanguage)
        {
            hr = ThemeLoadLocFromFile(m_pTheme, m_sczLanguage, m_hModule);
            ExitOnFailure(hr, "Failed to localize from /lang.");
        }
        else
        {
            hr = ThemeLoadLocFromFile(m_pTheme, L"en-us.wxl", m_hModule);
            ExitOnFailure(hr, "Failed to localize from default language.");
        }

        // Register the window class and create the window.
        wc.style = 0;
        wc.lpfnWndProc = CStandardUserExperience::WndProc;
        wc.cbClsExtra = 0;
        wc.cbWndExtra = 0;
        wc.hInstance = m_hModule;
        wc.hIcon = reinterpret_cast<HICON>(m_pTheme->hIcon);
        wc.hCursor = ::LoadCursor(NULL, IDC_ARROW);
        wc.hbrBackground = m_pTheme->rgFonts[m_pTheme->dwFontId].hBackground;
        wc.lpszMenuName = NULL;
        wc.lpszClassName = STDUX_WINDOW_CLASS;
        if (!::RegisterClassW(&wc))
        {
            ExitWithLastError(hr, "Failed to register window.");
        }

        m_fRegistered = TRUE;

        // Calculate the window style based on the theme style and command display value.
        dwWindowStyle = m_pTheme->dwStyle;
        if (BURN_DISPLAY_NONE == m_command.display)
        {
            dwWindowStyle &= ~WS_VISIBLE;
        }

        m_hWnd = ::CreateWindowExW(0, wc.lpszClassName, m_pTheme->wzCaption, dwWindowStyle, CW_USEDEFAULT, CW_USEDEFAULT, m_pTheme->nWidth, m_pTheme->nHeight, HWND_DESKTOP, NULL, m_hModule, this);
        ExitOnNullWithLastError(m_hWnd, hr, "Failed to create window.");

    LExit:
        ReleaseStr(sczThemePath);
        ReleaseStr(sczCommandLine);
        ReleaseStr(m_sczLanguage);

        if (FAILED(hr))
        {
            DestroyMainWindow();
        }

        return hr;
    }


    //
    // DestroyMainWindow - clean up all the window registration.
    //
    void DestroyMainWindow()
    {
        if (m_hWnd)
        {
            ::CloseWindow(m_hWnd);
            m_hWnd = NULL;
        }

        if (m_fRegistered)
        {
            ::UnregisterClassW(STDUX_WINDOW_CLASS, m_hModule);
            m_fRegistered = FALSE;
        }

        if (m_pTheme)
        {
            ThemeFree(m_pTheme);
            m_pTheme = NULL;
        }
    }


    //
    // ProcessCommandLine - process the provided command line arguments.
    //
    HRESULT ProcessCommandLine(
        __in_z LPCWSTR wzCommandLine
        )
    {
        HRESULT hr = S_OK;
        LPWSTR sczName = NULL;

        int argc = 0;
        LPWSTR* argv = ::CommandLineToArgvW(wzCommandLine, &argc);
        ExitOnNullWithLastError(argv, hr, "Failed to get command line.");

        for (int i = 0; i < argc; ++i)
        {
            if (argv[i][0] == L'-' || argv[i][0] == L'/')
            {
                if (CSTR_EQUAL == ::CompareStringW(LOCALE_INVARIANT, NORM_IGNORECASE, &argv[i][1], -1, L"lang", -1))
                {
                    if (i + 1 >= argc)
                    {
                        ExitOnRootFailure(hr = E_INVALIDARG, "Must specify a language.");
                    }

                    ++i;

                    hr = StrAllocString(&m_sczLanguage, &argv[i][0], 0);
                    ExitOnFailure(hr, "Failed to copy language.");

                    hr = StrAllocConcat(&m_sczLanguage, L".wxl", 0);
                    ExitOnFailure(hr, "Failed to concatenate wxl extension.");
                }
            }
        }

    LExit:
        ReleaseStr(sczName);

        if (argv)
        {
            ::LocalFree(argv);
        }

        return hr;
    }


    //
    // ModalPrompt - prompt the user to set the "m_nModalResult" value.
    //
    void ModalPrompt(
        __in STDUX_STATE modalState,
        __in DWORD dwUIHint,
        __in_z_opt LPCWSTR wzMessage
        )
    {
        HRESULT hr = S_OK;
        STDUX_STATE state = m_state; // remember the state so we can reset after going into the error state.

        m_nModalResult = IDNOACTION;

        if (wzMessage)
        {
            hr = ThemeSetTextControl(m_pTheme, STDUX_CONTROL_MESSAGE_TEXT, wzMessage);
            ExitOnFailure(hr, "Failed to set message text.");
        }

        SetState(modalState);
        ShowModalControls(dwUIHint);

        ::ResetEvent(m_hModalWait);

        ::WaitForSingleObject(m_hModalWait, INFINITE);

        ShowModalControls(0xF);

    LExit:
        m_state = state; // put the state back.
    }


    //
    // WndProc - standard windows message handler.
    //
    static LRESULT CALLBACK WndProc(
        __in HWND hWnd,
        __in UINT uMsg,
        __in WPARAM wParam,
        __in LPARAM lParam
        )
    {
        LRESULT lres = 0;
#pragma warning(suppress:4312)
        CStandardUserExperience* pUX = reinterpret_cast<CStandardUserExperience*>(::GetWindowLongPtr(hWnd, GWLP_USERDATA));

        switch (uMsg)
        {
        case WM_NCCREATE:
            {
            LPCREATESTRUCT lpcs = reinterpret_cast<LPCREATESTRUCT>(lParam);
            pUX = reinterpret_cast<CStandardUserExperience*>(lpcs->lpCreateParams);
#pragma warning(suppress:4244)
            ::SetWindowLongPtr(hWnd, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(pUX));
            }
            break;

        case WM_NCHITTEST:
            if (pUX->m_pTheme->dwStyle & WS_POPUP)
            {
                return HTCAPTION; // allow pop-up windows to be moved by grabbing any non-control.
            }
            break;

        case WM_NCDESTROY:
            lres = ::DefWindowProc(hWnd, uMsg, wParam, lParam);
            ::SetWindowLongPtr(hWnd, GWLP_USERDATA, 0);
            return lres;

        case WM_CREATE:
            if (!pUX->OnCreate(hWnd))
            {
                return -1;
            }
            break;

        case WM_CLOSE:
            if (STDUX_STATE_EXECUTED > pUX->m_state)
            {
                // Check with the user to verify they want to cancel.
                //pUX->PromptCancel(hWnd);
            }
            pUX->m_fCanceled = TRUE; // TODO: let the prompt set this instead.

            if (pUX->m_fCanceled)
            {
                ::DestroyWindow(hWnd);
            }
            return 0;

        case WM_DESTROY:
            ::PostQuitMessage(0);
            break;

        case WM_DRAWITEM:
            ThemeDrawControl(pUX->m_pTheme, reinterpret_cast<LPDRAWITEMSTRUCT>(lParam));
            return TRUE;

        case WM_CTLCOLORSTATIC:
            {
            HBRUSH hBrush = NULL;
            if (ThemeSetControlColor(pUX->m_pTheme, reinterpret_cast<HDC>(wParam), reinterpret_cast<HWND>(hWnd), &hBrush))
            {
                return reinterpret_cast<LRESULT>(hBrush);
            }
            }
            break;

        case WM_SETCURSOR:
            ThemeHoverControl(pUX->m_pTheme, hWnd, reinterpret_cast<HWND>(wParam));
            break;

        case WM_PAINT:
            // If there is anything to update, do so.
            if (::GetUpdateRect(hWnd, NULL, FALSE))
            {
                PAINTSTRUCT ps;
                ::BeginPaint(hWnd, &ps);
                ThemeDrawBackground(pUX->m_pTheme, &ps);
                ::EndPaint(hWnd, &ps);
            }
            return 0;

        case WM_STDUX_PREPARE_PACKAGES:
            pUX->OnPrepare();
            return 0;

        case WM_STDUX_CACHE_PACKAGES:
            pUX->OnCache();
            return 0;

        case WM_STDUX_EXECUTE_PACKAGES:
            pUX->OnExecute();
            return 0;

        case WM_COMMAND:
            switch (LOWORD(wParam))
            {
            case STDUX_CONTROL_TOU_LINK:
                pUX->OnClickTermsOfUse();
                break;

            case STDUX_CONTROL_INSTALL_BUTTON:
                pUX->OnClickInstallButton();
                break;

            case STDUX_CONTROL_REPAIR_BUTTON:
                pUX->OnClickRepairButton();
                break;

            case STDUX_CONTROL_UNINSTALL_BUTTON:
                pUX->OnClickUninstallButton();
                break;

            case STDUX_CONTROL_CLOSE_BUTTON:
                pUX->OnClickCloseButton();
                break;

            case STDUX_CONTROL_OK_BUTTON: __fallthrough;
            case STDUX_CONTROL_CANCEL_BUTTON: __fallthrough;
            case STDUX_CONTROL_RETRY_BUTTON: __fallthrough;
            case STDUX_CONTROL_ABORT_BUTTON: __fallthrough;
            case STDUX_CONTROL_IGNORE_BUTTON: __fallthrough;
            case STDUX_CONTROL_YES_BUTTON: __fallthrough;
            case STDUX_CONTROL_NO_BUTTON: __fallthrough;
                pUX->OnClickModalButton(LOWORD(wParam));
                break;

            default:
                break;
            }
            return 0;
        }

        return ::DefWindowProc(hWnd, uMsg, wParam, lParam);
    }


    //
    // OnCreate - finishes loading the theme and tells the core we're ready for data.
    //
    BOOL OnCreate(
        __in HWND hWnd
        )
    {
        HRESULT hr = S_OK;

        hr = ThemeLoadControls(m_pTheme, hWnd);
        ExitOnFailure(hr, "Failed to load theme controls.");

        // Okay, we're ready for packages now.
        SetState(STDUX_STATE_PREPARING);

        ::PostMessageW(hWnd, WM_STDUX_PREPARE_PACKAGES, 0, 0);

    LExit:
        return SUCCEEDED(hr);
    }


    //
    // OnPrepare - start the processing of packages.
    //
    void OnPrepare()
    {
        HRESULT hr = S_OK;

        Trace(REPORT_STANDARD, "CStandardUserExperience::OnPrepare()");

        // Tell the core we're ready for the packages to be processed now.
        hr = m_pCore->Prepare(NULL, BURN_ACTION_UNINSTALL == m_command.action ? REQUEST_STATE_ABSENT : REQUEST_STATE_PRESENT);
        ExitOnFailure(hr, "Failed to prepare chain.");

    LExit:
        return;
    }


    //
    // OnCache - cache the prepared changes.
    //
    void OnCache()
    {
        HRESULT hr = S_OK;

        Trace(REPORT_STANDARD, "CStandardUserExperience::OnCache()");

        hr = m_pCore->Cache(NULL);
        ExitOnFailure(hr, "Failed to cache install packages.");

        SetState(STDUX_STATE_CACHING);

    LExit:
        return;
    }


    //
    // OnExecute - execute the cached packages.
    //
    void OnExecute()
    {
        HRESULT hr = S_OK;

        Trace(REPORT_STANDARD, "CStandardUserExperience::OnExecute()");

        hr = m_pCore->Execute(REQUEST_STATE_NONE);
        ExitOnFailure(hr, "Failed to execute install.");

        SetState(STDUX_STATE_EXECUTING);

    LExit:
        return;
    }


    //
    // OnClickInstallButton - start the install by caching the packages.
    //
    void OnClickInstallButton()
    {
        this->OnCache();
    }


    //
    // OnClickRepairButton - start the repair.
    //
    void OnClickRepairButton()
    {
        HRESULT hr = S_OK;

        hr = m_pCore->Execute(REQUEST_STATE_REPAIR);
        ExitOnFailure(hr, "Failed to execute repair.");

        SetState(STDUX_STATE_EXECUTING);

    LExit:
        return;
    }


    //
    // OnClickUninstallButton - start the uninstall.
    //
    void OnClickUninstallButton()
    {
        this->OnExecute();
    }


    //
    // OnClickCloseButton - close the application.
    //
    void OnClickCloseButton()
    {
        ::PostMessageW(m_hWnd, WM_CLOSE, 0, 0);
    }


    //
    // OnClickModalButton - handle the modal button clicks.
    //
    void OnClickModalButton(
        int nButton
        )
    {
        Trace1(REPORT_STANDARD, "CStandardUserExperience::OnClickModalButton(%d) - enter", nButton);

        switch (nButton)
        {
        case STDUX_CONTROL_OK_BUTTON:
            m_nModalResult = IDOK;
            break;

        case STDUX_CONTROL_CANCEL_BUTTON:
            m_nModalResult = IDCANCEL;
            break;

        case STDUX_CONTROL_RETRY_BUTTON:
            m_nModalResult = IDRETRY;
            break;

        case STDUX_CONTROL_ABORT_BUTTON:
            m_nModalResult = IDABORT;
            break;

        case STDUX_CONTROL_IGNORE_BUTTON:
            m_nModalResult = IDIGNORE;
            break;

        case STDUX_CONTROL_YES_BUTTON:
            m_nModalResult = IDYES;
            break;

        case STDUX_CONTROL_NO_BUTTON:
            m_nModalResult = IDNO;
            break;
        }

        ::SetEvent(m_hModalWait);

        return;
    }


    //
    // OnClickTermsOfUse - show the terms of use.
    //
    void OnClickTermsOfUse()
    {
        HRESULT hr = S_OK;
        LPWSTR sczResUri = NULL;

        hr = PathRelativeToModule(&sczResUri, L"tou.htm", m_hModule);
        ExitOnFailure(hr, "Failed to create URI to TOU.");

        hr = ShellExec(sczResUri);
        ExitOnFailure(hr, "Failed to launch URI to TOU.");

    LExit:
        ReleaseStr(sczResUri);
        return;
    }


    //
    // SetState
    //
    void SetState(
        __in STDUX_STATE state
        )
    {
        if (m_state != state)
        {
            Trace2(REPORT_STANDARD, "CStandardUserExperience::SetState() changing from state %d to %d", m_state, state);
            m_state = state;

            int nShowTermsOfUse = (m_state < STDUX_STATE_EXECUTING) ? SW_SHOW : SW_HIDE;
            int nShowInstall = (BURN_DISPLAY_FULL == m_command.display && BURN_ACTION_INSTALL == m_command.action && m_state <= STDUX_STATE_PREPARED) ? SW_SHOW : SW_HIDE;
            int nShowRepair = (FALSE && BURN_DISPLAY_FULL == m_command.display && m_state <= STDUX_STATE_PREPARED) ? SW_SHOW : SW_HIDE;
            int nShowUninstall = (BURN_DISPLAY_FULL == m_command.display && BURN_ACTION_UNINSTALL == m_command.action && m_state <= STDUX_STATE_PREPARED) ? SW_SHOW : SW_HIDE;
            int nShowClose = (BURN_DISPLAY_FULL == m_command.display && STDUX_STATE_EXECUTED == m_state) ? SW_SHOW : SW_HIDE;

            int nShowMessage = ((STDUX_STATE_PREPARED < m_state && m_state <= STDUX_STATE_EXECUTED) || STDUX_STATE_ERROR == m_state) ? SW_SHOW : SW_HIDE;
            int nShowProgressBar = (STDUX_STATE_PREPARED < m_state && m_state <= STDUX_STATE_EXECUTED) ? SW_SHOW : SW_HIDE;

            ::ShowWindow(m_pTheme->rgControls[STDUX_CONTROL_TOU_LINK].hWnd, nShowTermsOfUse);
            ::ShowWindow(m_pTheme->rgControls[STDUX_CONTROL_INSTALL_BUTTON].hWnd, nShowInstall);
            ::ShowWindow(m_pTheme->rgControls[STDUX_CONTROL_REPAIR_BUTTON].hWnd, nShowRepair);
            ::ShowWindow(m_pTheme->rgControls[STDUX_CONTROL_UNINSTALL_BUTTON].hWnd, nShowUninstall);
            ::ShowWindow(m_pTheme->rgControls[STDUX_CONTROL_CLOSE_BUTTON].hWnd, nShowClose);
            ::ShowWindow(m_pTheme->rgControls[STDUX_CONTROL_MESSAGE_TEXT].hWnd, nShowMessage);
            ::ShowWindow(m_pTheme->rgControls[STDUX_CONTROL_PROGRESS_BAR].hWnd, nShowProgressBar);
        }
    }


    //
    // ShowModalControls
    //
    void ShowModalControls(
        __in DWORD dwControls
        )
    {
        int nControl = dwControls & 0xF;
        int nShowOK = (MB_OK == nControl || MB_OKCANCEL == nControl) ? SW_SHOW : SW_HIDE;
        int nShowCancel = (STDUX_STATE_EXECUTING == m_state || MB_OKCANCEL == nControl || MB_RETRYCANCEL == nControl || MB_YESNOCANCEL == nControl) ? SW_SHOW : SW_HIDE;
        int nShowRetry = (MB_RETRYCANCEL == nControl || MB_ABORTRETRYIGNORE == nControl) ? SW_SHOW : SW_HIDE;
        int nShowAbort = (MB_ABORTRETRYIGNORE == nControl) ? SW_SHOW : SW_HIDE;
        int nShowIgnore = (MB_ABORTRETRYIGNORE == nControl) ? SW_SHOW : SW_HIDE;
        int nShowYes = (MB_YESNO == nControl || MB_YESNOCANCEL == nControl) ? SW_SHOW : SW_HIDE;
        int nShowNo = (MB_YESNO == nControl || MB_YESNOCANCEL == nControl) ? SW_SHOW : SW_HIDE;

        ::ShowWindow(m_pTheme->rgControls[STDUX_CONTROL_OK_BUTTON].hWnd, nShowOK);
        ::ShowWindow(m_pTheme->rgControls[STDUX_CONTROL_CANCEL_BUTTON].hWnd, nShowCancel);
        ::ShowWindow(m_pTheme->rgControls[STDUX_CONTROL_RETRY_BUTTON].hWnd, nShowRetry);
        ::ShowWindow(m_pTheme->rgControls[STDUX_CONTROL_ABORT_BUTTON].hWnd, nShowAbort);
        ::ShowWindow(m_pTheme->rgControls[STDUX_CONTROL_IGNORE_BUTTON].hWnd, nShowIgnore);
        ::ShowWindow(m_pTheme->rgControls[STDUX_CONTROL_YES_BUTTON].hWnd, nShowYes);
        ::ShowWindow(m_pTheme->rgControls[STDUX_CONTROL_NO_BUTTON].hWnd, nShowNo);
    }


    HRESULT ShellExec(
        __in LPCWSTR wzTarget
        )
    {
        HRESULT hr = S_OK;

        HINSTANCE hinst = ::ShellExecuteW(m_hWnd, L"open", wzTarget, NULL, NULL, SW_SHOWDEFAULT);
        if (hinst <= HINSTANCE(32))
        {
            switch (int(hinst))
            {
            case ERROR_FILE_NOT_FOUND:
                hr = HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND);
                break;
            case ERROR_PATH_NOT_FOUND:
                hr = HRESULT_FROM_WIN32(ERROR_PATH_NOT_FOUND);
                break;
            case ERROR_BAD_FORMAT:
                hr = HRESULT_FROM_WIN32(ERROR_BAD_FORMAT);
                break;
            case SE_ERR_ASSOCINCOMPLETE:
            case SE_ERR_NOASSOC:
                hr = HRESULT_FROM_WIN32(ERROR_NO_ASSOCIATION);
                break;
            case SE_ERR_DDEBUSY:
            case SE_ERR_DDEFAIL:
            case SE_ERR_DDETIMEOUT:
                hr = HRESULT_FROM_WIN32(ERROR_DDE_FAIL);
                break;
            case SE_ERR_DLLNOTFOUND:
                hr = HRESULT_FROM_WIN32(ERROR_DLL_NOT_FOUND);
                break;
            case SE_ERR_OOM:
                hr = E_OUTOFMEMORY;
                break;
            case SE_ERR_ACCESSDENIED:
                hr = E_ACCESSDENIED;
                break;
            default:
                hr = E_FAIL;
            }

            ExitOnFailure1(hr, "ShellExec failed with return code %d", int(hinst));
        }

    LExit:
        return hr;
    }


public:
    //
    // Constructor - intitialize member variables.
    //
    CStandardUserExperience(
        __in HMODULE hModule,
        __in BURN_COMMAND* pCommand
        )
    {
        m_hModule = hModule;
        memcpy_s(&m_command, sizeof(m_command), pCommand, sizeof(BURN_COMMAND));

        m_cReferences = 1;
        m_pTheme = NULL;
        m_fRegistered = FALSE;
        m_hWnd = NULL;
        m_hwndHover = NULL;
        m_pCore = NULL;

        m_fCanceled = FALSE;
        m_state = STDUX_STATE_INITIALIZING;
        m_nModalResult = IDNOACTION;
        m_hModalWait = NULL;

        m_fAllowRestart = FALSE;
        m_sczLanguage = NULL;
    }


    //
    // Destructor - release member variables.
    //
    ~CStandardUserExperience()
    {
        DestroyMainWindow();

        if (m_pTheme)
        {
            ThemeFree(m_pTheme);
            m_pTheme = NULL;
        }

        ThemeUninitialize();

        if (m_hModalWait)
        {
            ::CloseHandle(m_hModalWait);
        }

        ReleaseNullObject(m_pCore);
    }

private:
    long m_cReferences;

    HMODULE m_hModule;
    BURN_COMMAND m_command;

    IBurnCore* m_pCore;
    THEME* m_pTheme;
    BOOL m_fRegistered;
    HWND m_hWnd;
    HWND m_hwndHover;

    BOOL m_fCanceled;
    STDUX_STATE m_state;
    INT m_nModalResult;
    HANDLE m_hModalWait;

    BOOL m_fAllowRestart;
    LPWSTR m_sczLanguage;
};


//
// CreateUserExperience - creates a new IBurnUserExperience object.
//
HRESULT CreateUserExperience(
    __in HMODULE hModule,
    __in BURN_COMMAND* pCommand,
    __out IBurnUserExperience **ppUX
    )
{
    HRESULT hr = S_OK;
    CStandardUserExperience* pUX = NULL;

    pUX = new CStandardUserExperience(hModule, pCommand);
    ExitOnNull(pUX, hr, E_OUTOFMEMORY, "Failed to create new standard UX object.");

    *ppUX = pUX;
    pUX = NULL;

LExit:
    ReleaseObject(pUX);
    return hr;
}
