//-------------------------------------------------------------------------------------------------
// <copyright file="buffutil.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
// </copyright>
// 
// <summary>
//    Binary serialization helper functions.
// </summary>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"


// constants

#define BUFFER_INCREMENT 128


// helper function declarations

static HRESULT EnsureBufferSize(
    __inout BYTE** ppbBuffer,
    __in SIZE_T cbSize
    );


// functions

extern "C" HRESULT BuffReadNumber(
    __in_bcount(cbBuffer) BYTE* pbBuffer,
    __in SIZE_T cbBuffer,
    __inout SIZE_T* piBuffer,
    __out DWORD* pdw
    )
{
    Assert(pbBuffer);
    Assert(piBuffer);
    Assert(pdw);

    // verify buffer size
    if (sizeof(DWORD) > cbBuffer - *piBuffer)
    {
        return E_INVALIDARG;
    }

    *pdw = *(DWORD*)(pbBuffer + *piBuffer);
    *piBuffer += sizeof(DWORD);
    return S_OK;
}

extern "C" HRESULT BuffReadNumber64(
    __in_bcount(cbBuffer) BYTE* pbBuffer,
    __in SIZE_T cbBuffer,
    __inout SIZE_T* piBuffer,
    __out DWORD64* pdw64
    )
{
    Assert(pbBuffer);
    Assert(piBuffer);
    Assert(pdw64);

    // verify buffer size
    if (sizeof(DWORD64) > cbBuffer - *piBuffer)
    {
        return E_INVALIDARG;
    }

    *pdw64 = *(DWORD64*)(pbBuffer + *piBuffer);
    *piBuffer += sizeof(DWORD64);
    return S_OK;
}

extern "C" HRESULT BuffReadString(
    __in_bcount(cbBuffer) BYTE* pbBuffer,
    __in SIZE_T cbBuffer,
    __inout SIZE_T* piBuffer,
    __inout_z LPWSTR* pscz
    )
{
    Assert(pbBuffer);
    Assert(piBuffer);
    Assert(pscz);

    HRESULT hr = S_OK;
    DWORD cch = 0;

    // verify buffer size
    if (sizeof(DWORD) > cbBuffer - *piBuffer)
    {
        ExitFunction1(hr = E_INVALIDARG);
    }

    // read character count
    cch = *(DWORD*)(pbBuffer + *piBuffer);
    *piBuffer += sizeof(DWORD);

    // verify buffer size
    if (cch * sizeof(WCHAR) > cbBuffer - *piBuffer)
    {
        ExitFunction1(hr = E_INVALIDARG);
    }

    // allocate string
    hr = StrAlloc(pscz, cch + 1); // + 1 to add null terminator
    ExitOnFailure(hr, "Failed to allocate string.");

    // read character data
    memcpy_s(*pscz, cbBuffer - *piBuffer, pbBuffer + *piBuffer, cch * sizeof(WCHAR));
    *piBuffer += cch * sizeof(WCHAR);

LExit:
    return hr;
}

extern "C" HRESULT BuffReadStream(
    __in_bcount(cbBuffer) BYTE* pbBuffer,
    __in SIZE_T cbBuffer,
    __inout SIZE_T* piBuffer,
    __out BYTE** ppbStream,
    __out SIZE_T* pcbStream
    )
{
    Assert(pbBuffer);
    Assert(piBuffer);
    Assert(ppbStream);
    Assert(pcbStream);

    HRESULT hr = S_OK;
    DWORD64 cb = 0;

    // verify buffer size
    if (sizeof(DWORD64) > cbBuffer - *piBuffer)
    {
        ExitFunction1(hr = E_INVALIDARG);
    }

    // read stream size
    cb = *(DWORD64*)(pbBuffer + *piBuffer);
    *piBuffer += sizeof(DWORD64);

    // verify buffer size
    if (cb > cbBuffer - *piBuffer)
    {
        ExitFunction1(hr = E_INVALIDARG);
    }

    // allocate buffer
    *ppbStream = (BYTE*)MemAlloc((SIZE_T)cb, TRUE);
    ExitOnNull(*ppbStream, hr, E_OUTOFMEMORY, "Failed to allocate stream.");

    // read stream data
    memcpy_s(*ppbStream, cbBuffer - *piBuffer, pbBuffer + *piBuffer, (SIZE_T)cb);
    *piBuffer += (SIZE_T)cb;

    // return stream size
    *pcbStream = (SIZE_T)cb;

LExit:
    return hr;
}

extern "C" HRESULT BuffWriteNumber(
    __inout BYTE** ppbBuffer,
    __inout SIZE_T* piBuffer,
    __in DWORD dw
    )
{
    Assert(ppbBuffer);
    Assert(piBuffer);

    HRESULT hr = S_OK;

    // make sure we have a buffer with sufficient space
    hr = EnsureBufferSize(ppbBuffer, *piBuffer + sizeof(DWORD));
    ExitOnFailure(hr, "Failed to ensure buffer size.");

    // copy data to buffer
    *(DWORD*)(*ppbBuffer + *piBuffer) = dw;
    *piBuffer += sizeof(DWORD);

LExit:
    return hr;
}

extern "C" HRESULT BuffWriteNumber64(
    __inout BYTE** ppbBuffer,
    __inout SIZE_T* piBuffer,
    __in DWORD64 dw64
    )
{
    Assert(ppbBuffer);
    Assert(piBuffer);

    HRESULT hr = S_OK;

    // make sure we have a buffer with sufficient space
    hr = EnsureBufferSize(ppbBuffer, *piBuffer + sizeof(DWORD64));
    ExitOnFailure(hr, "Failed to ensure buffer size.");

    // copy data to buffer
    *(DWORD64*)(*ppbBuffer + *piBuffer) = dw64;
    *piBuffer += sizeof(DWORD64);

LExit:
    return hr;
}

extern "C" HRESULT BuffWriteString(
    __inout BYTE** ppbBuffer,
    __inout SIZE_T* piBuffer,
    __in_z LPCWSTR scz
    )
{
    Assert(ppbBuffer);
    Assert(piBuffer);
    Assert(scz);

    HRESULT hr = S_OK;
    DWORD cch = (DWORD)lstrlenW(scz);
    SIZE_T cb = cch * sizeof(WCHAR);

    // make sure we have a buffer with sufficient space
    hr = EnsureBufferSize(ppbBuffer, *piBuffer + (sizeof(DWORD) + cb));
    ExitOnFailure(hr, "Failed to ensure buffer size.");

    // copy character count to buffer
    *(DWORD*)(*ppbBuffer + *piBuffer) = cch;
    *piBuffer += sizeof(DWORD);

    // copy data to buffer
    memcpy_s(*ppbBuffer + *piBuffer, cb, scz, cb);
    *piBuffer += cb;

LExit:
    return hr;
}

extern "C" HRESULT BuffWriteStream(
    __inout BYTE** ppbBuffer,
    __inout SIZE_T* piBuffer,
    __in BYTE* pbStream,
    __in SIZE_T cbStream
    )
{
    Assert(ppbBuffer);
    Assert(piBuffer);
    Assert(pbStream);

    HRESULT hr = S_OK;
    DWORD64 cb = cbStream;

    // make sure we have a buffer with sufficient space
    hr = EnsureBufferSize(ppbBuffer, *piBuffer + cbStream + sizeof(DWORD64));
    ExitOnFailure(hr, "Failed to ensure buffer size.");

    // copy byte count to buffer
    *(DWORD64*)(*ppbBuffer + *piBuffer) = cb;
    *piBuffer += sizeof(DWORD64);

    // copy data to buffer
    memcpy_s(*ppbBuffer + *piBuffer, cbStream, pbStream, cbStream);
    *piBuffer += cbStream;

LExit:
    return hr;
}


// helper functions

static HRESULT EnsureBufferSize(
    __inout BYTE** ppbBuffer,
    __in SIZE_T cbSize
    )
{
    HRESULT hr = S_OK;
    SIZE_T cbTarget = ((cbSize % BUFFER_INCREMENT) + 1) * BUFFER_INCREMENT;

    if (*ppbBuffer)
    {
        if (MemSize(*ppbBuffer) < cbTarget)
        {
            LPVOID pv = MemReAlloc(*ppbBuffer, cbTarget, TRUE);
            ExitOnNull(pv, hr, E_OUTOFMEMORY, "Failed to reallocate buffer.");
            *ppbBuffer = (BYTE*)pv;
        }
    }
    else
    {
        *ppbBuffer = (BYTE*)MemAlloc(cbTarget, TRUE);
        ExitOnNull(*ppbBuffer, hr, E_OUTOFMEMORY, "Failed to allocate buffer.");
    }

LExit:
    return hr;
}
