//-------------------------------------------------------------------------------------------------
// <copyright file="BurnViewTestShim.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"

namespace MS
{
namespace Tools
{
namespace WindowsInstallerXml
{
namespace Test
{
namespace Bootstrapper
{
//
// Initialize - call IBurnView::Initialize().
//
void BurnViewTestShim::Initialize(BurnModelTestShim^ model, BurnControllerTestShim^ controller)
{
    interior_ptr<IBurnModel*> ppModel = &(model->m_pModel);
    interior_ptr<IBurnController*> ppController = &(controller->m_pController);
    interior_ptr<IBurnView*> ppView = &(this->m_pView);
    IBurnView* pThisView = *ppView;

    HRESULT hr = pThisView->Initialize(*ppModel, *ppController);
    if (FAILED(hr))
    {
        throw gcnew BurnTestShimException(hr, "Failed while loading view.");
    }

    this->m_fManuallyInitialized = TRUE;
}


//
// Run - call IBurnView::Run().
//
void BurnViewTestShim::Run()
{
    interior_ptr<IBurnView*> ppView = &(this->m_pView);
    IBurnView* pThisView = *ppView;

    HRESULT hr = pThisView->Run();
    if (FAILED(hr))
    {
        throw gcnew BurnTestShimException(hr, "Failed while running view.");
    }
}


//
// Uninitialize - call IBurnView::Uninitialize().
//
void BurnViewTestShim::Uninitialize()
{
    interior_ptr<IBurnView*> ppView = &(this->m_pView);
    IBurnView* pThisView = *ppView;

    pThisView->Uninitialize();
    this->m_fManuallyInitialized = FALSE;
}


//
// SetPackageProperty - call IBurnCore::SetPackageProperty().
//
void BurnViewTestShim::SetPackageProperty(String^ packageId, String^ name, String^ value)
{
    pin_ptr<const WCHAR> wzPackageId = PtrToStringChars(packageId);
    pin_ptr<const WCHAR> wzProperty = PtrToStringChars(name);
    pin_ptr<const WCHAR> wzValue = PtrToStringChars(value);
    interior_ptr<IBurnView*> ppView = &(this->m_pView);
    IBurnView* pThisView = *ppView;

    HRESULT hr;
    IBurnCore* pCore = NULL;
    try
    {
        hr = pThisView->QueryInterface(IID_PPV_ARGS(&pCore));
        if (FAILED(hr))
        {
            throw gcnew BurnTestShimException(hr, "Failed to query for core from view.");
        }

        hr = pCore->SetPackageProperty(wzPackageId, wzProperty, wzValue);
        if (FAILED(hr))
        {
            throw gcnew BurnTestShimException(hr, "Failed to set package property.");
        }
    }
    finally
    {
        ReleaseObject(pCore);
    }
}


//
// GetControlText - Make sure all the controls have been localized properly.
//
String^ BurnViewTestShim::GetControlText(int nControlIdx)
{
    HRESULT hr = S_OK;
    interior_ptr<IBurnView*> ppView = &(this->m_pView);
    IBurnView* pThisView = *ppView;
    LPWSTR wzText = NULL;
    String^ sControlText;
    
    hr = pThisView->GetControlText(nControlIdx, &wzText);

    // do not throw an exception on failure
    sControlText = gcnew String(wzText);

    return sControlText;
}


//
// Constructor
//
BurnViewTestShim::BurnViewTestShim() : m_fManuallyInitialized(FALSE)
{
    IBurnView* x = NULL;

    HRESULT hr = BurnCreateView(SW_SHOWNORMAL, &x);
    if (FAILED(hr))
    {
        throw gcnew BurnTestShimException(hr, "Failed to create view.");
    }

    this->m_pView = x;
}


BurnViewTestShim::BurnViewTestShim(BurnMode^ mode) : m_fManuallyInitialized(FALSE)
{
    IBurnView* x = NULL;

    HRESULT hr = E_INVALIDARG;
    if (BurnMode::Normal == *mode)
    {
        hr = BurnCreateView(SW_SHOWNORMAL, &x);
    }
    else if (BurnMode::Elevated == *mode)
    {
        hr = BurnCreateViewElevated(&x);
    }

    if (FAILED(hr))
    {
        throw gcnew BurnTestShimException(hr, "Failed to create view.");
    }

    this->m_pView = x;
}


//
// Destructor
//
BurnViewTestShim::~BurnViewTestShim()
{
    this->Destroy();
}


//
// Finalizer
//
BurnViewTestShim::!BurnViewTestShim()
{
    this->Destroy();
}


//
// Destroy - cleans up the object (called by destructor and finalizer).
//
void BurnViewTestShim::Destroy()
{
    // If we manually initialized, then automatically unitialize in the finalizer to be helpful.
    if (this->m_fManuallyInitialized)
    {
        this->Uninitialize();
    }

    ReleaseNullObject(this->m_pView);
}
}
}
}
}
}
