﻿// <copyright file="ModelTest.cs" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

namespace MS.Tools.WindowsInstallerXml.Test.Bootstrapper
{
    using System;

    using Microsoft.VisualStudio.TestTools.UnitTesting;

    /// <summary>
    /// Unit test for Burn model.
    /// </summary>
    [TestClass]
    public class ModelTest
    {
        [TestMethod]
        public void ModelCommandLineTest()
        {
            BurnModelTestShim model = new BurnModelTestShim(String.Empty);
            BurnCommand command = model.Command;
            Assert.AreEqual(BurnMode.Normal, model.Mode);
            Assert.AreEqual(BurnAction.Install, command.action);
            Assert.AreEqual(BurnDisplay.Full, command.display);
            Assert.AreEqual(BurnRestart.Prompt, command.restart);

            model = new BurnModelTestShim(@"path\to\setup.exe /quiet");
            command = model.Command;
            Assert.AreEqual(BurnMode.Normal, model.Mode);
            Assert.AreEqual(BurnAction.Install, command.action);
            Assert.AreEqual(BurnDisplay.None, command.display);
            Assert.AreEqual(BurnRestart.Automatic, command.restart);

            model = new BurnModelTestShim(@"path\to\setup.exe /passive /norestart");
            command = model.Command;
            Assert.AreEqual(BurnMode.Normal, model.Mode);
            Assert.AreEqual(BurnAction.Install, command.action);
            Assert.AreEqual(BurnDisplay.Passive, command.display);
            Assert.AreEqual(BurnRestart.Never, command.restart);

            model = new BurnModelTestShim(@"path\to\setup.exe /forcerestart /uninstall");
            command = model.Command;
            Assert.AreEqual(BurnMode.Normal, model.Mode);
            Assert.AreEqual(BurnAction.Uninstall, command.action);
            Assert.AreEqual(BurnDisplay.Full, command.display);
            Assert.AreEqual(BurnRestart.Always, command.restart);

            model = new BurnModelTestShim(@"path\to\setup.exe /log path\to\setup.log");
            command = model.Command;
            Assert.AreEqual(BurnMode.Normal, model.Mode);
            Assert.AreEqual(BurnAction.Install, command.action);
            Assert.AreEqual(BurnDisplay.Full, command.display);
            Assert.AreEqual(BurnRestart.Prompt, command.restart);
            Assert.AreEqual(@"path\to\setup.log", model.LogPath);

            string pipeName = String.Concat("BurnPipe.", Guid.NewGuid().ToString("N"));
            string token = Guid.NewGuid().ToString("B");
            model = new BurnModelTestShim(String.Format(@"path\to\setup.exe /burn.elevated {0} {1}", pipeName, token));
            Assert.AreEqual(BurnMode.Elevated, model.Mode);
            Assert.AreEqual(pipeName, model.ElevatedPipeName);
            Assert.AreEqual(token, model.ElevatedToken);
        }

        [TestMethod]
        [DeploymentItem(@"data\setup.exe")]
        public void ModelLoadTest()
        {
            using (BurnModelTestShim model = new BurnModelTestShim(String.Empty))
            {
                model.Load("setup.exe");
                Assert.AreEqual<uint>(3, model.PackageCount);

                BurnPackageTestShim package = model.GetPackage(1); // remember package indices are 1 based, so this gets the first package.
                Assert.AreEqual(new Version("1.1.2.3"), package.PackageVersion);
                Assert.IsTrue(package.IsPerMachine);

                BurnPayloadTestShim payload = package.GetPayload(1); // remember payloads indices are also 1 based.
                Assert.AreEqual(PayloadSourceType.Embedded, payload.SourceType);
                Assert.AreEqual("setup.exe", payload.SourcePath);

                BurnMsiPackageTestShim msiPackage = package as BurnMsiPackageTestShim;
                Assert.IsNotNull(msiPackage);
                Assert.AreEqual(package.PackageVersion, msiPackage.PackageVersion);
                Assert.AreEqual(new Guid("834bd917-0dae-419b-a8fc-2065eb47481a"), new Guid(msiPackage.ProductCode));
            }
        }

        [TestMethod]
        [DeploymentItem(@"data\setup.exe")]
        public void ModelPropertyTest()
        {
            using (BurnModelTestShim model = new BurnModelTestShim(@"path\to\setup.exe PACKAGE2TARGETDIR=C:\Pkg2"))
            {
                model.Load("setup.exe");

                BurnPackageTestShim package = model.GetPackage(2); // remember package indices are 1 based, so this gets the second package.
                Assert.IsFalse(package.IsPerMachine);

                BurnPropertiesTestShim properties = package.Properties;

                Assert.AreEqual(1, properties.Count);
                Assert.AreEqual(@"C:\Pkg2", properties.GetStringValueByIndex(0));
                Assert.AreEqual(@"C:\Pkg2", properties.GetStringValueByName("TARGETDIR"));
            }
        }

        [TestMethod]
        [DeploymentItem(@"data\setup.exe")]
        public void ModelChainLogTest()
        {
            using (BurnModelTestShim model = new BurnModelTestShim(@"path\to\setup.exe"))
            {
                model.Load("setup.exe");

                Assert.AreEqual(Environment.ExpandEnvironmentVariables(@"%TEMP%\foo\global.log"), model.LogPath);
            }
        }

        [TestMethod]
        [DeploymentItem(@"data\setup.exe")]
        public void ModelPackageLogTest()
        {
            using (BurnModelTestShim model = new BurnModelTestShim(@"path\to\setup.exe"))
            {
                model.Load("setup.exe");

                BurnPackageTestShim package = model.GetPackage(1); // remember package indices are 1 based, so this gets the first package.

                Assert.AreEqual(Environment.ExpandEnvironmentVariables(@"%TEMP%\foo\pkg1v.log"), package.LogPath);
                Assert.AreEqual(LoggingLevel.Verbose, package.LogLevel);
            }
        }

        [TestMethod]
        [DeploymentItem(@"data\setup_default_logging.exe")]
        public void ModelDefaultLoggingTest()
        {
            using (BurnModelTestShim model = new BurnModelTestShim(@"path\to\setup.exe"))
            {
                model.Load("setup_default_logging.exe");
                Assert.IsTrue(model.LogPath.StartsWith(Environment.ExpandEnvironmentVariables(@"%TEMP%\vstesthost.0")));

                BurnPackageTestShim package1 = model.GetPackage(1); // remember package indices are 1 based, so this gets the first package.
                Assert.IsTrue(package1.LogPath.StartsWith(Environment.ExpandEnvironmentVariables(@"%TEMP%\vstesthost.0")));
                Assert.IsTrue(package1.LogPath.Contains("Package1."));
                Assert.AreEqual(LoggingLevel.Normal, package1.LogLevel);

                BurnPackageTestShim package2 = model.GetPackage(2);
                Assert.IsTrue(package2.LogPath.StartsWith(Environment.ExpandEnvironmentVariables(@"%TEMP%\vstesthost.0")));
                Assert.IsTrue(package2.LogPath.Contains("Package2."));
                Assert.AreEqual(LoggingLevel.Debug, package2.LogLevel);

                BurnPackageTestShim package3 = model.GetPackage(3);
                Assert.AreEqual(Environment.ExpandEnvironmentVariables(@"%TEMP%\foo\pkg3_non_default.log"), package3.LogPath);
                Assert.AreEqual(LoggingLevel.Normal, package3.LogLevel);
            }
        }

        [TestMethod]
        [DeploymentItem(@"data\setup_chain_logging.exe")]
        public void ModelDefaultChainLoggingTest()
        {
            using (BurnModelTestShim model = new BurnModelTestShim(@"path\to\setup.exe"))
            {
                model.Load("setup_chain_logging.exe");
                Assert.AreEqual(model.LogPath, Environment.ExpandEnvironmentVariables(@"%TEMP%\foo\chain_defaultlog"));

                BurnPackageTestShim package1 = model.GetPackage(1); // remember package indices are 1 based, so this gets the first package.
                Assert.IsTrue(package1.LogPath.StartsWith(Environment.ExpandEnvironmentVariables(@"%TEMP%\foo\chain_defaultlog.Package1")));
                Assert.AreEqual(LoggingLevel.Normal, package1.LogLevel);

                BurnPackageTestShim package2 = model.GetPackage(2);
                Assert.IsTrue(package2.LogPath.StartsWith(Environment.ExpandEnvironmentVariables(@"%TEMP%\foo\chain_defaultlog.Package2")));
                Assert.AreEqual(LoggingLevel.Normal, package2.LogLevel);

                BurnPackageTestShim package3 = model.GetPackage(3);
                Assert.IsTrue(package3.LogPath.StartsWith(Environment.ExpandEnvironmentVariables(@"%TEMP%\foo\chain_defaultlog.Package3")));
                Assert.AreEqual(LoggingLevel.Normal, package3.LogLevel);
            }
        }
    }
}
