﻿// <copyright file="ViewTest.cs" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
//-------------------------------------------------------------------------------------------------

namespace MS.Tools.WindowsInstallerXml.Test.Bootstrapper
{
    using System;
    using System.IO;
    using System.IO.Pipes;
    using System.Text;
    using System.Threading;

    using Microsoft.VisualStudio.TestTools.UnitTesting;

    /// <summary>
    /// Unit test for Burn model.
    /// </summary>
    [TestClass]
    public class ViewTest
    {
        [TestMethod]
        [DeploymentItem(@"data\setup.exe")]
        public void ViewLoadTest()
        {
            using (BurnModelTestShim model = new BurnModelTestShim(String.Empty))
            {
                using (BurnViewTestShim view = new BurnViewTestShim())
                {
                    using (BurnControllerTestShim controller = new BurnControllerTestShim(model, view))
                    {
                        model.Load("setup.exe");
                        view.Initialize(model, controller);
                        view.SetPackageProperty("Package1", "property", "value");

                        BurnPackageTestShim package = model.GetPackage(1);
                        BurnPropertiesTestShim properties = package.Properties;
                        Assert.AreEqual("value", properties.GetStringValueByName("property"));
                    }
                }
            }
        }

        [TestMethod]
        [DeploymentItem(@"data\setup.exe")]
        public void ViewLocalizationTest()
        {
            using (BurnModelTestShim model = new BurnModelTestShim("/lang ps-ps"))
            {
                using (BurnViewTestShim view = new BurnViewTestShim())
                {
                    using (BurnControllerTestShim controller = new BurnControllerTestShim(model, view))
                    {
                        model.Load("setup.exe");
                        view.Initialize(model, controller);
                        string sText;

                        sText = view.GetControlText(0);
                        Assert.AreEqual("[Tèrms òf Úsé ÅÇÉ ÅÇ]", sText);

                        sText = view.GetControlText(1);
                        Assert.AreEqual("[Ìñståll ÅÇÉ ]", sText);

                        sText = view.GetControlText(2);
                        Assert.AreEqual("[Rêpªïr ÅÇÉ]", sText);

                        sText = view.GetControlText(3);
                        Assert.AreEqual("[Ûñìñställ ÅÇÉ Å]", sText);

                        sText = view.GetControlText(4);
                        Assert.AreEqual("[Çlòsë ÅÇÉ]", sText);

                        sText = view.GetControlText(5);
                        Assert.AreEqual("[ÔK ÅÇ]", sText);

                        sText = view.GetControlText(6);
                        Assert.AreEqual("[Çæñ¢èl ÅÇÉ]", sText);

                        sText = view.GetControlText(7);
                        Assert.AreEqual("[Rétrÿ ÅÇÉ]", sText);

                        sText = view.GetControlText(8);
                        Assert.AreEqual("[Äþòrt ÅÇÉ]", sText);

                        sText = view.GetControlText(9);
                        Assert.AreEqual("[Ìgñórè ÅÇÉ]", sText);

                        sText = view.GetControlText(10);
                        Assert.AreEqual("[Ýês ÅÇ]", sText);

                        sText = view.GetControlText(11);
                        Assert.AreEqual("[Ñö ÅÇ]", sText);

                        sText = view.GetControlText(12);
                        Assert.AreEqual("[Prõgréss ÅÇÉ ]", sText);

                        sText = view.GetControlText(13);
                        Assert.AreEqual("[Ìñîtìálízîñg... ÅÇÉ ÅÇÉ]", sText);
                    }
                }
            }
        }

        [TestMethod]
        public void ViewElevatedTest()
        {
            string pipeName = String.Concat("BurnPipe.", Guid.NewGuid().ToString("N"));
            string token = Guid.NewGuid().ToString("N");
            using (BurnModelTestShim model = new BurnModelTestShim(String.Format(@"path\to\setup.exe /burn.elevated {0} {1}", pipeName, token)))
            {
                using (BurnViewTestShim view = new BurnViewTestShim(BurnMode.Elevated))
                {
                    using (BurnControllerTestShim controller = new BurnControllerTestShim(model, view))
                    {
                        ElevatedViewParentContext context = new ElevatedViewParentContext() { PipeName = pipeName, Token = token };
                        Thread t = new Thread(new ParameterizedThreadStart(ViewTest.ElevatedViewParentThread));

                        t.Start(context);
                        context.InitializedEvent.WaitOne();

                        view.Initialize(model, controller);
                        t.Join();

                        Assert.IsTrue(context.Completed);
                    }
                }
            }
        }

        private static void ElevatedViewParentThread(object obj)
        {
            ElevatedViewParentContext context = (ElevatedViewParentContext)obj;
            using (NamedPipeServerStream pipeServer = new NamedPipeServerStream(context.PipeName))
            {
                context.InitializedEvent.Set();

                try
                {
                    pipeServer.WaitForConnection();

                    byte[] token = Encoding.Unicode.GetBytes(context.Token);
                    byte[] tokenSize = BitConverter.GetBytes((UInt32)token.Length);

                    pipeServer.Write(tokenSize, 0, tokenSize.Length);
                    pipeServer.Write(token, 0, token.Length);

                    context.Completed = true;
                }
                catch (IOException)
                {
                }
            }
        }

        /// <summary>
        /// Context used by parent of elevated view.
        /// </summary>
        private class ElevatedViewParentContext
        {
            private ManualResetEvent initialized = new ManualResetEvent(false);

            public string PipeName { get; set; }
            public string Token { get; set; }
            public bool Completed { get; set; }
            public ManualResetEvent InitializedEvent { get { return this.initialized; } }
        }
    }
}
