//-----------------------------------------------------------------------
// <copyright file="FileTests.cs" company="Microsoft">
//     Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
// <summary>Tests for authoring files into a package</summary>
//-----------------------------------------------------------------------

namespace Microsoft.Tools.WindowsInstallerXml.Test.Tests.Integration.BuildingPackages.Files
{
    using System;
    using System.IO;
    using System.Text;
    using System.Collections.Generic;
    using Microsoft.VisualStudio.TestTools.UnitTesting;

    using Microsoft.Tools.WindowsInstallerXml.Test;

    /// <summary>
    /// Tests for authoring files into a package
    /// </summary>
    [TestClass]
    public class FileTests
    {
        private static readonly string TestDataDirectory = Environment.ExpandEnvironmentVariables(@"%WIX_ROOT%\test\data\Integration\BuildingPackages\Files\FileTests");

        [TestMethod]
        [Description("Verify the simple use of the File element")]
        [Priority(1)]
        public void SimpleFile()
        {
            string sourceFile = Path.Combine(FileTests.TestDataDirectory, @"SimpleFile\product.wxs");
            string msi = Builder.BuildPackage(sourceFile, "test.msi");
            Verifier.VerifyResults(Path.Combine(FileTests.TestDataDirectory, @"SimpleFile\expected.msi"), msi, "BindImage", "File", "MsiAssembly", "MsiAssemblyName");
        }

        [TestMethod]
        [Description("Verify a companion file can be specified")]
        [Priority(2)]
        [Ignore]
        public void CompanionFile()
        {
        }

        [TestMethod]
        [Description("Verify that the publicKeyToken in the MsiAssemblyName table is not 'NEUTRAL'")]
        [TestProperty("Bug Link", "http://sourceforge.net/tracker/index.php?func=detail&aid=1801826&group_id=105970&atid=642714")]
        [Priority(3)]
        public void AssemblyNullPublicKey()
        {
            // Run Candle
            Candle candle = new Candle();
            candle.SourceFiles.Add(Path.Combine(FileTests.TestDataDirectory, @"AssemblyNullPublicKey\product.wxs"));
            candle.Run();

            // Run Light
            Light light = new Light(candle);

            string assembly1 = Path.Combine(WixTests.SharedFilesDirectory, @"TestBinaries\bin\Assembly1.dll");
            light.ExpectedWixMessages.Add(new WixMessage(282, String.Format("Assembly {0} in component Component1 has no strong name and has been marked to be placed in the GAC. All assemblies installed to the GAC must have a valid strong name.", assembly1), WixMessage.MessageTypeEnum.Error));
            light.ExpectedExitCode = 282;

            light.Run();
        }

        [TestMethod]
        [Description("Verify that the specifing @DefaultVersion for a non versioned file does not cause its addition to the msifilehash table")]
        [TestProperty("Bug Link", "http://sourceforge.net/tracker/index.php?func=detail&aid=1666461&group_id=105970&atid=642714")]
        [Priority(3)]
        public void DefaultVersion()
        {
            Candle candle = new Candle();
            candle.SourceFiles.Add(Path.Combine(FileTests.TestDataDirectory, @"DefaultVersion\product.wxs"));
            candle.Run();

            Light light = new Light(candle);
            light.ExpectedWixMessages.Add(new WixMessage(1103, "The DefaultVersion '3.0.0.0' was used for file 'TextFile1.txt' which has no version. No entry for this file will be placed in the MsiFileHash table. For unversioned files, specifying a version that is different from the actual file may result in unexpected versioning behavior during a repair or while patching. Version the resource to eliminate this warning.", WixMessage.MessageTypeEnum.Warning));
            light.Run();

            // verify unversioned file is not added to the msifilehash table
            Verifier.VerifyNotTableExists(light.OutputFile, "MsiFileHash");
        }
    }
}